#!/usr/bin/nodejs

const http = require('http');

const DomTest = require('./domtest');
const url = require('url');

// check argument
const prog = require('commander');
prog
  .version('0.1.0')
  .option('-p, --port <port>', 'port number to listen', parseInt)
  .option('-H, --host <hostname>', 'host to listen')
  .parse(process.argv);

const port = prog.port || 4000;
const host = prog.host || '127.0.0.1';

// error handler
const server = http.createServer();
server.on('clientError', (err, socket) => {
  socket.end('HTTP/1.1 400 Bad Request\r\n\r\n');
});

// request handler
server.on('request', onRequest);

console.log(`app.js: Listening to http://${host}:${port}/`);
server.listen(port, host);


function onRequest(req, resp) {
  console.log(`${Date.now()}:${req.method} ${req.url}`);
  // check request
  if (req.method != 'POST') {
      sendJSON(resp, 400, {error: 1, message: "invalid request"});
      return;
  }
  
  const contentType = req.headers['Content-Type'] || 'application/json';
  if (contentType != 'application/json') {
      sendJSON(resp, 400, {error: 1, message: "Content-Type is not application/json"});
      return;
  }
  
  // parse request
  var requestBody = '';
  req.on('data', (chunk) => {
    requestBody += chunk.toString();
  });

  req.on('end', () => {
    req.parsedBody = JSON.parse(requestBody);
    processRequest(req, resp);
  });

}

function sendJSON(resp, code, obj) {
  const message = JSON.stringify(obj);
  resp.writeHead(code,
                 { 'Content-Type': 'application/json',
                   'Content-Length': message.length });
  resp.end(message);
}

function processRequest(req, resp) {
  const targetUrl = req.parsedBody.url;

  // check url
  try {
    const parsed_url = url.parse(targetUrl);
    if (!parsed_url.host) {
      sendJSON(resp, 400, {error: 1, message: "invalid url"});
      return;
    }
  } catch (err) {
    sendJSON(resp, 400, {error: 1, message: "url parsing error"});
    return;
  }

  const options = {};
  if (req.parsedBody.ignoreMessages) {
    options.ignoreMessages = req.parsedBody.ignoreMessages;
  }
  if (req.parsedBody.finishTimeout) {
    options.finishTimeout = req.parsedBody.finishTimeout;
  }
  
  // start test
  console.log(`run test for ${targetUrl}`);
  const domtest = new DomTest();
  domtest.run(targetUrl, options, (error) => {
    if (error) {
      sendJSON(resp, 500, {error: error, message: "domtest running error" });
      return;
    }

    const errors = stringifyErrors(domtest.runtimeErrors);
    sendJSON(resp, 200, {error: 0,
                         runtimeErrorCount: domtest.errorCounter,
                         runtimeErrors: errors, });
  });
}

function stringifyErrors(errors) {
  const result = [];
  errors.forEach((val) => {
    const rs = {};
    if (val.type) {
      rs.type = val.type;
    }
    if (val.detail) {
      rs.detail = val.detail.toString();
    }
    result.push(rs);
  });
  return result;
}
