package Newslash::Command::usermod;
use Mojo::Base 'Newslash::Command';

use Mojo::Util 'getopt';
use Data::Dumper;

has description => 'Change user properties';
has usage => sub { shift->extract_usage };

has experimental => 1;
has danger => 0;

sub run {
    my ($self, @args) = @_;
    return if !$self->check_force_option(\@args);

    my $app = $self->app;
    my $seclev;
    my $author;
    my $password;

    getopt \@args,
      'u|username=s' => \my $nickname,
      'password=s' => \$password,
      'seclev=i'     => \$seclev,
      'author=i'     => \$author;

    if (!$nickname) {
        $app->log->error("usermod: nickname not given!");
        return;
    }

    my $users = $app->model('users');
    my $user = $users->select(nickname => $nickname);
    if (!$user) {
        $app->log->error("usermod: user '$nickname' not found!");
        return;
    }

    my %param;
    if (defined $seclev) { $param{"seclev"} = $seclev; };
    if (defined $password) { $param{"passwd"} = $password; };
    if (defined $author) { $param{"author"} = $author; };

    my $rs = $users->update(uid => $user->{uid}, %param);
    if (!$rs) {
        my $err = $users->last_error || "";
        $app->log->info("usermod: update user info failed - $err");
        return;
    }
    $app->log->info("usermod: update user info done ($rs)");
}

1;

=encoding utf8

=head1 NAME

Newslash::Command::usermod - newslash administer command (usermod)

=head1 SYNOPSIS

  Usage: APPLICATION usermod [OPTIONS]

      ./newslash_web usermod -u foobar --seclev 100 --author 1 --password hogehoge

    Options:
      -u                     target's user name (nickname)
      --seclev <seclev>      set seclev
      --password <password>  update password
      --author <0/1>         add/remove author flag

=head1 DESCRIPTION

  L<Newslash::Command::usermod> updates newslash's user database.

=head1 ATTRIBUTES

L<Newslash::Command::usermod> inherits all attributes from
L<Newslash::Command> and implements the following new ones.

=head2 description

  my $description = $daemon->description;
  $daemon         = $daemon->description('Foo');

Short description of this command, used for the command list.

=head2 usage

  my $usage = $daemon->usage;
  $daemon   = $daemon->usage('Foo');

Usage information for this command, used for the help screen.

=head1 METHODS

L<Newslash::Command::usermod> inherits all methods from
L<Newslash::Command> and implements the following new ones.

=head2 run

  $daemon->run(@ARGV);

Run this command.

=head1 SEE ALSO

L<Mojolicious>, L<Mojolicious::Guides>, L<https://mojolicious.org>.

=cut
