package Newslash::Model::Discussions;
use Newslash::Model::Base -base;

use Data::Dumper;
use Mojo::Log;
use POSIX qw(strftime);

use constant TABLE_COLUMNS => qw(
                                    id
                                    dkid
                                    stoid
                                    sid
                                    title
                                    url
                                    topic
                                    ts
                                    type
                                    uid
                                    commentcount
                                    flags
                                    primaryskid
                                    last_update
                                    approved
                                    commentstatus
                                    archivable
                      );

#========================================================================

=head2 select($query_type, $value)

get a discussion.

=over 4

=item Parameters

=over 4

=item $query_type

query key, one of "id", "sid", "stoid"

=item $value

value for query

=back

=item Return value

HASH of discussion contents

=back

=cut

sub select {
  my ($self, $query_type, $value) = @_;

  if ($query_type !~ m/\A(id|sid|stoid|discussion_id)\z/) {
      return;
  }
  $query_type = "id" if $query_type eq "discussion_id";

  my $dbh = $self->connect_db;
  my $sql = <<"EOSQL";
SELECT discussions.*, discussion_kinds.name AS kind
  FROM discussions
  LEFT JOIN discussion_kinds ON discussions.dkid = discussion_kinds.dkid
  WHERE discussions.$query_type = ?
EOSQL

  my $sth = $dbh->prepare($sql);

  $sth->execute($value);
  my $discuss = $sth->fetchrow_hashref;
  $sth->finish;
  $self->disconnect_db();

  if (!$discuss) {
    return undef;
  }

  if ($discuss->{url} =~ m|^/poll/|) {
      $discuss->{dkid} = 5;
      $discuss->{kind} = "poll";
  }

  $discuss->{discussion_id} = $discuss->{id};
  return $discuss;
}

#========================================================================

=head2 create(\%options)

create a discussion.

=over 4

=item Parameters

=over 4

=item \%options

options

=item Return value

discussion id

=back

=cut

# extract from "discussion_kinds" table
my $discussion_kinds = {
                        feed => 7,
                        journal => 3,
                        "journal-story" => 4,
                        poll => 5,
                        project => 8,
                        story => 1,
                        submission => 6,
                        user_createduse => 2
                       };

sub create {
    #my ($self, $params, $options) = @_;
    my $self = shift;
    my $params = {@_};

    return if $self->check_readonly;
    if (!$params
        || !$params->{title}
        || length $params->{title} == 0
        || !$params->{url}) {
        return undef;
    }
    my $default_comment_status = "enabled";
    my $mainpage_skid = 1; # mainpage

    my $dkid = 1;
    if ($params->{kind}) {
        # getDescriptions('discussion_kinds') method is used in slash's original implement,
        # but here, use static table.
        $dkid = $discussion_kinds->{$params->{kind}} || 1;
    }
    my $uid = $params->{uid} || 0;

    my $dbh = $self->connect_db;
    my $sql = <<"EOSQL";
INSERT INTO discussions
    (dkid, stoid, sid, title, url, topic, ts, type, uid, primaryskid, commentstatus)
  VALUES
    (?,    ?,     ?,   ?,     ?,   ?,     ?,  ?,    ?,   ?,           ?)
EOSQL
    my $rs = $dbh->do($sql, undef,
                      $dkid, #dkid
                      $params->{stoid} || 0,
                      $params->{sid} || '',
                      $params->{title},
                      $params->{url},
                      $params->{topic} || 0,
                      $params->{ts} || strftime("%F %T", gmtime),
                      $params->{type} || "open", #type
                      $uid, #uid
                      $params->{primaryskid} || $mainpage_skid,
                      $params->{commentstatus} || $default_comment_status,
                     );
    my $discussion_id = 0;
    if ($rs) {
        $discussion_id = $dbh->last_insert_id(undef, undef, undef, undef);
    } else {
        $self->set_errorno($dbh->{mysql_errno});
    }

    $self->disconnect_db();

    return if !$discussion_id;
    return $discussion_id;
}

sub update {
    my $self = shift;
    my $params = {@_};
    return if $self->check_readonly;

    if (!$params->{id}) {
        if ($params->{discussion_id}) {
            $params->{id} = $params->{discussion_id};
        }
        else {
            return;
        }
    }
    my $id = delete $params->{id};
    my $options = {};
    if ($params->{options}) {
        $options = delete $params->{options};
    }

    my @keys = grep {defined $params->{$_}} TABLE_COLUMNS;
    my @values = map {$params->{$_}} @keys;

    my $dbh = $options->{dbh} || $self->connect_db;
    my $set_clause = join(", ", map {"$_ = ?"} @keys);
    my $sql = <<"EOSQL";
UPDATE discussions SET $set_clause WHERE id = ?
EOSQL

    push @values, $id;
    my $rs = $dbh->do($sql, undef, @values);
    if (!$rs) {
        $self->set_errorno($dbh->{mysql_errno});
    }

    if (!$options->{dbh}) {
        $self->disconnect_db();
    }

    return $rs;

}

sub is_comment_allowed {
    my $self = shift;
    my $params = {@_};
    my $uid = $params->{uid};
    my $discuss = $params->{discussion};
    my $did = $params->{discussion_id} || $params->{id};
    return if !$uid || (!$did && !$discuss);

    $discuss = $self->select(discussion_id => $did) if !$discuss;

    my $relations = $self->new_instance_of("Relations");
    my $users = $self->new_instance_of("Users");
    return if !$discuss;

    my $status = $discuss->{commentstatus};
    my $the_uid = $discuss->{uid};

    if ($status eq "disabled" || $discuss->{type} eq 'archived') {
        return 0;
    }

    if ($status eq "enabled") {
        return 1;
    }

    if ($uid == $the_uid) {
        return 1;
    }

    if ($status eq "friends_only") {
        return 1 if $relations->is_friend(of => $the_uid, uid => $uid);
    }

    if ($status eq "friends_fof_only") {
        return 1 if $relations->is_friend_or_fof(of => $the_uid, uid => $uid);
    }

    if ($status eq "no_foe") {
        return 1 if !$relations->is_foe(of => $the_uid, uid => $uid);
    }

    if ($status eq "friends_fof_only") {
        return 1 if !$relations->is_foe_or_eof(of => $the_uid, uid => $uid);
    }

    if ($status eq "logged_in") {
        return 1 if !$users->is_anonymous_uid($uid);
    }

    return 0;
}

1;
