package Newslash::Model::Feeds;
use Newslash::Model::Base -base;

use Newslash::Model::Feeds::Items;
use XML::Feed;
use LWP::UserAgent;



sub items {
    return shift->new_instance_of("Newslash::Model::Feeds::Items");
}

sub on_start_up {
    my $self = shift;

    # create tables
    return 1 if $self->check_readonly;

    my $sqls = {
                ns_feed_items => <<"EOSQL",
CREATE TABLE IF NOT EXISTS ns_feed_items (
  id           mediumint(8) unsigned NOT NULL PRIMARY KEY AUTO_INCREMENT,
  feed_id      mediumint(8) unsigned NOT NULL,
  url          varchar(191) UNIQUE NOT NULL,
  title        mediumtext NOT NULL,
  content      mediumtext,
  content_type varchar(128),
  modified     datetime,
  create_time  timestamp NOT NULL DEFAULT '1980-01-01 00:00:00'
) DEFAULT CHARSET=utf8mb4
EOSQL
                ns_feeds => <<"EOSQL",
CREATE TABLE IF NOT EXISTS ns_feeds (
  feed_id             mediumint(8) unsigned NOT NULL PRIMARY KEY AUTO_INCREMENT,
  name                varchar(128) NOT NULL UNIQUE,
  url                 varchar(191) UNIQUE,
  title               mediumtext,
  description         mediumtext,
  create_time         datetime NOT NULL DEFAULT '1980-01-01 00:00:00',
  update_time         timestamp DEFAULT CURRENT_TIMESTAMP,
  subscribers         mediumint NOT NULL DEFAULT 0,
  latest_fetch_time   datetime,
  latest_fetch_result smallint unsigned,
  status              enum('active', 'inactive', 'deleted') DEFAULT 'active'
) DEFAULT CHARSET=utf8mb4
EOSQL
               };

    my $dbh = $self->connect_db;
    for my $table_name (keys %$sqls) {
        if (!$self->table_exists($table_name)) {

            my $rs = $dbh->do($sqls->{$table_name}, undef);
            if ($rs) {
                $self->warn("create table: $table_name");
            }
            else {
                $self->warn("cannot create table: $table_name");
            }
        }
    }
    $self->disconnect_db;


    return 1;
}

sub select {
    my $self = shift;
    my $params = {@_};

    my $unique_keys = { id => "ns_feeds.feed_id",
                        name => "ns_feeds.name",
                        feed_id => "ns_feeds.feed_id",
                        url => "ns_feeds.url",
                      };
    my $keys = { create_time => "ns_feeds.create_time",
                 update_time => "ns_feeds.update_time",
                 status => "ns_feeds.status",
               };
    my $timestamp = "ns_feeds.create_time";

    my ($where_clause, $where_values, $unique) = $self->build_where_clause(unique_keys => $unique_keys,
                                                                           keys => $keys,
                                                                           timestamp => $timestamp,
                                                                           params => $params);
    my ($limit_clause, $limit_values) = $self->build_limit_clause(params => $params);
    my ($orderby_clause, $orderby_values) = $self->build_order_by_clause(keys => $keys,
                                                                         params => $params);

    my @attrs;
    push @attrs, @$where_values, @$limit_values, @$orderby_values;

    my $dbh = $self->connect_db;
    my $sql = <<"EOSQL";
SELECT * FROM ns_feeds
    $where_clause
    $orderby_clause
    $limit_clause
EOSQL

    #warn($sql);
    #warn(Dumper(@attrs));

    my $sth = $dbh->prepare($sql);
    $sth->execute(@attrs);
    my $feeds = $sth->fetchall_arrayref({});

    if (!$feeds) {
        $self->disconnect_db();
        return;
    }
    if (@$feeds == 0) {
        $self->disconnect_db();
        return $unique ? undef : [];
    }

    return $feeds->[0] if $unique;
    return $feeds;
}

sub create {
    my $self = shift;
    my $params = {@_};
    my $url = $params->{url};
    return if !$url;

    my $name = $params->{name};
    return if !$name;

    my $dbh = $self->connect_db;
    my $sql = <<"EOSQL";
INSERT INTO ns_feeds
    (url, name, title, description, create_time, status)
  VALUES
    (?,   ?,    ?,     ?,           NOW(),       ?)
EOSQL

    my $rs = $dbh->do($sql, undef,
                      $url,
                      $name,
                      $params->{title} || "",
                      $params->{description} || "",
                      $params->{status} || "active",
                     );
    return if !$rs;
    my $id = $dbh->last_insert_id(undef, undef, undef, undef);
    return $id;
}

sub update {
    my $self = shift;
    my $params = {@_};
    my $feed_id = $params->{id} || $params->{feed_id};
    return if !$feed_id;

    my @cols;
    my @attrs;
    for my $k (qw|url name title description latest_fetch_time latest_fetch_result status|) {
        if (defined $params->{$k}) {
            push @cols, "$k = ?";
            push @attrs, $params->{$k};
        }
    }
    push @cols, "update_time = NOW()";

    my $columns = join(", ", @cols);
    my $sql = <<"EOSQL";
UPDATE ns_feeds 
  SET $columns
  WHERE feed_id = ?
EOSQL

    push @attrs, $feed_id;
    my $dbh = $self->connect_db;
    my $rs = $dbh->do($sql, undef, @attrs);
    return if !$rs;
    return $feed_id;
}

sub delete {
    my $self = shift;
    my $params = {@_};
    my $feed_id = $params->{id} || $params->{feed_id};
    return if !$feed_id;

    my $sql = <<"EOSQL";
DELETE FROM ns_feeds
  WHERE feed_id = ?
EOSQL

    my $dbh = $self->connect_db;
    my $rs = $dbh->do($sql, undef, $feed_id);
    return $rs;
}

sub fetch_url {
    my ($self, $url_ref, @opts) = @_;
    my $options = {};
    if (@opts == 1 && ref($opts[0]) eq "HASH") {
        $options = $opts[0];
    }
    else {
        $options = {@opts};
    }
    my $timeout = $options->{timeout} || $options->{fetch_timeout} || 60;

    my $url = $$url_ref;

    my $ua = LWP::UserAgent->new;
    if ($options->{http_proxy}) {
        $ua->proxy(http => $options->{http_proxy});
    }
    if ($options->{https_proxy}) {
        $ua->proxy(https => $options->{https_proxy});
    }

    my $req = HTTP::Request->new(GET => $url);
    $ua->timeout($timeout);
    my $res = $ua->request($req);

    if (!$res->is_success) {
        $self->set_error($res->message, $res->code);
        return;
    }

    my $content = $res->content;
    #my $feed = XML::Feed->parse(\$content);
    my $feed = eval { XML::Feed->parse(\$content) };
    if (!$feed) {
        $self->set_error("invalid_feed", -1);
        return;
    }

    my $new_url = $res->request->uri;
    if ($new_url ne $url) {
        $$url_ref = $url;
    }

    return $feed;
}

sub export_all {
    my $self = shift;
    my $target = shift;

    if (ref($target) eq "SCALAR") {
        $$target = $self->export_json(table => "ns_feeds",
                                      sort_key => "name",
                                      exclude => [qw(feed_id
                                                     update_time
                                                     subscribers
                                                     latest_fetch_time
                                                     latest_fetch_result)],
                                     );
        return 1;
    }
    my $output = $self->export_json(file => $target,
                                    table => "ns_feeds",
                                    sort_key => "name",
                                    exclude => [qw(feed_id
                                                   update_time
                                                   subscribers
                                                   latest_fetch_time
                                                   latest_fetch_result)],
                                   );
    return if !$output;
    return 1;
}

sub import_all {
    my $self = shift;
    my $target = shift;

    my $output;

    if (ref($target) eq "SCALAR") {
        my $rs = $self->import_json(table => "ns_feeds",
                                    json => $$target,
                                    unique_key => "name",
                                    exclude => [qw(feed_id
                                                   update_time
                                                   subscribers
                                                   latest_fetch_time
                                                   latest_fetch_result)],
                                     );
        return $rs;
    }
    my $rs = $self->import_json(table => "ns_feeds",
                                file => $target,
                                unique_key => "name",
                                exclude => [qw(feed_id
                                               update_time
                                               subscribers
                                               latest_fetch_time
                                               latest_fetch_result)],
                               );
    return $rs;
}

1;
