package Newslash::Model::Journals;
use Newslash::Model::Base -base;

#use Newslash::Model::Util;
use URI::Escape;
use Data::Dumper;
use POSIX qw(strftime);
use Newslash::Util::TextFormatter;

use constant JOURNAL_FORMATTER_TABLE => "ns_journal_formatter";

use constant DEFAULT_ALLOWED_TAGS => {
                                      'b' => [],
                                      'i' => [],
                                      'p' => [],
                                      'br' => [],
                                      'a' => ['href',],
                                      'ol' => ['start',],
                                      'ul' => [],
                                      'li' => [],
                                      'dl' => [],
                                      'dt' => [],
                                      'dd' => [],
                                      'em' => [],
                                      'strong' => [],
                                      'tt' => [],
                                      'blockquote' => ['title', 'cite',],
                                      'div' => [],
                                      'ecode' => [],
                                      'del' => [],
                                      'ins' => [],
                                      'sub' => [],
                                      'sup' => [],
                                      'quote' => [],
                                      'strike' => [],
                                     };

sub create_table {
    my $self = shift;
    my $dbh = $self->connect_db;
    my $table_name = JOURNAL_FORMATTER_TABLE;

    return 1 if $self->check_readonly;

    my $sql = <<"EOSQL";
CREATE TABLE IF NOT EXISTS $table_name (
  journal_id     int(8) unsigned NOT NULL,
  formatter      varchar(128)    NOT NULL,
  KEY journal_id  (journal_id)
) DEFAULT CHARSET=utf8mb4
EOSQL
    my $rs = $dbh->do($sql, undef);
    $self->disconnect_db;
    return 1 if $rs;

    $self->logger->warn("Journals: journal formatter table create failed.");
    return;
}

sub on_start_up {
    my $self = shift;
    if (!$self->table_exists(JOURNAL_FORMATTER_TABLE)) {
        $self->logger->warn("Journals: journal formatter table not exists, create it.");
        return $self->create_table;
    }
    return 1;
}


sub count {
    my $self = shift;
    my $params = {@_};
    my $join = 'LEFT JOIN firehose ON (journals.id = firehose.srcid AND firehose.type = "journal")';

    my @clauses;
    if ($params->{hide_future}) {
	push @clauses,  'journals.date < NOW()';
    }
    if ($params->{public_only}) {
	push @clauses, 'firehose.public != "no"';
    }
    my $where = join " AND ", @clauses;

    return $self->generic_count(table => "journals",
                                target => "id",
                                timestamp => "date",
                                join => $join,
                                where => $where,
                                @_);
}

sub count_by_author {
    my $self = shift;
    my $args = {@_};
    my $limit = $args->{limit} || 10;
    my $years = $args->{years};

    my @attr;
    my $where_clause = "";
    if ($years =~ m/^\d+$/ ) {
        $where_clause = "WHERE date >= DATE_ADD(NOW(), INTERVAL - ? YEAR)";
        push @attr, $years;
    }

    my $sql = <<"EOSQL";
SELECT users.*, COUNT(id) AS count
  FROM journals
  JOIN users USING(uid)
  $where_clause
  GROUP BY uid
  ORDER BY COUNT(id) DESC
  LIMIT ?
EOSQL
    push @attr, $limit;

    my $dbh = $self->connect_db;
    my $sth = $dbh->prepare($sql);
    $sth->execute(@attr);
    my $rs = $sth->fetchall_arrayref({});

    if (!defined $rs) {
        $self->set_error("select failed", $dbh->errstr, $dbh->err);
        $self->disconnect_db;
        return;
    }
    $self->disconnect_db;
    return $rs;
}



#========================================================================

=head2 select()

get a journal.

=over 4

=item Parameters

=over 4

=item $query_type

query key, "id"

=item $value

value for query

=back

=item Return value

HASH of story contents

=back

=cut

sub select {
    my $self = shift;
    my $params = {@_};

    my $unique_keys = { id => "journals.id",
                        journal_id => "journals.id",
                        discussion_id => "journals.discussion",
                        discussion => "journals.discussion",
                      };
    my $keys = { uid => "journals.uid",
                 user_id => "journals.uid",
                 karma => "users_info.karma",
                 popularity => "firehose.popularity",
               };
    my $datetime_keys = { create_time => 'journals.date',
                          update_time => 'journals.last_update',
                        };

    my ($where_clause, $where_values, $unique) = $self->build_where_clause(unique_keys => $unique_keys,
                                                                           keys => $keys,
                                                                           datetime_keys => $datetime_keys,
                                                                           timestamp => 'journals.date',
                                                                           params => $params);
    my ($limit_clause, $limit_values) = $self->build_limit_clause(params => $params);
    my ($orderby_clause, $orderby_values) = $self->build_order_by_clause(keys => $datetime_keys,
                                                                        params => $params);
    # show future story?
    my @where_clauses;
    if (!$params->{show_future}) {
        push @where_clauses, "journals.date <= NOW()";
    }

    # show non-public story?
    if (!$params->{show_nonpublic} && !$params->{public}) {
        push @where_clauses, "firehose.public != 'no'";
    }

    if (@where_clauses) {
        if ($where_clause) {
            $where_clause = $where_clause . " AND ";
        }
        else {
            $where_clause = "WHERE ";
        }
        $where_clause = $where_clause . join(" AND ", @where_clauses);
    }

    my @attr;
    push @attr, @$where_values, @$limit_values, @$orderby_values;

    my $dbh = $self->connect_db;
    my $formatter_table = JOURNAL_FORMATTER_TABLE;
    my $sql = <<"EOSQL";
SELECT
    journals.*,
    journals_text.*,
    users.nickname as author,
    firehose.toptags,
    firehose.public,
    firehose.popularity,
    firehose.editorpop,
    firehose.neediness,
    firehose.activity,
    firehose.globjid,
    firehose.id AS fhid,
    discussions.commentcount AS comment_count,
    discussions.type AS discussion_type,
    discussions.commentstatus AS comment_status,
    $formatter_table.formatter
  FROM journals
    LEFT JOIN journals_text ON journals.id = journals_text.id
    LEFT JOIN users ON journals.uid = users.uid
    LEFT JOIN users_info ON journals.uid = users_info.uid
    LEFT JOIN firehose ON (journals.id = srcid AND firehose.type = 'journal')
    LEFT JOIN topics ON journals.tid = topics.tid
    LEFT JOIN discussions ON journals.discussion = discussions.id
    LEFT JOIN $formatter_table ON $formatter_table.journal_id = journals.id
  $where_clause
  $orderby_clause
  $limit_clause
EOSQL

    my $sth = $dbh->prepare($sql);
    $sth->execute(@attr);
    my $rs = $sth->fetchall_arrayref({});
    $self->disconnect_db() if !$params->{dbh};

    if (@$rs == 0) {
        return if $unique;
        return [];
    }

    my $tags = $self->new_instance_of("Tags");
    my $whole_tags = $tags->select_for_items("journal", "id", $rs);
    if ($whole_tags) {
        for my $item (@$rs) {
            $item->{tags} = $whole_tags->{$item->{id}} || [];
        }
    }

    for my $j (@$rs) {
        $self->_generalize($j);
    }

    return $rs->[0] if $unique;
    return $rs;
}

sub generate_bodytext {
    my ($self, $journals) = @_;
    my $target = (ref($journals) eq "HASH") ? [$journals,] : $journals;

    for my $j (@$target) {
        if ($j->{formatter} eq "modern") {
            $j->{bodytext} = $self->article_to_bodytext($j->{article},
                                                        $j->{posttype},
                                                        {formatter => "modern"});
            $j->{fulltext} = $self->article_to_fulltext($j->{article},
                                                        $j->{posttype},
                                                        {formatter => "modern"});
        }
        else {
            $j->{introtext} = $self->article_to_introtext($j->{article},
                                                          $j->{posttype});
            $j->{bodytext} = $self->article_to_bodytext($j->{article},
                                                        $j->{posttype});
            $j->{fulltext} = $self->article_to_fulltext($j->{article},
                                                        $j->{posttype});
        }
        $j->{body_text} = $j->{bodytext}; # for compatible
        $j->{full_text} = $j->{fulltext}; # for compatible
    }
    return;
}

sub _generalize {
    my ($self, $journal, $options) = @_;

    # NTO-nized
    $journal->{journal_id} = $journal->{id};
    $journal->{create_time} = $journal->{date};
    $journal->{update_time} = $journal->{last_update};

    my @tags = split /\s+/, $journal->{toptags};
    $journal->{toptags} = \@tags;
    my $primary_topic = {};
    for my $item (qw(tid keyword textname series image width height submittable searchable storypickable usesprite)) {
        $primary_topic->{$item} = $journal->{$item};
    }
    $journal->{primary_topic} = $primary_topic;
    $journal->{topics} = [$primary_topic,];
    $journal->{content_type} = 'journal';

    $journal->{title} = $journal->{description};

    if (!$journal->{formatter}) {
        $journal->{formatter} = "legacy";
    }

    ## parse process is not lightweight, so bypassed here
    #
    # if ($journal->{formatter} eq "modern") {
    #     $journal->{bodytext} = $self->article_to_bodytext($journal->{article},
    #                                                       $journal->{posttype},
    #                                                       {formatter => "modern"});
    #     $journal->{fulltext} = $self->article_to_fulltext($journal->{article},
    #                                                       $journal->{posttype},
    #                                                       {formatter => "modern"});
    # }
    # elsif ($journal->{formatter} eq "legacy") {
    #     $journal->{introtext} = $self->article_to_introtext($journal->{article},
    #                                                       $journal->{posttype});
    #     $journal->{bodytext} = $self->article_to_bodytext($journal->{article},
    #                                                       $journal->{posttype});
    #     $journal->{fulltext} = $self->article_to_fulltext($journal->{article},
    #                                                       $journal->{posttype});
    # }


    $journal->{intro_text} = $journal->{introtext}; # for compatible
    $journal->{discussion_id} = $journal->{discussion};

    # no public flag given, public is 'yes'
    $journal->{public} = 'yes' if !$journal->{public};

    # add tags_string
    if ($journal->{tags}) {
        my $tags = $self->new_instance_of("Tags");
        my @tagnames;
        for my $tag (@{$journal->{tags}}) {
            next if $tags->is_reserved_name($tag->{tagname});
            push @tagnames, $tag->{tagname};
        }
        $journal->{tags_string} = join(" ", @tagnames);
    }

    return $journal;
}


=head2 create(\%params, \%user, \%extra_params, \$opts)

create a journal.

=over 4

=item Parameters

=over 4

=item \%params

parameters

=item \user

author's user object

=item extra_params

extra parameters

=item opts

options

=back

=item Return value

journal id

=back

=cut

use constant COMMENT_STATUS => ('disabled', 'enabled', 'friends_only', 'friends_fof_only',
                                'no_foe', 'no_foe_eof' ,'logged_in');

sub create {
    #my ($self, $params, $user, $extra_params, $opts) = @_;
    my $self = shift;
    return if $self->check_readonly;

    my $params = {@_};
    my $user = $params->{user};
    return if (!$params || !$user);

    # check title exists
    my $title = $params->{description} || $params->{title};
    if (!$title) {
        $self->set_error("no title given", 1);
        return;
    }
    # check title exists
    my $article = $params->{article} || $params->{body_text} || $params->{bodytext};
    if (!$article) {
        $self->set_error("no journal text given", 1);
        return;
    }

    # check promotetype
    # check discussion
    my $comment_status = $params->{commentstatus} || $params->{comment_status};
    if (!$comment_status) {
        $self->set_error("no comment_status given", 1);
        return;
    }
    if (!grep {$comment_status eq $_} COMMENT_STATUS) {
        $self->set_error("invalid comment_status given", 1);
        return;
    }

    # posttype: 1: PLAINTEXT, 2: HTML, 3: EXTRANS, 4: CODE, 77: FULLHTML
    my $post_type = $params->{posttype} || $params->{post_type} || 2;
    my $topic_id = $params->{tid} || $params->{topic_id} || 2068, # tid=2068 is journal.,
    my $promo_type = $params->{promotype} || $params->{promo_type} || 'publish';

    my $dbh = $self->start_transaction;

    my $discussion_id;
    if ($comment_status ne 'disabled') {
        # is discussion enabled, create discussion
        my $discussions = $self->new_instance_of('Discussions');
        my $escaped_nickname = uri_escape($user->{nickname});

        $discussion_id = $discussions->create(title => $title,
                                              url => "http://srad.jp/~$escaped_nickname/journal/",
                                              topic => $topic_id,
                                              kind => 'journal',
                                              commentstatus => $comment_status,
                                              uid => $user->{uid} || $user->{user_id},
                                             );
        if (!$discussion_id) {
            $self->rollback;
            $self->set_error("discussion create failed", $discussions->last_errorno);
            return;
        }
    }

    # create journal
    my @values;
    my $sql = <<"EOSQL";
INSERT INTO journals
    (uid, date,  description, posttype, discussion, tid, promotetype, srcid_32, srcid_24)
  VALUES
    (?,   NOW(), ?,           ?,        ?,          ?,   ?,
       CAST(CONV(?, 16, 10) AS UNSIGNED),
       CAST(CONV(?, 16, 10) AS UNSIGNED))
EOSQL

    push @values, ($user->{uid} || $user->{user_id});
    push @values, $title;
    push @values, $post_type;
    push @values, $discussion_id;
    push @values, $topic_id;
    # promotype: publicize: make_submission, publish: post firehose, post: no firehose
    push @values, $promo_type;
    # srcid_32 and srcid_24 is decimal.
    # warn: Hexadecimal number > 0xffffffff non-portable
    push @values, $user->{srcids}{32};
    push @values, $user->{srcids}{24};

    my $rs = $dbh->do($sql, undef, @values);
    if (!$rs) {
        $self->rollback;
        $self->set_error($dbh->errstr, $dbh->err);
        return;
    }
    my $journal_id = $dbh->last_insert_id(undef, undef, undef, undef);

    # insert into journals_text
    $sql = <<"EOSQL";
INSERT INTO journals_text
    (id, article, introtext)
  VALUES
    (?,  ?,       ?)
EOSQL

    my $introtext = $self->article_to_introtext($article, $post_type, $params);
    my $bodytext = $self->article_to_bodytext($article, $post_type, $params);

    $rs = $dbh->do($sql, undef, $journal_id, $article, $introtext);
    if (!$rs) {
        $self->rollback;
        $self->set_error($dbh->errstr, $dbh->err);
        return;
    }

    # update discussion info
    if ($discussion_id) {
        my $discussions = $self->new_instance_of('Discussions');
        my $escaped_nickname = uri_escape($user->{nickname});
        $discussions->update(id => $discussion_id,
                             url => "http://srad.jp/~$escaped_nickname/journal/$journal_id");
    }

    # insert formatter
    if ($params->{formatter}) {
        my $formatter_table = JOURNAL_FORMATTER_TABLE;
        $sql = <<"EOSQL";
INSERT INTO $formatter_table
    (journal_id, formatter)
  VALUES
    (?,          ?)
EOSQL
        $rs = $dbh->do($sql, undef, $journal_id, $params->{formatter});
        if (!$rs) {
            $self->rollback;
            $self->set_error($dbh->errstr, $dbh->err);
            return;
        }
    }

    # firehose item is not created, so select() cannot use.
    #my $journal = $self->select(id => $journal_id, dbh => $dbh);
    my $journal = {
                   description => $title,
                   id => $journal_id,
                   promotetype => $promo_type,
                   uid => $user->{uid},
                   tid => $params->{tid} || 2068,
                   posttype => $post_type,
                   date => strftime("%F %T", gmtime),
                   discussion => $discussion_id,
                  };
    # TODO: update firehose's timestamp.
    # create firehose item
    $rs = $self->create_firehose_item($journal, $introtext, $bodytext, $user);
    if (!$rs) {
        $self->rollback;
        return;
    }
    my $globj_id = $rs->{globj_id};

    # set tags
    my $tags = $self->new_instance_of("Tags");
    if ($params->{tags_string}) {
        my @tag_names = split(/\s+/, $params->{tags_string});
        for my $tag_name (@tag_names) {
            next if $tags->is_reserved_name($tag_name);
            my $new_id = $tags->set_tag(globj_id => $globj_id,
                                        uid => $user->{uid},
                                        name => $tag_name);
            if (!$new_id) {
                $self->logger->warn("setting tag failed: $tag_name to $globj_id");
            }
        }
    }
    # add "journal" tag
    my $new_id = $tags->set_tag(globj_id => $globj_id,
                                uid => $user->{uid},
                                name => "journal");
    if (!$new_id) {
        $self->logger->warn("setting tag failed: journal to $globj_id");
    }

    # update users_journal:
    #   - set latest user's journal id to users_journal.jid
    #   - set latest user's journal date to users_journal.date
    #   - set number of journals to users_journal.count
    #
    $sql = <<"EOSQL";
UPDATE users_journal
  SET jid = ?,
      count = count + 1,
      date = ?
  WHERE uid = ?
EOSQL
    $rs = $dbh->do($sql, undef, $journal->{id}, $journal->{date}, $journal->{uid});
    if (!$rs) {
        $self->rollback;
        $self->set_error("update users_journal failed", $dbh->err);
        return;
    }

    #my $achievements = getObject('Slash::Achievements');
    #if ($achievements) {
    #    $achievements->setUserAchievement('journal_posted', $user->{uid});
    #}

    # if url given, create url object. But, current implement has no url form.
    #if ($form->{url_id}) {
    #    my $url_id = $form->{url_id};
    #    my $globjid = $slashdb->getGlobjidCreate('journals', $id);
    #    $slashdb->addUrlForGlobj($url_id, $globjid);
    #}

    # transfer tag info
    #getObject("Slash::Tags")->transferTags($fhitem->{globjid}, $journal_globjid);

    # call hooks, in current srad.jp, call Slash::Journal::promoteJournal.
    # slashHook('journal_save_success', { id => $id });

    # if promotetype is "publicize", create submission
    if ($promo_type && $promo_type eq "publicize") {
        my $submissions = $self->new_instance_of("Submissions");
        my $sub_id = $submissions->create(user => $user,
                                          title => $title,
                                          introtext => $article,
                                          tid => $topic_id,
                                         );
        if (!$sub_id) {
            $self->rollback;
            $self->set_error("submission create failed", $submissions->last_errorno);
            return;
        }
    }

    # create messages

    # done
    $self->commit;
    return $journal_id;
}

=head2 update(\%params, \%user, \%extra_params, \$opts)

update a journal.

=over 4

=item Parameters

=over 4

=item \%params

parameters

=item \user

author's user object

=item extra_params

extra parameters

=item opts

options

=back

=item Return value

journal id

=back

=cut

sub update {
    #my ($self, $params, $user, $extra_params, $opts) = @_;
    my $self = shift;
    return if $self->check_readonly;

    my $params = {@_};
    my $user = $params->{user};
    return if (!$params || !$user);

    my $id = $params->{id} || $params->{journal_id};
    return if !$id;

    # get current information
    my $current = $self->select(id => $id);
    return if !$current;

    my $dbh = $self->start_transaction;

    my @sql_clauses;
    my @sql_params;
    my $updated = 0;

    my $article = $params->{article} || $params->{bodytext} || $params->{body_text};
    my $description = $params->{description} || $params->{title};
    my $posttype = $params->{posttype} || $params->{post_type};
    my $discussion_id = $params->{discussion} || $params->{discussion_id} || $current->{discussion};
    my $tid = $params->{tid};
    my $commentstatus = $params->{commentstatus} || $params->{comment_status};
    my $formatter = $params->{formatter};

    # check formatter
    if ($formatter) {
        my $formatter_table = JOURNAL_FORMATTER_TABLE;
        my $sql = "UPDATE $formatter_table SET formatter = ? WHERE journal_id = ?";
        my $rs = $dbh->do($sql, undef, $formatter, $id);
        if (!defined $rs) {
            $self->rollback;
            $self->set_error("update $formatter_table failed", $dbh->err);
            return;
        }
        $updated += $rs;
        # force update journals table's last_update column.
        push @sql_clauses, "last_update = DEFAULT";
    }
    else {
        my $formatter_table = JOURNAL_FORMATTER_TABLE;
        my $sql = "DELETE FROM $formatter_table WHERE journal_id = ?";
        my $rs = $dbh->do($sql, undef, $id);
        if (!defined $rs) {
            $self->rollback;
            $self->set_error("delete from $formatter_table failed", $dbh->err);
            return;
        }
        $updated += 1;
        # force update journals table's last_update column.
        push @sql_clauses, "last_update = DEFAULT";
    }

    # check article body
    if ($article) {
        my $introtext = $self->article_to_introtext($article, $posttype || 2, $params);

        my $sql = "UPDATE journals_text SET article = ?, introtext = ? WHERE id = ?";
        my $rs = $dbh->do($sql, undef, $article, $introtext, $id);
        if (!defined $rs) {
            $self->rollback;
            $self->set_error("update journals_text failed", $dbh->err);
            return;
        }
        $updated += $rs;
        # force update journals table's last_update column.
        push @sql_clauses, "last_update = DEFAULT";
    }

    # check commentstatus
    if ($commentstatus) {
        if ($commentstatus ne 'disabled' && !$discussion_id) {
            # is discussion enabled and not created, create it.
            my $t = $description;
            if (!$t) {
                my $journal = $self->select(id => $id);
                if (!$journal) {
                    $self->rollback;
                    $self->set_error("cannot select target journal", $dbh->err);
                    return;
                }
                $t = $journal->{title};
            }
            my $discussions = $self->new_instance_of('Discussions');
            my $escaped_nickname = uri_escape($user->{nickname});
            $discussion_id = $discussions->create(title => $t,
                                                  url => "http://srad.jp/~$escaped_nickname/journal/$id",
                                                  topic => $tid || 2068, # tid=2068 is journal.
                                                  kind => 'journal',
                                                  commentstatus => $commentstatus,
                                                  uid => $user->{uid} || $user->{user_id},
                                                 );
            if (!$discussion_id) {
                $self->rollback;
                $self->set_error("discussion create failed", $discussions->last_errorno);
                return;
            }
            push @sql_clauses, "discussion = ?";
            push @sql_params, $discussion_id;
        }
    }

    # check given params
    my $tbl = { description => $description,
                posttype => $posttype,
                tid => $tid,
                promotetype => $params->{promotetype} || $params->{promote_type}
              };
    for my $k (qw(description posttype tid promotetype)) {
        if (defined $tbl->{$k}) {
            push @sql_clauses, "$k = ?";
            push @sql_params, $tbl->{$k};
        }
    }

    # update srcids
    push @sql_clauses, "srcid_32 = CAST(CONV(?, 16, 10) AS UNSIGNED)";
    push @sql_clauses, "srcid_24 = CAST(CONV(?, 16, 10) AS UNSIGNED)";
    push @sql_params, $user->{srcids}{32};
    push @sql_params, $user->{srcids}{24};

    # do update table
    my $set_clauses = join(", ", @sql_clauses);
    my $sql = "UPDATE journals SET $set_clauses WHERE id = ?";
    push @sql_params, $id;

    my $rs = $dbh->do($sql, undef, @sql_params);
    if (!defined $rs) {
        $self->rollback;
        $self->set_error("update journals_text failed", $dbh->err);
        return;
    }
    $updated += $rs;

    my $q_sql = <<"EOSQL";
SELECT id, globjid FROM firehose
  WHERE type = 'journal'
    AND srcid = ?
EOSQL
    my $firehose_item = $dbh->selectall_arrayref($q_sql, {Slice => {}}, $id);
    if (!$firehose_item || !@$firehose_item) {
        $self->rollback;
        $self->set_error("cannot get firehose items", $dbh->err);
        return;
    }
    my $firehose_id = $firehose_item->[0]->{id};
    my $globj_id = $firehose_item->[0]->{globjid};
    # update firehose_text
    if ($article || $description) {
        my @sql_clauses;
        my @sql_params;
        if ($article) {
            my $introtext = $self->article_to_introtext($article, $posttype || 2, $params);
            my $bodytext = $self->article_to_bodytext($article, $posttype || 2, $params);
            push @sql_clauses, "introtext = ?";
            push @sql_params, $introtext;
            push @sql_clauses, "bodytext = ?";
            push @sql_params, $introtext . $bodytext; # TODO!
        }
        if ($description) {
            push @sql_clauses, "title = ?";
            push @sql_params, $description;
        }
        my $set_clauses = join(", ", @sql_clauses);
        my $sql = "UPDATE firehose_text SET $set_clauses WHERE id = ?";
        push @sql_params, $firehose_id;
        my $rs = $dbh->do($sql, undef, @sql_params);
        if (!defined $rs) {
            $self->rollback;
            $self->set_error("update journals_text failed", $dbh->err);
            return;
        }
        $updated += $rs;
    }

    # update tags_string
    if (defined $params->{tags_string}) {
        my $tags = $self->new_instance_of("Tags");
        my @tag_names = split(/\s+/, $params->{tags_string});
        my @tag_filtered;
        for my $tag_name (@tag_names) {
            next if $tags->is_reserved_name($tag_name);
            push @tag_filtered, $tag_name;
        }
        my $rs = $tags->set(globj_id => $globj_id,
                            uid => $user->{uid},
                            names => \@tag_filtered);
        if ($rs) {
            $updated++;
        }
    }

    # if no updated item, done.
    if (!$updated) {
        return $id;
    }

    # update firehose
    @sql_clauses = ();
    @sql_params = ();
    if ($discussion_id) {
        push @sql_clauses, "discussion = ?";
        push @sql_params, $discussion_id;
    }
    if ($params->{tid}) {
        push @sql_clauses, "tid = ?";
        push @sql_params, $tid;
    }
    push @sql_clauses, "last_update = NOW()";
    $set_clauses = join(", ", @sql_clauses);
    $sql = "UPDATE firehose SET $set_clauses WHERE id = ?";
    push @sql_params, $firehose_id;
    $rs = $dbh->do($sql, undef, @sql_params);
    if (!$rs) {
        $self->rollback;
        $self->set_error("update firehose failed", $dbh->err);
        return;
    }

    # done
    $self->commit;
    return $id;
}

sub hard_delete {
    my $self = shift;
    my $params = {@_};
    my $journal_id = $params->{journal_id};

    return if !$journal_id;
    my $error;

    my $dbh = $self->start_transaction;

    for my $table (qw(journals journals_text)) {
        my $sql = "DELETE FROM $table WHERE id = ?";
        my $rs = $dbh->do($sql, undef, $journal_id);
        if (!defined $rs || $rs == 0) {
            Mojo::Log->new->warn("DELETE FROM $table failed. id is $journal_id.");
            $error = 1;
        }
    }
    # delete formatter_table
    my $formatter_table = JOURNAL_FORMATTER_TABLE;
    my $rs = $dbh->do("DELETE FROM $formatter_table WHERE journal_id = ?", undef, $journal_id);
    if (!defined $rs) {
        Mojo::Log->new->warn("DELETE FROM $formatter_table failed. id is $journal_id.");
        $error = 1;
    }

    # delete from firehose
    my $firehose = $self->new_instance_of("Firehose");
    $firehose->hard_delete("journal", $journal_id);

    $self->commit;
    return !$error;


}


=pod

HTML形式：（HTML）
利用可能なHTMLタグを使って、投稿の体裁を整えることができます。

テキスト形式(HTML OK!)：（EXTRANS）
HTMLタグを使用できます。HTML形式との違いは、改行のポイントで改行文字として<BR>が
自動的に挿入され、さらに空白文字は多少知的に改行なしのスペースに変わります。

ホントのテキスト形式：（PLAINTEXT）
&、 <、>が除外され、HTMLタグの利用ができないホントのテキスト形式です。

コード：（CODE）
HTMLを使えないテキスト形式ですが、さらに等幅フォントが使われ、適当にインデントもされます。

=cut

sub article_to_introtext {
    my ($self, $article, $posttype, $options) = @_;
    my $html = $self->article_to_fulltext($article, $posttype, $options);
    return if !$html;
    my $intro_text = $self->_extract_introtext($html, $posttype, $options);
    $intro_text =~ s/\s+$//g;

    if ($intro_text) {
        $intro_text = Newslash::Util::TextFormatter::tidy_html($intro_text);
    }
    return $intro_text;
}

sub _extract_introtext {
    my ($self, $html, $posttype, $options) = @_;
    my $min_chars = 50;
    my $max_chars = 500;
    my $intro_text;
    return if !$html;

    if (length $html < $min_chars) {
        return $html;
    }

    my $linebreak = qr{(?:
                           <br>\s*<br> |
                           </?p> |
                           </(?:
                               div | (?:block)?quote | [oud]l
                           )>
                       )}x;
    $intro_text = $1 if $html =~ m/^(.{$min_chars,$max_chars})?$linebreak/s;
    $intro_text =~ s/\s+$//g;
    if (!$intro_text) {
        return Newslash::Util::TextFormatter::contextual_strip($html, $max_chars, $min_chars);
    }
    return $intro_text;
}

sub article_to_bodytext {
    my ($self, $article, $posttype, $options) = @_;
    my $html = $self->article_to_fulltext($article, $posttype, $options);
    my $min_chars = 50;
    if (length $html < $min_chars) {
        return "";
    }
    my $intro_text = $self->_extract_introtext($html, $posttype, $options);
    my $body_text = substr $html, length($intro_text);
    $body_text =~ s/\s+$//g;
    if ($body_text) {
        $body_text = Newslash::Util::TextFormatter::tidy_html($body_text);
    }
    return $body_text;
}

sub article_to_fulltext {
    my ($self, $article, $posttype, $options) = @_;
    $options ||= {};

    my $html;
    my $allowed_tags = $options->{allowed_tags} || $self->options(Editor => "allowed_tags") || DEFAULT_ALLOWED_TAGS;

    my $formatter = $options->{formatter};
    if ($formatter) {
        if ($formatter eq "modern") {
            $html = Newslash::Util::TextFormatter::clean_html($allowed_tags, $article);
            return $html;
        }
    }

    # compatibility mode
    # posttype: 1: PLAINTEXT, 2: HTML, 3: EXTRANS, 4: CODE, 77: FULLHTML
    $html = Newslash::Util::TextFormatter::strip_by_mode($article, $posttype, allowed_tags => $allowed_tags);
    return $html;
}

sub create_firehose_item {
    my ($self, $journal, $introtext, $bodytext, $user) = @_;
    return if $self->check_readonly;
    return if !$journal;

    # create globjs
    my $globjs = $self->new_instance_of('Globjs');
    my $globj_id = $globjs->create(journals => $journal->{id});
    if (!$globj_id) {
        $self->set_error($globjs->last_error, $globjs->last_errorno);
        return;
    }

    # create item
    # TODO: sprite_info column don't have default value...
    my $sql = <<"EOSQL";
INSERT INTO firehose
    (uid, globjid, discussion, type, createtime, popularity, editorpop, public, attention_needed, primaryskid, tid, srcid,
     url_id, email, emaildomain, name, ipid, subnetid, sprite_info)
  VALUES
    (?,  ?,        ?,          ?,    ?,          ?,          ?,         ?,      ?,                ?,           ?,   ?,
     ?,      ?,     ?,           ?,    ?,    ?,        "")
EOSQL
    my $dbh = $self->connect_db;

    my $promo_type = $journal->{promotype} || $journal->{promo_type};

    my $publicize  = $promo_type eq 'publicize';
    my $publish    = $promo_type eq 'publish';
    my $color_lvl  = $publicize ? 5 : $publish ? 6 : 7; # post == 7
    my $editor_lvl = $publicize ? 5 : $publish ? 6 : 8; # post == 8

    my $mods = $self->new_instance_of('Moderations');
    my $popularity = $mods->getEntryPopularityForColorLevel($color_lvl);
    my $editorpop  = $mods->getEntryPopularityForColorLevel($editor_lvl);

    my $primary_skid = 1; # journal's skid is 1 (mainpage)

    my $rs = $dbh->do($sql, undef,
                      $journal->{uid} || $journal->{user_id}, #uid
                      $globj_id, #globjid
                      $journal->{discussion} || 0,
                      "journal", #type
                      $journal->{date} || $journal->{create_time}, #createtime
                      $popularity, #popularity
                      $editorpop, #editorpop
                      "yes", #public
                      "yes", #attention_needed
                      $primary_skid, #primaryskid
                      $journal->{tid} || $journal->{topic_id}, #tid
                      $journal->{id}, #srcid
                      0, #url_id
                      "", #email
                      "", #emaildomain
                      "", #name
                      $user->{ipid} || $user->{ip_id}, #ipid
                      $user->{subnetid} || $user->{subnet_id} #subnetid
                     );
    if (!$rs) {
        $self->set_error("insert into firehose failed", $dbh->err);
        $self->disconnect_db;
        return;
    }
    my $fh_id = $dbh->last_insert_id(undef, undef, undef, undef);

    # insert firehose_text
    $sql = <<"EOSQL";
INSERT INTO firehose_text
    (title, introtext, bodytext, media)
  VALUES
    (?,     ?,         ?,        NULL)
EOSQL
    $rs = $dbh->do($sql, undef,
                   $journal->{description},
                   $introtext,
                   $introtext . $bodytext); # TODO!
    if (!$rs) {
        $self->set_error("insert into firehose_text failed", $dbh->err);
        $self->disconnect_db;
        return;
    }

    # create tag
    my $tags_upbote_tagname = "nod";
    my $tags = $self->new_instance_of("Tags");
    my $tag_id = $tags->set_tag(uid => $journal->{uid} || $journal->{user_id},
                                name => $tags_upbote_tagname,
                                globj_id => $globj_id,
                                private => 1);
    if (!$tag_id) {
        $self->set_error("insert into tags failed", $tags->last_errorno);
        $self->disconnect_db;
        return;
    }

    $self->disconnect_db();
    return { fh_id => $fh_id,
             globj_id => $globj_id };
}

1;
