package Newslash::Model::Metamoderations;
use Newslash::Model::Base -base;

use Data::Dumper;
use List::Util qw(sum any);
use POSIX;

my $m2_params = {
                 m2_mintokes => 0,
                 m2_consensus => 9,

                 count_threshold => 3,
                 agreement_threshold_ratio => 0.66,
                 prevent_params => {
                                    alpha => 0.85,
                                    beta => 0.43,
                                    min => 0.5,
                                    max => 1.0,
                                    min_count => 3,
                                    max_count => 20,
                                    reduction => 1.0,
                                   },
                 enforce_params => {
                                    alpha => 0.85,
                                    beta => 0.086,
                                    min => 0.05,
                                    max => 0.15,
                                    min_count => 3,
                                    max_count => 20,
                                    reduction => 0.1,
                                   },
                 token_unfair_params => {
                                       alpha => 0.56,
                                       beta => 21,
                                       min => 1.0,
                                       max => 20.0,
                                       min_count => 3,
                                       max_count => 20,
                                      },
                 token_fair_params => {
                                       alpha => 0.56,
                                       beta => 4.2,
                                       min => 0.2,
                                       max => 4.0,
                                       min_count => 3,
                                       max_count => 20,
                                      },
                };

sub create {
    my ($self, $user, $moderation_id, $value) = @_;
    return if $self->check_readonly;
    my $moderations = $self->new_instance_of("Moderations");
    my $moderation = $moderations->select(id => $moderation_id);
    if (!$moderation) {
        $self->set_error("invalid moderation id");
        return;
    }

    # user validation
    if (!$moderations->is_metamoderatable($user, $moderation)
        && !$user->{is_admin}
        && !$user->{editor} ) {
        $self->set_error("you cannot metamoderate the moderation");
        return;
    }

    # when the user already metamoderated, return.
    my $m2check = $self->select(moderation_id => $moderation->{id},
                                uid => $user->{uid});
    if (!$m2check || @$m2check != 0) {
        $self->set_error("you have already metamoderated");
        return;
    }

    my $metamods = $self->select(moderation_id => $moderation->{id});
    if (!defined $metamods) {
        $self->set_error("metamoderations retrive error");
        return;
    }
    my $mods = $moderations->select(cid => $moderation->{cid});
    if (!defined $mods) {
        $self->set_error("moderations retrive error");
        return;
    }

    # calc unchanged moderation's fixed scores
    my $unchanged_scores = [];
    for my $mod (@$mods) {
        next if $mod->{id} == $moderation->{id};
        my $mms = $self->select(id => $mod->{id});
        push @$unchanged_scores, $self->_get_fixed_scores($mod, $mms);
    }
    my $base_scores = {
                      points => sum map { $a->{points} } $unchanged_scores,
                      karma => sum map { $a->{karma} } $unchanged_scores,
                      };

    # calc changed moderation's old score
    my $old_score = $self->_get_fixed_scores($moderation, $metamods);
    my $old_total = {
                     points => floor($base_scores->{points} + $old_score->{points}),
                     karma => floor($base_scores->{karma} + $old_score->{karma}),
                    };
    my $old_result = $self->_get_votes_result($moderation, $metamods);
    my $old_moderator_scores = $self->_get_fixed_moderator_scores($moderation, $old_result);

    $self->start_transaction;

    # create metamod
    my $m2id = $self->_create_metamodlog($user, $moderation, $value);
    if (!$m2id) {
        $self->rollback;
        $self->set_error("metamodlog create failed");
        return;
    }

    # calc changed moderation's new score
    my $new_metamod = {
                       val => $value,
                       uid => $user->{uid},
                       mmid => $moderation->{id},
                       active => 1,
                     };
    push @$metamods, $new_metamod;
    my $new_score = $self->_get_fixed_scores($moderation, $metamods);
    my $new_total = {
                     points => floor($base_scores->{points} + $new_score->{points}),
                     karma => floor($base_scores->{karma} + $new_score->{karma}),
                    };
    my $delta = {
                 points => $new_total->{points} - $old_total->{points},
                 karma => $new_total->{karma} - $old_total->{karma},
                };

    # update comment's point
    my $comments = $self->new_instance_of("Comments");
    my $rs = $comments->update(cid => $moderation->{cid},
                               points => { add => $delta->{points} });
    if (!defined $rs) {
        $self->rollback;
        $self->set_error("comments update failed");
        return;
    }

    # update comment author's karma
    my $users = $self->new_instance_of("Users");
    $rs = $users->info->update(uid => $moderation->{cuid},
                               karma => { add => $delta->{karma} });
    if (!defined $rs) {
        $self->rollback;
        $self->set_error("users update failed");
        return;
    }

    # update moderator's info
    my $new_result = $self->_get_votes_result($moderation, $metamods);
    my $new_moderator_scores = $self->_get_fixed_moderator_scores($moderation, $old_result);
    my $u_delta = {};
    for my $k (qw(karma tokens up_fair down_fair up_unfair down_unfair)) {
        $u_delta->{$k} = $old_result->{$k} - $new_result->{$k};
    }
    $users->info->update(uid => $moderation->{uid},
                         tokens => { add => $u_delta->{tokens} },
                         up_fair => { add => $u_delta->{up_fair} },
                         down_fair => { add => $u_delta->{down_fair} },
                         up_unfair => { add => $u_delta->{up_unfair} },
                         down_unfair => { add => $u_delta->{down_unfair} });

    # update metamoderator's info
    my $new_params = {};
    $new_params->{m2_fair} = 1 if $value > 0;
    $new_params->{m2_unfair} = 1 if $value < 0;
    if (($value > 0) && ($moderation->{val} > 0)) {
        $new_params->{m2voted_up_fair} = 1;
    }
    if (($value > 0) && ($moderation->{val} < 0)) {
        $new_params->{m2voted_down_fair} = 1;
    }
    if (($value < 0) && ($moderation->{val} > 0)) {
        $new_params->{m2voted_up_unfair} = 1;
    }
    if (($value < 0) && ($moderation->{val} < 0)) {
        $new_params->{m2voted_down_unfair} = 1;
    }
    $users->info->update(uid => $user->{uid}, %$new_params);

    $self->commit;
    return $m2id;
}

sub _get_fixed_moderator_scores {
    my ($self, $mod, $result) = @_;
    my $rs = {
              karma => 0,
              tokens => 0,
              up_fair => 0,
              down_fair => 0,
              up_unfair => 0,
              down_unfair => 0,
             };

    return $rs if $result->{result} == 0;

    my $max_ratio = 0;
    my $p;
    my $votes = $result->{votes};

    if ($result->{result} == 1) {
        $p = $m2_params->{token_fair_params};
    }
    elsif ($result->{result} == -1) {
        $p = $m2_params->{token_unfair_params};
    }

    if ($p) {
        if ($votes <= $p->{min_count}) {
            $max_ratio = $p->{min};
        }
        elsif ($votes >= $p->{max_count}) {
            $max_ratio = $p->{max};
        }
        else {
            $max_ratio = 
              log(($votes - $p->{min_count}) * $p->{alpha})/log(10) * $p->{beta} + $p->{min};
        }
    }
    $rs->{tokens} = $result->{result} * $max_ratio * $result->{ratio};

    if ($result->{result} == 1) {
        $rs->{karma} = 0.5;
        $rs->{up_fair} = 1 if $mod->{val} > 0;
        $rs->{down_fair} = 1 if $mod->{val} < 0;
    }
    elsif ($result->{result} == -1) {
        $rs->{karma} = -1;
        $rs->{up_unfair} = 1 if $mod->{val} > 0;
        $rs->{down_unfair} = 1 if $mod->{val} < 0;
    }
    return $rs;
}

sub _get_fixed_scores {
    my ($self, $mod, $metamods) = @_;
    return if !defined $mod;
    my $rs = {
              points => 0,
              karma => 0,
             };
    return $rs if !$mod->{active};
    return $rs if !defined $metamods;

    my $affect = $self->_get_votes_result($mod, $metamods);

    if ($affect->{result} > 0) {
        # nod
        $rs->{points} = $mod->{val} * (1.0 + $affect->{enforce});
        $rs->{karma} = $mod->{val} * (1.0 + $affect->{enforce});
    }
    elsif ($affect->{result} < 0) {
        # nix
        $rs->{points} = $mod->{val} * (1.0 - $affect->{prevent});
        $rs->{karma} = $mod->{val} * (1.0 - $affect->{prevent});
    }
    else {
        # no agreement
        $rs->{points} = $mod->{val};
        $rs->{karma} = $mod->{val};
    }
    return $rs;
}

sub _get_votes_result {
    my ($self, $mod, $metamods) = @_;
    my $count = @$metamods;
    my $rs = {
              result => 0,
              ratio => 0,
              enforce => 0,
              prevent => 0,
              votes => $count,
             };
    if ($count < $m2_params->{count_threshold}) {
        return $rs;
    }

    my $nods = grep { $_->{val} == 1 && $_->{active} } @$metamods;
    my $nixes = grep { $_->{val} == -1 && $_->{active} } @$metamods;

    my $fraction = $nods / $count;
    if ($fraction > $m2_params->{agreement_threshold_ratio}) {
        $rs->{result} = 1;
        $rs->{ratio} = $fraction;
    }
    else {
        $fraction = $nixes / $count;
        if ($fraction > $m2_params->{agreement_threshold_ratio}) {
            $rs->{result} = -1;
            $rs->{ratio} = $fraction;
        }
    }

    if ($rs->{result} == 0) {
        # no agreement
        return $rs;
    }
    elsif ($rs->{result} < 0) {
        # agreement is nix
        $rs->{prevent} =
          $self->_calc_ratio($m2_params->{prevent_params}, $count, $fraction);
    }
    elsif ($rs->{result} > 0) {
        # agreement is nod
        $rs->{enforce} =
          $self->_calc_ratio($m2_params->{enfoce_params}, $count, $fraction);
    }
    return $rs;

}

sub _calc_ratio {
    my ($self, $p, $votes, $agreement_ratio) = @_;
    my $max_prevent_ratio;
    if ($votes <= $p->{min_count}) {
        $max_prevent_ratio = $p->{min};
    }
    elsif ($votes >= $p->{max_count}) {
        $max_prevent_ratio = $p->{max};
    }
    else {
        $max_prevent_ratio =
          log(($votes - $p->{min_count}) * $p->{alpha})/log(10) * $p->{beta} + $p->{min};
    }
    return $max_prevent_ratio - $p->{reduction} * (1.0 - $agreement_ratio);
}

sub select {
    my $self = shift;
    my $params = {@_};
    my $rs;

    if ($params->{cid}) {
        $rs = $self->select_by_cid(%$params);
    }

    elsif ($params->{moderation_id}) {
        $params->{mmid} = $params->{moderation_id};
        my $uniques = [qw(id)];
        my $keys = [qw(uid val mmid)];
        $rs = $self->generic_select(table => "metamodlog",
                                   uniques => $uniques,
                                   keys => $keys,
                                   params => $params);
    }
    return $rs;
}

sub select_by_cid {
    my $self = shift;
    my $params = {@_};
    my $cid = $params->{cid};
    return if !$cid;

    my $moderations = $self->new_instance_of('Moderations');
    my $sql = <<"EOSQL";
SELECT metamodlog.*, moderatorlog.cid FROM moderatorlog
  RIGHT JOIN metamodlog ON moderatorlog.id = metamodlog.mmid
  WHERE moderatorlog.cid = ? AND metamodlog.active = 1;
EOSQL
    my $dbh = $self->connect_db;
    my $sth = $dbh->prepare($sql);
    $sth->execute($cid);
    my $rs = $sth->fetchall_arrayref({});
    $self->disconnect_db;
    return $rs;
}

sub _create_metamodlog {
    my ($self, $user, $moderation, $value) = @_;
    return if $self->check_readonly;
    my $dbh = $self->connect_db;
    my $sql = <<"EOSQL";
INSERT INTO metamodlog
    (mmid, uid, val, ts, active)
  VALUES
    (?,    ?,   ?,   NOW(), 1)
EOSQL

    my $rs = $dbh->do($sql, undef, $moderation->{id}, $user->{uid}, $value);
    if (!$rs) {
        $self->disconnect_db;
        return;
    }
    my $m2id = $dbh->last_insert_id(undef, undef, undef, undef);
    $self->disconnect_db;
    return $m2id;
}



# import from ::Metamod
sub getM2Needed {
    my ($self, $user, $comment, $reason, $adjust) = @_;
    $adjust ||= 0;

    my $mod = $self->getModForM2Inherit($user, $comment, $reason);
    my $m2base = $mod ? $mod->{m2needed} : $m2_params->{m2_consensus};
    my $m2needed = $m2base + $adjust;
    $m2needed++ if !($m2needed % 2); # m2needed always odd value

    return $m2needed;
}


# import from ::Metamod
sub getModForM2Inherit {
    my ($self, $user, $comment, $reason, $mod_id) = @_;

    my $moderations = $self->new_instance_of("Moderations");
    my $reasons = $moderations->reasons;
    my @m2able_reasons = sort grep { $reasons->{$_}->{m2able} } (keys %$reasons);
    return if !@m2able_reasons;

    my $m2able = join(", ", @m2able_reasons);
    my @params = ( $comment->{cid},
                   $reason->{id},
                   $user->{uid},
                   $comment->{cuid} );
    my $id_clause = "";
    if ($mod_id) {
        $id_clause = " AND id != ?";
        push @params, $mod_id;
    }

    my $sql = <<"EOSQL";
SELECT * FROM moderatorlog
  WHERE cid = ?
    AND reason = ?
    AND uid != ?
    AND cuid != ?
    AND reason in ($m2able)
    AND active = 1 $id_clause
  ORDER BY id ASC LIMIT 1
EOSQL
    my $dbh = $self->connect_db;
    my $sth = $dbh->prepare($sql);
    $sth->execute(@params);
    my $rs = $sth->fetchrow_hashref;
    $self->disconnect_db;

    return $rs;
}

sub getInheritedM2sForMod {
    my ($self, $user, $comment, $reason, $active, $mod_id) = @_;
    return if !$active;

    my $mod = $self->getModForM2Inherit($user, $comment, $reason, $mod_id);
    my $p_mid = defined $mod ? $mod->{id} : undef;
    return if !$p_mid;

    my $sql = "SELECT * FROM metamodlog WHERE mmid = ?";
    my $dbh = $self->connect_db;
    my $sth = $dbh->prepare($sql);
    $sth->execute($p_mid);
    my $m2s = $sth->fetchall_arrayref({});
    if (!$m2s) {
        return;
    }
    $self->disconnect_db;

    return $m2s;
}

sub applyInheritedM2s {
    my ($self, $mod_id, $m2s) = @_;
    my $users = $self->new_instance_of("Users");
    for my $m2 (@$m2s) {
        my $user = $users->select(uid => $m2->{uid});
        my $m2 = { is_fair => $m2->{val} == 1 ? 1 : 0 };
        $self->createInherited($mod_id, $m2, $user);
    }
}

# createMetaMod with inherited = 1 and multi_max = 0
sub createInherited {
    my ($self, $mod_id, $m2, $user) = @_;
    return if $self->check_readonly;
    my $mods = $self->new_instance_of("Moderations");

    # We need to know whether the M1 IDs are for moderations
    # that were up or down.
    #my $m1_list = join(",", keys %$m2s);
    #my $m2s_vals = $self->sqlSelectAllHashref("id", "id, val",
    #                                          "moderatorlog", "id IN ($m1_list)");
    #for my $m1id (keys %$m2s_vals) {
    #    $m2s->{$m1id}{val} = $m2s_vals->{$m1id}{val};
    #}
    my $mod = $mods->select(id => $mod_id);
    return if !$mod;
    $m2->{val} = $mod->{val};

    # Whatever happens below, as soon as we get here, this user has
    # done their M2 for the day and gets their list of OK mods cleared.
    # The only exceptions are admins who didn't metamod any of their
    # saved mods, and inherited M2s.  In the case of inherited mods
    # the m2_user isn't actively m2ing, their m2s are just being
    # applied to another mod.

    my $voted_up_fair = 0;
    my $voted_down_fair = 0;
    my $voted_up_unfair = 0;
    my $voted_down_unfair = 0;

    my $mod_uid = $mod->{uid};
    my $is_fair = $m2->{is_fair};

    my $rows = 0;
    if ($user->{is_admin}
        || $user->{tokens} >= $m2_params->{m2_mintokens}) {

        my $sql = <<"EOSQL";
UPDATE moderatorlog
  SET m2count = ?,
      m2status = IF(m2count >= m2needed AND MOD(m2count, 2) = 1, 1, 0)
  WHERE id = ? 
    AND m2status = 0
    AND m2count < m2needed
    AND active=1
EOSQL
        my $dbh = $self->connect_db;
        my $rs = $dbh->do($sql, undef, $mod_id);
        $rows += $rs || 0; # if no error, returns 0E0 (true!), we want a numeric answer
        $self->disconnect_db;
    }

    my $users = $self->new_instance_of("Users");
    if ($is_fair  && $m2->{val} > 0) {
        ++$voted_up_fair;
        $users->info->update(uid => $mod_uid, up_fair => {add => 1});
        #$ui_hr->{-up_fair}      = "up_fair+1";
    } elsif ($is_fair  && $m2->{val} < 0) {
        ++$voted_down_fair;
        $users->info->update(uid => $mod_uid, down_fair => {add => 1});
        #$ui_hr->{-down_fair}    = "down_fair+1";
    } elsif (!$is_fair && $m2->{val} > 0) {
        ++$voted_up_unfair;
        $users->info->update(uid => $mod_uid, up_unfair => {add => 1});
        #$ui_hr->{-up_unfair}    = "up_unfair+1";
    } elsif (!$is_fair && $m2->{val} < 0) {
        ++$voted_down_unfair;
        $users->info->update(uid => $mod_uid, down_unfair => {add => 1});
        #$ui_hr->{-down_unfair}  = "down_unfair+1";
    }

    my $active;
    if ($rows) {
        # If a row was successfully updated, insert a row
        # into metamodlog.
        $active = 1;
    } else {
        # If a row was not successfully updated, probably the
        # moderation in question was assigned to more than
        # $consensus users, and the other users pushed it up to
        # the $consensus limit already.  Or this user has
        # gotten bad M2 and has negative tokens.  Or the mod is
        # no longer active.
        $active = 0;
    }
    my $sql = <<"EOSQL";
INSERT INTO metamodlog
    (mmid, uid, val, ts,    active)
  VALUES
    (?,    ?,   ?,   NOW(), ?)
EOSQL
    my $dbh = $self->connect_db;
    my $rs = $dbh->do($sql, undef,
                      $mod_id, $user->{uid}, ($is_fair ? '+1' : '-1'), $active);
    if (!$rs) {
        return;
    }
    $self->disconnect_db;

    $users->info->update(uid => $mod_uid,
                         m2voted_up_fair => {add => $voted_up_fair},
                         m2voted_down_fair => {add => $voted_down_fair},
                         m2voted_up_unfair => {add => $voted_up_unfair},
                         m2voted_down_unfair => {add => $voted_down_unfair}
                        );
    return 1;
}

sub adjustForNewMod {
    my ($self, $user, $comment, $reason, $active, $mod_id) = @_;
    my $i_m2s = $self->getInheritedM2sForMod($user, $comment, $reason, $active, $mod_id);
    $self->applyInheritedM2s($mod_id, $i_m2s);
}
