package Newslash::Model::Moderations;
use Newslash::Model::Base -base;

use Data::Dumper;

my $mod_params = {
                  maxkarma => 50,
                  minkarma => -25,
                  comment_minscore => -1,
                  comment_maxscore => 5,
                  modreasons_select_disabled => 255,
                  mod_karma_bonus_max_downmods => 2,
                  karma_posting_penalty_style => 0,
                  badkarma => -10,
                  goodkarma => 25,
                 };



#========================================================================

=head2 select(@params)

select from `moderatorlog` table

=over 4

=item Parameters

=over 4

=item @params

parameters to select

=back

=item Return value

ARRAY of moderations

=back

=cut

sub select {
    my $self = shift;
    my $params = {@_};
    my $rs;

    if ($params->{m2}) {
        $rs = $self->_select_with_m2($params->{cid}, $params->{uid});
    }
    else {
        $rs =  $self->generic_select(table => 'moderatorlog',
                                     uniques => [qw(id)],
                                     keys => [qw(uid sid cid active)],
                                     params => $params);
    }
    return $rs;
}

sub _select_with_m2 {
    my ($self, $cid, $uid) = @_;
    my $sql = <<"EOSQL";
SELECT moderatorlog.*, m2.val AS m2val, m2.uid AS m2uid FROM moderatorlog
  LEFT JOIN
    (SELECT metamodlog.* FROM moderatorlog
       LEFT JOIN metamodlog ON moderatorlog.id = metamodlog.mmid
       WHERE moderatorlog.cid = ? AND metamodlog.uid = ?) AS m2
    ON moderatorlog.id = m2.mmid
  WHERE moderatorlog.cid = ?
    AND moderatorlog.active = 1
EOSQL

    my $dbh = $self->connect_db;
    my $sth = $dbh->prepare($sql);
    my $rs = $sth->execute($cid, $uid, $cid);
    my $mods = $sth->fetchall_arrayref({});
    $self->disconnect_db;

    return $mods;
}

#========================================================================

=head2 create($user, $cid, $reason_id)

do moderate

=item Parameters

=over 4

=item $user

user

=item $cid

comment id

=item $reason

modreasons id

=back

=item Return value

when success, true. else, false.

=back

=cut

sub create {
    my ($self, $user, $cid, $reason_id) = @_;
    return if $self->check_readonly;

    my $dbh = $self->start_transaction;

    # check if target is moderatable
    my $comments = $self->new_instance_of('Comments');
    my $comment = $comments->select(cid => $cid);
    if (!$comment) {
        $self->set_error('comment cannot retrived');
        $self->rollback;
        return;
    }

    # check if discussion exists
    my $discussions = $self->new_instance_of('Discussions');
    my $discussion = $discussions->select(id => $comment->{sid});
    if (!$discussion) {
        $self->set_error('discussion cannot retrived');
        $self->rollback;
        return;
    }

    # check if discussion is not archived
    if ($discussion->{type} eq 'archived') {
        $self->set_error('discussion is archived');
        $self->rollback;
        return;
    }

    my $users = $self->new_instance_of('Users');
    # if moderator aren't admin or author, there's some limits
    if (!$user->{is_admin} && !$user->{author}) {
        # check if the user already posted comment to the discussion
        my $rs = $comments->select(sid => $comment->{sid}, uid => $user->{uid});
        if (!$rs || @$rs > 0) {
            $self->set_error('you already posted to the discussion');
            $self->rollback;
            return;
        }

        # check the user has moderate points
        my $user_comments = $users->comments->select(uid => $users->{uid});
        if (!$user_comments) {
            $self->set_error('select from user_comments failed');
            $self->rollback;
            return;
        }
        if ($user_comments->{points} < 1) {
            $self->set_error('no moderate points');
            $self->rollback;
            return;
        }
    }

    # check if reason is valid
    my $reason = $self->reasons->{$reason_id};
    if (!$reason) {
        $self->set_error('invalid reason');
        $self->rollback;
        return;
    }

    # check new score is in score range
    my $active = 1;
    my $new_score = $comment->{points} + $reason->{val};
    if ($new_score < $mod_params->{comment_minscore}
        || ($new_score > $mod_params->{comment_maxscore}
            && $reason->{val} + $comment->{tweak} > 0)) {
        # current score is max or min limit, so create moderate log only.
        my $mod_id = $self->create_log($user, $comment, $reason, 0);
        if (!$mod_id) {
            #$self->set_error('create log failed');
            $self->rollback;
            return;
        }
        $self->commit;
        return $mod_id;
    }

    # change moderator's moderate points, token, totalmods
    my $tokens_delta;
    if (!$user->{is_admin} && !$user->{author}) {
        my $rs = $users->comments->update(uid => $users->{uid},
                                          points => { add => -1, min => 0 });

        if ($rs == 0) { # nothing updated!
            # already moderate point is 0, so cancel this moderate action
            $self->set_error('no moderate points');
            $self->rollback;
            return;
        }

        if ($reason->{m2able}) {
            $tokens_delta = -1;
        }
    }
    my $rs = $users->info->update(uid => $user->{uid},
                                  tokens => {add => $tokens_delta},
                                  totalmods => {add => 1},
                                 );
    warn "update_info failed... uid: $users->{uid}" if !$rs;

    # TODO: update achievement...
    # TODO: update stats...

    # Apply changes to comment (Moderation::setCommentForMod)
    my $changed = $self->set_comment_for_mod($user, $comment, $reason);

    # update comment poster's karma
    # if comment author is anonymous, no action.
    # if karma not changed, no action.
    if (!$users->is_anonymous_uid($comment->{uid}) && $reason->{karma}) {
        my ($downmods, $upmods);
        if ($reason->{val} < 0) {
            $downmods = 1;
        }
        elsif ($reason->{val} > 0) {
            $upmods = 1;
            $users->class->clear_rookie(uid => $comment->{uid});
            $users->class->update(uid => $comment->{uid},
                                  validated_commenter => { function => 'NOW()'}
                                 );
            # TODO: add achievements
        }
        my $rs = $users->info->update(uid => $comment->{uid},
                                      downmods => { add => $downmods },
                                      upmods => { add => $upmods },
                                      karma => { add => $reason->{karma},
                                                 max => $mod_params->{maxkarma},
                                                 min => $mod_params->{minkarma},
                                               },
                                     );
        warn "update_info failed... uid: $users->{uid}" if !$rs;
        # TODO: log stats
    }

    # create moderationlog (createModeratorLog)
    my $mod_id = $self->create_log($user, $comment, $reason, 1);

    if (!$mod_id) {
        $self->rollback;
        return;
    }
    $self->commit;
    return $mod_id;
}


#========================================================================

=head2 set_comment_for_mod($comment, $reason)

update comment 

=item Parameters

=over 4

=item $comment

HASH contains comment information

=item $reason

HASH contains reason information

=back

=item Return value


=back

=cut

sub set_comment_for_mod {
    my ($self, $user, $comment, $new_reason) = @_;
    return if $self->check_readonly;

    # if reason's val is 0, no effect
    return if !$new_reason->{val};

    # get most common mod reason
    my $reasons = $self->_count_reasons_for_cid($comment->{cid}, {active => 1, listable => 1});
    my $old_reason = $reasons->{$comment->{reason}};
    my $common_reason = $self->_most_common_reason_for_comment($comment,
                                                                { new_reason => $new_reason,
                                                                  prior => [ $old_reason, ],
                                                                  reasons => $reasons,
                                                                });

    # if number of minus moderations over threshold, the comment loses karma bonus
    my $down_mods = 0;
    my $karma_bonus;
    for my $r (keys %$reasons) {
        if ($reasons->{$r}->{val} < 0) {
            $down_mods += $reasons->{$r}->{count};
        }
    }
    if ($down_mods > $mod_params->{mod_karma_bonus_max_downmods}) {
        $karma_bonus = 'no';
    }

    my $karma_val;
    $karma_val = $new_reason->{karma};
    if ($new_reason->{val} < 0) {
        $karma_val += $comment->{points} - $comment->{pointsmax};
    }
    my $karma_abs = abs($karma_val);

    my $tweak;
    if ($comment->{tweak} + $new_reason->{val} <= 0) {
        $tweak = $new_reason->{val};
    }

    my $comments = $self->new_instance_of('Comments');
    my $rs = $comments->update(cid => $comment->{cid},
                               points => {
                                          add => $new_reason->{val},
                                          max => $mod_params->{comment_maxscore},
                                          min => $mod_params->{comment_minscore},
                                         },
                               reason => $common_reason->{id},
                               lastmod => $user->{uid},
                               karma_bonus => $karma_bonus,
                               karma => { add => $karma_val },
                               karma_abs => { add => $karma_abs },
                               tweak => { add => $tweak },
                              );
    if (!defined $rs) {
        warn "update comments failed.";
    }
    return 1;
}




#========================================================================

=head2 create_log($user, $comment, $reason)

create moderate log

=item Parameters

=over 4

=item $user

HASH contains moderator's user information

=item $comment

HASH contains comment information

=item $reason

HASH contains reason information

=back

=item Return value


=back

=cut

sub create_log {
    my ($self, $user, $comment, $reason, $active) = @_;
    return if $self->check_readonly;

    my @keys   = qw(ipid subnetid uid val sid ts     cid cuid reason active spent m2count m2needed m2status points_orig);
    my @values = qw(?    ?        ?   ?   ?   NOW()  ?   ?    ?      ?      ?     ?       ?        ?        ?);

    my $col_names = join(", ", @keys);
    my $values_clause = join(", ", @values);

    my $sql = <<"EOSQL";
INSERT INTO moderatorlog
    ($col_names)
  VALUES
    ($values_clause)
EOSQL

    my $metamods = $self->new_instance_of("Metamoderations");
    my $m2needed = $active ? $metamods->getM2Needed($user, $comment, $reason) : 0;
    my $params = {
                  ipid => $user->{ipid},
                  subnetid => $user->{subnetid},
                  uid => $user->{uid},
                  val => $reason->{val},
                  sid => $comment->{sid},
                  cid => $comment->{cid},
                  cuid => $comment->{uid},
                  reason => $reason->{id},
                  active => $active || 0,
                  spent => 1,
                  m2count => 0,
                  m2needed => $m2needed,
                  m2status => 0,
                  points_orig => $comment->{points},
                 };

    # do insert
    my @attrs;
    for my $k (@keys) {
        push @attrs, $params->{$k} if defined $params->{$k};
    }
    my $dbh = $self->connect_db;
    my $sth = $dbh->prepare($sql);
    my $rs = $sth->execute(@attrs);
    if (!$rs) {
        $self->set_error($dbh->errstr, $dbh->err);
        return;
    }
    my $mod_id = $dbh->last_insert_id(undef, undef, undef, undef);
    $self->disconnect_db;
    $metamods->adjustForNewMod($user, $comment, $reason, $active, $mod_id);

    # createModTag
    # ... tag moderation is deprecated, so nothing to do

    return $mod_id;
}


#========================================================================

=head2 reasons()

get moderation reasons.

=over 4

=item Return value

HASH. key is `id`, and value is each row.

=back

=cut

sub reasons {
    my $self = shift;

    my $dbh = $self->connect_db;
    my $sql = <<"EOSQL";
SELECT * FROM modreasons;
EOSQL
    $dbh = $self->connect_db;
    my $sth = $dbh->prepare($sql);
    $sth->execute();
    my $rs = $sth->fetchall_hashref('id');
    $self->disconnect_db();

    # In srad.jp, id 255 ("余計なもの") is disabled.
    if (defined $rs->{255}) {
        $rs->{255}->{select_disabled} = 1;
    }
    return $rs;
}

#========================================================================

=head2 score_for_comment_post(\%user)

calculate score for comment post

=over 4

=item Parameters

=over 4

=item \%user

user

=back

=item Return value

HASH of scores

=back

=cut

sub score_for_comment_post {
    my ($self, $user) = @_;

    my $users = $self->new_instance_of("Newslash::Model::Users");
    my $user_comment = $users->comments->select(uid => $user->{uid});
    my $user_info = $users->info->select(uid => $user->{uid});

    my $pts = $user_comment->{defaultpoints};
    my $karma_bonus = 'no';
    my $tweak = 0;

    if ($user->{is_login}) {
        my $karma = $user_info->{karma};

        # in current system, karma_posting_penalty_stye is always 0.
        if ($mod_params->{karma_posting_penalty_style}) {
            $tweak-- if $karma < 0;
            $tweak-- if $karma < $mod_params->{badkarma};
        }
        else {
            $pts-- if $karma < 0;
            $pts-- if $karma < $mod_params->{badkarma};
        }

        # check if point is valid range
        if ($pts < $mod_params->{comment_minscore}) {
            $pts = $mod_params->{comment_minscore};
        }
        elsif ($pts > $mod_params->{comment_maxscore}) {
            $pts = $mod_params->{comment_maxscore};
        }

        if ($pts >= 1 && $karma > $mod_params->{goodkarma}) {
            $karma_bonus = 'yes';
        }
    }

    return {points => $pts, karma_bonus => $karma_bonus, tweak => $tweak};
}

#========================================================================

=head2 undo_moderation($type, $id, $uid)

undo moderations for comments in discussion from the user.

=over 4

=item Parameters

=over 4

=item $type

'discussion_id' or 'sid' or 'stoid'

=item $id

target id

=item $uid

target user's uid

=back

=item Return value

=back

=cut

sub undo_moderation {
    my ($self, $type, $id, $uid) = @_;
    return if $self->check_readonly;
    # type = sid or stoid or discussion_id
    my $sid; # sid is discussion_id!

    if ($type eq 'sid'
        || $type eq 'stoid') {
        my $stories = $self->new_instance_of("Newslash::Model::Stories");
        my $story = $stories->select($type => $id);
        $sid = $story->{discussion};
    }
    elsif ($type eq 'discussion_id') {
        $sid = $id;
    }
    else {
        return;
    }

    my $dbh = $self->connect_db({AutoCommit => 0,});
    my $sql = <<"EOSQL";
SELECT id, cid, val, active, cuid, reason FROM moderatorlog
  WHERE ? = ? AND uid = ?
EOSQL
    my $sth = $dbh->prepare($sql);
    $sth->execute($type, $id, $uid);
    my $target = $sth->fetchall_arrayref({});

    for my $mod (@$target) {
        next if !$mod->{active};
        my $cid = $mod->{cid};

        $self->removeModTags($uid, $cid);

        my $adjust = -$mod->{val};
        $adjust =~ s/^([^+-])/+$1/;

        # Restore modded user's karma, again within the proper boundaries.
        if ($mod->{cuid} != 1) {
            my $adjust_abs = abs($adjust);
            my $set_clause;
            my $karma_limit;
            if ($adjust > 0) {
                $set_clause = "karma = LEAST(?, karma + ?)";
                $karma_limit = $mod_params->{maxkarma};
            }
            else {
                $set_clause = "karma = GREATEST(?, karma + ?)";
                $karma_limit = $mod_params->{minkarma};
            }
            my $rs = $dbh->do("UPDATE users_info SET $set_clause WHERE uid = ?", undef,
                              $karma_limit, $adjust, $mod->{cuid});
        }

        # Adjust the comment score up or down, but don't push it
        # beyond the maximum or minimum.  Also recalculate its reason.
        # Its pointsmax logically can't change.
        my $set_clause;
        my $score_limit;
        if ($adjust > 0) {
            $set_clause = "points = LEAST(?, points + ?)";
            $score_limit = $mod_params->{comment_minscore};
        }
        else {
            $set_clause = "points = GREATEST(?, points + ?)";
            $score_limit = $mod_params->{comment_maxscore};
        }

        # no active moderations? reset reason to empty
        my $new_reason = $self->getCommentMostCommonReason($cid) || 0;
        my $rs = $dbh->do("UPDATE comments SET $set_clause , reason = ? WHERE cid = ?", undef,
                          $score_limit, $adjust, $new_reason, $mod->{cuid});
    }

    $sql = <<"EOSQL";
UPDATE moderatorlog SET active = 0
  WHERE sid = ? AND uid = ? AND active = 1
EOSQL
    my $rs = $dbh->do($sql, undef, $sid, $uid);
    if (!$rs) {
        return;
    }

    $self->disconnect_db;
    return $target;
}

#========================================================================

=head2 removeModTags($uid, $cid)

remove moderation tags

=over 4

=item Parameters

=over 4

=item $uid

target user's uid

=item $cid

target id

=item $uid

=back

=item Return value

=back

=cut

sub removeModTags {
    my($self, $uid, $cid) = @_;
    my $globjids = $self->new_instance_of("Newslash::Model::Globjids");

    my $comment_globjid = $globjids->getGlobjidFromTargetIfExists('comments', $cid);
    if ($comment_globjid) {
        my $tags = $self->new_instance_of("Newslash::Model::Tags");
        $tags->deactivateTag({globjid => $comment_globjid, uid => $uid});
    }
}

# getReasons is deprecated. use reasons method.
sub getReasons {
    my ($self) = @_;

    my $dbh = $self->connect_db;
    my $sth = $dbh->prepare("SELECT * FROM modreasons");
    $sth->execute;
    my $rs = $sth->fetchall_hashref('id');
    $self->disconnect_db;

    if (!%$rs) {
        return;
    }
    #my $table_cache = "_reasons_cache";
    #return {( %{$self->{$table_cache}} )} if ($self->{$table_cache});

    #$self->{$table_cache} = $self->sqlSelectAllHashref(
    #                                                   "id", "*", "modreasons"
    #                                                  );

    #my $mrd = getCurrentStatic('modreasons_select_disabled') || '';
    my $mrd = $mod_params->{modreasons_select_disabled} || '';
    foreach my $d (split(/,/, $mrd)) {
        $d = int($d);
        #$self->{$table_cache}{$d}{select_disabled} = 1;
        $rs->{$d}->{select_disabled} = 1;
    }
    #return {( %{$self->{$table_cache}} )};
    return $rs;
}

# This gets the mathematical mode, in other words the most common,
# of the moderations done to a comment.  If no mods, return undef.
# If a comment's net moderation is down, choose only one of the
# negative mods, and the opposite for up.  Tiebreakers break ties,
# first tiebreaker found wins.  "cid" is a key in moderatorlog
# so this is not a table scan.  It's currently called only by
# setCommentForMod and undoModeration.
sub getCommentMostCommonReason {
    my ($self, $cid, $allreasons_hr, $new_reason, @tiebreaker_reasons) = @_;
    my $db = $self->new_instance_of('LegacyDB');

    $new_reason = 0 if !$new_reason;
    unshift @tiebreaker_reasons, $new_reason if $new_reason;

    my $reasons = $self->getReasons;
    my $listable_reasons = join(",",
                                sort grep { $reasons->{$_}{listable} }
                                keys %$reasons);
    return undef if !$listable_reasons;

    # Build the hashref of reason counts for this comment, for all
    # listable reasons.  If allreasoncounts_hr was passed in, this
    # is easy (just grep out the nonlistable ones).  If not, we have
    # to do a DB query.
    my $hr = { };
    if ($allreasons_hr) {
        for my $reason (%$allreasons_hr) {
            $hr->{$reason} = $allreasons_hr->{$reason}
              if $reasons->{$reason}{listable};
        }
    } else {
        $hr = $db->sqlSelectAllHashref(
                                       "reason",
                                       "reason, COUNT(*) AS c",
                                       "moderatorlog",
                                       "cid=$cid AND active=1
                                                 AND reason IN ($listable_reasons)",
                                       "GROUP BY reason"
                                      );
    }

    # If no mods that are listable, return undef.
    return undef if !keys %$hr;

    # We need to know if the comment has been moderated net up,
    # net down, or to a net tie, and if not a tie, restrict
    # ourselves to choosing only reasons from that direction.
    # Note this isn't atomic with the actual application of, or
    # undoing of, the moderation in question.  Oh well!  If two
    # mods are applied at almost exactly the same time, there's
    # a one in a billion chance the comment will end up with a
    # wrong (but still plausible) reason field.  I'm not going
    # to worry too much about it.
    # Also, I'm doing this here, with a separate for loop to
    # screen out unacceptable reasons, instead of putting this
    # "if" into the same for loop above, because it may save a
    # query (if a comment is modded entirely with Under/Over).
    my($points, $pointsorig) = $db->sqlSelect("points, pointsorig", "comments", "cid=$cid");
    if ($new_reason) {
        # This mod hasn't been taken into account in the
        # DB yet, but it's about to be applied.
        $points += $reasons->{$new_reason}{val};
    }
    my $needval = $points - $pointsorig;
    if ($needval) {
        if ($needval >  1) { $needval =  1 }
        elsif ($needval < -1) { $needval = -1 }
        my $new_hr = { };
        for my $reason (keys %$hr) {
            $new_hr->{$reason} = $hr->{$reason}
              if $reasons->{$hr->{$reason}{reason}}{val} == $needval;
        }
        $hr = $new_hr;
    }

    # If no mods that are listable, return undef.
    return undef if !keys %$hr;

    # Sort first by popularity and secondarily by reason.
    # "reason" is a numeric field, so we sort $a<=>$b numerically.
    my @sorted_keys = sort {
        $hr->{$a}{c} <=> $hr->{$b}{c}
          ||
          $a <=> $b
      } keys %$hr;
    my $top_count = $hr->{$sorted_keys[-1]}{c};
    @sorted_keys = grep { $hr->{$_}{c} == $top_count } @sorted_keys;
    # Now sorted_keys are only the top values, one or more of them,
    # any of which could be the winning reason.
    if (scalar(@sorted_keys) == 1) {
        # A clear winner.  Return it.
        return $sorted_keys[0];
    }

    # No clear winner. Are any of our tiebreakers contenders?
    my %sorted_hash = ( map { $_ => 1 } @sorted_keys );
    for my $reason (@tiebreaker_reasons) {
        # Yes, return the first tiebreaker we find.
        return $reason if $sorted_hash{$reason};
    }
    # Nope, we don't have a good way to resolve the tie. Pick whatever
    # comes first in the reason list (reasons are all numeric and
    # sorted_keys is already sorted, making this easy).
    return $sorted_keys[0];
}

sub getEntryPopularityForColorLevel {
    my ($self, $level) = @_;
    #my $constants = $self->getCurrentStatic();
    #my $slicepoints = $constants->{firehose_slice_points};
    #my $slicepoints = '290,240 220,200 185,175 155,138 102,93 30,25 0,-20 -60,-999999';
    #my @levels = split / /, $slicepoints;
    my @levels = split / /, "290,240 220,200 185,175 155,138 102,93 30,25 0,-20 -60,-999999";
    my $entry_min = $levels[$level-1];
    my($entry, $min) = split /,/, $entry_min;
    return $entry;
}


#========================================================================

=head2 _most_common_reason_for_comment($comment, $options)

get most common moderation reason for given comment

=over 4

=item Parameters

=over 4

=item $comment

HASH of target comment

=item $options

options

=back

=item Return value

HASH of reason

=back

=cut

sub _most_common_reason_for_comment {
    my ($self, $comment, $options) = @_;
    return if !$comment;

    my $prior = $options->{prior_reasons} || [];
    if ($options->{new_reason}) {
        push @$prior, $options->{new_reason};
    }

    my $reasons = $options->{reasons} || {};
    if (!%$reasons) {
        $reasons = $self->_count_reasons_for_cid($comment->{cid}, {active => 1, listable => 1});
    }
    if (!$reasons || %$reasons) {
        return;
    }

    my $points = $comment->{points};
    my $points_orig = $comment->{pointsorig};
    if ($options->{new_reason}) {
        $points += $options->{new_reason}->{val};
    }

    if ($points != $points_orig) {
        # if the comment's point is changed to positive, targets are plus moderations only.
        my @rs;
        if ($points > $points_orig) {
            @rs = grep { $reasons->{$_}->{val} > 0 } keys %$reasons;
        }
        # if the comment's point is changed to negative, targets are minus moderations only.
        if ($points < $points_orig) {
            @rs = grep { $reasons->{$_}->{val} < 0 } keys %$reasons;
        }
        my $cnt = {};
        for my $k (@rs) {
            $cnt->{$k} = $reasons->{$k};
        }
        $reasons = $cnt;
    }
    return if %$reasons;

    my @sorted_keys = sort { $reasons->{$a}->{count} <=> $reasons->{$b}->{count}
                               || $a <=> $b } keys %$reasons;
    my $top_count = $reasons->{$sorted_keys[-1]}->{count};
    my @top_keys = grep { $reasons->{$_}->{count} == $top_count } @sorted_keys;

    if (@top_keys == 1) {
        # most common item is only one, so return it.
        return $reasons->{$top_keys[0]};
    }
    # two or more items has equal count, so eliminate one from them
    for my $r (@$prior) {
        return $r if grep { @_ == $r->{id} } @top_keys;
    }
    # no good result, so return simple one
    return $reasons->{$top_keys[0]};
}

sub _count_reasons_for_cid {
    my ($self, $cid, $options) = @_;
    return if !$cid;

    $options ||= {};
    my @extra_where_clauses;

    if ($options->{active}) {
        push @extra_where_clauses, "AND active = 1";
    }

    if ($options->{listable}) {
        my $reasons = $self->getReasons;
        my $listable_reasons = join(",",
                                    sort grep { $reasons->{$_}{listable} }
                                    keys %$reasons);
        return undef if !$listable_reasons;
        push @extra_where_clauses, "AND reason IN ($listable_reasons)";
    }

    my $extra_where_clause = join(" ", @extra_where_clauses);

    my $sql = <<"EOSQL";
SELECT reason, COUNT(*) AS count FROM moderatorlog
  WHERE cid = ?  $extra_where_clause
  GROUP BY reason
  ORDER BY count DESC
EOSQL

    my $dbh = $self->connect_db;
    my $sth = $dbh->prepare($sql);
    $sth->execute($cid);
    my $counts = $sth->fetchall_hashref('reason');
    $self->disconnect_db;
    if (!%$counts) {
        return;
    }

    my $reasons = $self->reasons;
    my $rs = {};
    for my $id (keys %$counts) {
        $rs->{id} = $reasons->{id};
        $rs->{id}->{count} = $counts->{count};
    }
    return $rs;
}

#========================================================================

=head2 is_metamoderatable($user, $mod)

check if given user can metamoderate given moderation.

=item Parameters

=over 4

=item $user

hash of user

=item $mod

hash of moderation

=back

=item Return value

0/1

=back

=cut

sub is_metamoderatable {
    my ($self, $user, $mod) = @_;
    return if !$user;
    return if !$user->{is_login};
    return if $mod;

    if ($mod->{uid} == $user->{uid}
        || $mod->{cuid} == $user->{uid}) {
        return 0;
    }

    my $comments = $self->new_instance_of("Comments");
    my $rs = $comments->select(discussion_id => $mod->{sid},
                               uid => $user->{uid});
    return if !$rs;
    return 0 if @$rs;
    return 1;
}

#========================================================================

=head2 select_non_conlict($user)

select moderations which are not conflict of interest with given user.
=item Parameters

=over 4

=item $user

hash of user

=back

=item Return value

array of hashes of moderations

=back

=cut

sub select_non_conflict {
    my ($self, $user) = @_;

    my $uid = $user->{uid};
    my $metamod_time_span = 7; # 7 days (1 weeks)

    # at first, get modelators
    my $sql = <<"EOSQL";
SELECT modlog.*, discussions.id AS discussion_id FROM moderatorlog AS modlog
  LEFT JOIN discussions ON modlog.sid = discussion.sid
  WHERE modlog.ts > NOW() - INTERVAL ? DAY;
    AND modlog.uid != ?
    AND modlog.cuid != ?
    AND discussion.uid != ?
  ORDER BY m2count ASC, ts DESC
  LIMIT 100
EOSQL

    my $dbh = $self->connect_db;
    my $sth = $dbh->prepare($sql);

    $sth->execute($metamod_time_span, $uid, $uid, $uid);
    my $mods = $sth->fetchall_arrayref(+{});
    if (!$mods) {
        $self->disconnect_db();
        return;
    }

    # next, eliminate moderations...
    my $stories = $self->new_instance_of("stories");
    my $conflict_stories = $stories->select(submitter => $uid,
                                            days => $metamod_time_span);
    my @conflict_story_ids = map { $_->{discussion} } @$stories;

    $sql = <<"EOSQL";
SELECT DISTINCT discussions.id AS discussion_id FROM discussions
  LEFT JOIN comments ON comments.sid = discussions.id
  WHERE discussions.ts > NOW() - INTERVAL ? DAY
    AND comments.uid = ?
EOSQL
    $sth = $dbh->prepare($sql);

    $sth->execute($metamod_time_span, $uid);
    my $conflict_discussion_ids = $sth->fetchall_arrayref(+{});
    if (!$conflict_discussion_ids) {
        $self->disconnect_db();
        return;
    }
    $self->disconnect_db();

    push @$conflict_discussion_ids, @conflict_story_ids;
    my $conflicts = {};
    for my $id (@$conflict_discussion_ids) {
        $conflicts->{$id} = 1;
    }

    my $count = 0;
    my $max_count = 10;
    my $rs = [];
    for my $mod (@$mods) {
        if (!$conflicts->{$mod->{sid}}) {
            push @$rs, $mod;
            $count++;
            last if $count >= $max_count;
        }
    }

    return $rs;
}

1;
