package Newslash::Model::Relations;
use Newslash::Model::Base -base;
use Data::Dumper;
use List::Util qw(any);
use Storable;

use constant {
      REL_FRIEND => "friends",
      REL_FREAK => "freaks",
      REL_FAN => "fans",
      REL_FOE => "foes",
      REL_FOF => "fofs",
      REL_EOF => "eofs",
  };

sub on_start_up {
    my $self = shift;
    return 1;
}

sub create {
    my ($self, $uid, $target_uid, $type) = @_;
    return if $self->check_readonly;

    if (!$uid || !$target_uid || !$type) {
        $self->set_error("no uid or target_uid or type given");
        return;
    }
    if (($type ne 'friend') && ($type ne 'foe')) {
        $self->set_error("invalid type");
        return;
    }

    # first, create forward direction relationship
    my $dbh = $self->start_transaction;

    my $sql = <<"EOSQL";
INSERT INTO people
    (uid, person, type)
  VALUES
    (?,   ?,      ?)
  ON DUPLICATE KEY UPDATE type = ?
EOSQL

    my $rs = $dbh->do($sql, undef, $uid, $target_uid, $type, $type);
    if (!$rs) {
        $self->rollback;
        $self->set_error("insert friend/foe failed");
        return $rs;
    }

    # next, create reverse direction relationship
    my $perceive = ($type eq "friend") ? "fan" : "freak";
    $sql = <<"EOSQL";
INSERT INTO people
    (uid, person, perceive)
  VALUES
    (?,   ?,      ?)
  ON DUPLICATE KEY UPDATE perceive = ?
EOSQL

    $rs = $dbh->do($sql, undef, $target_uid, $uid, $perceive, $perceive);
    if (!$rs) {
        $self->rollback;
        $self->set_error("insert fan/freak failed");
        return $rs;
    }

    $self->commit;
    return 1;
}

sub delete {
    my ($self, $uid, $target_uid) = @_;

    if (!$uid || !$target_uid) {
        $self->set_error("no uid or target uid given");
        return;
    }

    # first, delete forward direction relationship
    my $dbh = $self->start_transaction;

    my $sql = <<"EOSQL";
UPDATE people SET type = NULL
  WHERE uid = ?
    AND person = ?
EOSQL

    my $rs = $dbh->do($sql, undef, $uid, $target_uid);
    if (!defined $rs) {
        $self->rollback;
        $self->set_error("update type failed");
        return $rs;
    }

    # next, create reverse direction relationship
    $sql = <<"EOSQL";
UPDATE people SET perceive = NULL
  WHERE uid = ?
    AND person = ?
EOSQL

    $rs = $dbh->do($sql, undef, $target_uid, $uid);
    if (!defined $rs) {
        $self->rollback;
        $self->set_error("update perceive failed");
        return $rs;
    }

    # last, remove NULL-NULL relationship
    $sql = <<"EOSQL";
DELETE FROM people
  WHERE type IS NULL
    AND perceive IS NULL
EOSQL

    $rs = $dbh->do($sql, undef, $target_uid, $uid);
    if (!defined $rs) {
        $self->rollback;
        $self->set_error("delete relation failed");
        return $rs;
    }

    $self->commit;
    return 1;
}


sub friends {
    my $self = shift;
    my $uid = shift;
    $uid = shift if ($uid && $uid eq "uid");
    return if !$uid;

    my $rs = $self->select(uid => $uid, type => 'friend');
    return if !defined $rs;

    my @result = map { $_->{person} } @$rs;
    return \@result;
}

sub foes {
    my $self = shift;
    my $uid = shift;
    $uid = shift if ($uid && $uid eq "uid");
    return if !$uid;

    my $rs = $self->select(uid => $uid, type => 'foe');
    return if !defined $rs;

    my @result = map { $_->{person} } @$rs;
    return \@result;
}

sub freaks {
    my $self = shift;
    my $uid = shift;
    $uid = shift if ($uid && $uid eq "uid");
    return if !$uid;

    my $rs = $self->select(uid => $uid, perceive => 'freak');
    return if !defined $rs;

    my @result = map { $_->{person} } @$rs;
    return \@result;
}

sub fans {
    my $self = shift;
    my $uid = shift;
    $uid = shift if ($uid && $uid eq "uid");
    return if !$uid;

    my $rs = $self->select(uid => $uid, perceive => 'fan');
    return if !defined $rs;

    my @result = map { $_->{person} } @$rs;
    return \@result;
}

sub friends_of_friends {
    my $self = shift;
    my $uid = shift;
    $uid = shift if ($uid && $uid eq "uid");
    return if !$uid;

    my $sql = <<"EOSQL";
SELECT * FROM people
  WHERE type = "friend"
    AND uid IN (SELECT person FROM people WHERE uid = ? AND type = "friend")
EOSQL
    my $dbh = $self->connect_db;
    my $sth = $dbh->prepare($sql);
    $sth->execute($uid);
    my $rs = $sth->fetchall_arrayref({});
    $self->disconnect_db;
    return if !defined $rs;

    my $result = {};
    for my $item (@$rs) {
        my $u = $item->{person};
        $result->{$u} = [] if !defined $result->{$u};
        push @{$result->{$u}}, $item->{uid};
    }
    return $result;
}

sub foes_of_friends {
    my $self = shift;
    my $uid = shift;
    $uid = shift if ($uid && $uid eq "uid");
    return if !$uid;

    my $sql = <<"EOSQL";
SELECT * FROM people
  WHERE type = "foe"
    AND uid IN (SELECT person FROM people WHERE uid = ? AND type = "friend")
EOSQL
    my $dbh = $self->connect_db;
    my $sth = $dbh->prepare($sql);
    $sth->execute($uid);
    my $rs = $sth->fetchall_arrayref({});
    $self->disconnect_db;
    return if !defined $rs;

    my $result = {};
    for my $item (@$rs) {
        my $u = $item->{person};
        $result->{$u} = [] if !defined $result->{$u};
        push @{$result->{$u}}, $item->{uid};
    }
    return $result;
}

sub hash {
    my ($self, $uid) = @_;
    return if !$uid;

    my $result = {};

    # query friends and foes, fans, freaks
    my $rs = $self->select(uid => $uid);
    return if !$rs;
    for my $item (@$rs) {
        if ($item->{type} && $item->{type} eq "friend") {
            my $k = REL_FRIEND;
            $result->{$k} = {} if !defined $result->{$k};
            $result->{$k}->{$item->{person}} = 1;
        }
        if ($item->{type} && $item->{type} eq "foe") {
            my $k = REL_FOE;
            $result->{$k} = {} if !defined $result->{$k};
            $result->{$k}->{$item->{person}} = 1;
        }
        if ($item->{perceive} && $item->{perceive} eq "fan") {
            my $k = REL_FAN;
            $result->{$k} = {} if !defined $result->{$k};
            $result->{$k}->{$item->{person}} = 1;
        }
        if ($item->{perceive} && $item->{perceive} eq "freak") {
            my $k = REL_FREAK;
            $result->{$k} = {} if !defined $result->{$k};
            $result->{$k}->{$item->{person}} = 1;
        }
    }

    # query FOF and EOF
    $rs = $self->_relations_of_friends($uid);
    if ($rs) {
        for my $item (@$rs) {
            if ($item->{type} eq "friend") {
                my $p = $item->{person};
                my $k = REL_FOF;
                $result->{$k} = {} if !defined $result->{$k};
                $result->{$k}->{$p} = {} if !defined $result->{$k}->{$p};
                $result->{$k}->{$p}->{$item->{uid}} = 1;
            }
            if ($item->{type} eq "foe") {
                my $p = $item->{person};
                my $k = REL_EOF;
                $result->{$k} = {} if !defined $result->{$k};
                $result->{$k}->{$p} = {} if !defined $result->{$k}->{$p};
                $result->{$k}->{$p}->{$item->{uid}} = 1;
            }
        }
    }

    return $result;
}

sub _relations_of_friends {
    my ($self, $uid) = @_;
    return if !$uid;

    my $sql = <<"EOSQL";
SELECT * FROM people
  WHERE type IS NOT NULL
    AND uid IN (SELECT person FROM people WHERE uid = ? AND type = "friend")
EOSQL
    my $dbh = $self->connect_db;
    my $sth = $dbh->prepare($sql);
    $sth->execute($uid);
    my $rs = $sth->fetchall_arrayref({});
    $self->disconnect_db;
    return $rs;
}

sub is_friend{
    my $self = shift;
    my $params = {@_};
    my $uid = $params->{uid};
    my $of = $params->{of};

    my $f = $self->friends(uid => $of);
    return any { $_ == $uid } @$f;
}

sub is_foe{
    my $self = shift;
    my $params = {@_};
    my $uid = $params->{uid};
    my $of = $params->{of};

    my $f = $self->foes(uid => $of);
    return any { $_ == $uid } @$f;
}

sub is_fof {
    my $self = shift;
    my $params = {@_};
    my $uid = $params->{uid};
    my $of = $params->{of};

    my $fofs = $self->friends_of_friends(uid => $of);
    my $f = [];
    for my $u (keys %$fofs) {
        push @$f, @{$fofs->{$u}};
    }
    return any { $_ == $uid } @$f;
}

sub is_eof {
    my $self = shift;
    my $params = {@_};
    my $uid = $params->{uid};
    my $of = $params->{of};

    my $fofs = $self->foes_of_friends(uid => $of);
    my $f = [];
    for my $u (keys %$fofs) {
        push @$f, @{$fofs->{$u}};
    }
    return any { $_ == $uid } @$f;
}


sub is_friend_or_fof {
    my $self = shift;
    my $params = {@_};
    my $uid = $params->{uid};
    my $of = $params->{of};

    my $fofs = $self->friends_of_friends(uid => $of);
    my $f = [];
    for my $u (keys %$fofs) {
        return 1 if $u == $uid;
        push @$f, @{$fofs->{$u}};
    }
    return any { $_ == $uid } @$f;
}

sub is_foe_or_eof {
    my $self = shift;
    my $params = {@_};
    my $uid = $params->{uid};
    my $of = $params->{of};

    my $fofs = $self->foes_of_friends(uid => $of);
    my $f = [];
    for my $u (keys %$fofs) {
        return 1 if $u == $uid;
        push @$f, @{$fofs->{$u}};
    }
    return any { $_ == $uid } @$f;
}

sub select {
    my $self = shift @_;

    my $people = $self->generic_select(table => 'people',
                                       keys => [qw(uid person type perceive fof eof)],
                                       params => {@_});
    return $people;
}

1;
