package Newslash::Model::Stories;
use Newslash::Model::Base -base;

use Newslash::Model::SlashDB;

use Newslash::Model::Stories::Text;
use Newslash::Model::Stories::RelatedStories;

use DateTime;
use DateTime::Format::MySQL;
use DateTime::Format::ISO8601;
use List::Util qw(any all);

use Data::Dumper;
use DateTime;


sub key_definition {
    return {
            table => "stories",
            primary => "stoid",
            unique => [qw(sid)],
            datetime => [qw(time last_update day_published
                            stuckendtime archive_last_update
                          )],
            other => [qw(uid dept hits discussion primaryskid
                         tid submitter commentcount hitparade
                         is_archived in_trash
                         qid body_length word_count sponsor
                         stuck stuckpos fakeemail homepage
                       )],
            aliases => { user_id => "uid",
                         id => "stoid",
                         create_time => time,
                       }
           };
}

use constant FACULTIES => { 1000 => [qw(hits hitparade)] };

sub count_by_author {
    my $self = shift;
    my $sql = <<"EOSQL";
SELECT uid, COUNT(stoid) AS count
  FROM stories
  WHERE stories.time <= NOW()
  GROUP BY uid;
EOSQL

    my $dbh = $self->connect_db;
    my $sth = $dbh->prepare($sql);
    $sth->execute();
    my $rs = $sth->fetchall_hashref("uid");
    if (!defined $rs) {
        $self->set_error("select failed", $dbh->errstr, $dbh->err);
        $self->disconnect_db;
        return;
    }
    $self->disconnect_db;
    return $rs;
}

sub count {
    my $self = shift;
    my $params = {@_};

    my $join = 'LEFT JOIN firehose ON (stories.stoid = firehose.srcid AND firehose.type = "story")';

    my @clauses;
    if ($params->{hide_future}) {
	push @clauses,  'stories.time < NOW()';
    }
    if ($params->{public_only}) {
	push @clauses, 'firehose.public != "no"';
    }
    my $where = join " AND ", @clauses;

      return $self->generic_count(table => "stories",
                                target => "stoid",
                                timestamp => "time",
                                join => $join,
                                where => $where,
                                @_);
}

##### sub models
sub text            { return shift->_get_model('::Stories::Text'); }
sub related_stories { return shift->_get_model('::Stories::RelatedStories'); }

sub _get_model {
    my ($self, $class) = @_;
    $self->{sub_models} ||= {};
    if (!$self->{sub_models}->{$class}) {
        $self->{sub_models}->{$class} = $self->new_instance_of($class);
    }
    else {
        if ($self->transaction_mode && !$self->{sub_models}->{$class}->transaction_mode) {
            $self->{sub_models}->{$class}->use_transaction($self->{_tr_dbh});
        }
    }
    return $self->{sub_models}->{$class};
}

#========================================================================

=head2 select($query_type, $value)

get a story.

=over 4

=item Parameters

=over 4

=item $query_type

query key, "sid" or "stoid"

=item $value

value for query

=back

=item Return value

HASH of story contents

=back

=cut

sub select {
    my $self = shift;
    my $params = {@_};


    my $unique_keys = { id => "stories.stoid",
                        story_id => "stories.stoid",
                        stoid => "stories.stoid",
                        sid => "stories.sid",
                      };
    my $keys = { user_id => "stories.uid",
                 uid => "stories.uid",
                 topic_id => "stories.tid",
                 tid => "stories.tid",
                 discussion_id => "stories.discussion",
                 commentcount => "stories.commentcount",
                 hits => "stories.hits",
                 submitter => "stories.submitter",
                 create_time => "stories.time",
                 update_time => "stories.last_update",
                 public => "firehose.public",
               };
    my $datetime_keys = { create_time => "stories.time",
                          update_time => "stories.last_update",
                        };
    my $timestamp = "stories.time";

    my ($where_clause, $where_values, $unique) = $self->build_where_clause(unique_keys => $unique_keys,
                                                                           keys => $keys,
                                                                           datetime_keys => $datetime_keys,
                                                                           timestamp => $timestamp,
                                                                           params => $params);
    my ($limit_clause, $limit_values) = $self->build_limit_clause(params => $params);
    my ($orderby_clause, $orderby_values) = $self->build_order_by_clause(keys => $keys,
                                                                         params => $params);

    # TODO: give reasonable LIMIT Value...
    $limit_clause = "LIMIT 50" if !$limit_clause;

    # hide future story?
    my @where_clauses;
    if ($params->{hide_future}) {
        push @where_clauses, "stories.time <= NOW()";
    }

    # hide non-public story?
    if ($params->{public_only}) {
        push @where_clauses, "firehose.public != 'no'";
    }

    # exclude "preview" item
    push @where_clauses, "globj_types.maintable != 'preview'";

    if (@where_clauses) {
        if ($where_clause) {
            $where_clause = $where_clause . " AND ";
        }
        else {
            $where_clause = "WHERE ";
        }
        $where_clause = $where_clause . join(" AND ", @where_clauses);
    }

    my @attrs;
    push @attrs, @$where_values, @$limit_values, @$orderby_values;

    my $dbh = $self->connect_db;
    my $sql = <<"EOSQL";
SELECT stories.*,
       story_text.*,
       users.nickname as author,
       firehose.public,
       firehose.popularity,
       firehose.editorpop,
       firehose.neediness,
       firehose.activity,
       firehose.globjid,
       firehose.id AS fhid,
       discussions.type AS discussion_type,
       discussions.commentcount AS comment_count
  FROM stories
    LEFT JOIN story_text
      ON stories.stoid = story_text.stoid
    LEFT JOIN users
      ON stories.uid = users.uid
    LEFT JOIN firehose 
      ON (stories.stoid = firehose.srcid AND firehose.type = "story")
    LEFT JOIN discussions
      ON firehose.discussion = discussions.id
    LEFT JOIN globjs
      ON firehose.globjid = globjs.globjid
    LEFT JOIN globj_types
      ON globjs.gtid = globj_types.gtid
    $where_clause
    $orderby_clause
    $limit_clause
EOSQL

    $self->_last_query($sql, \@attrs);

    my $sth = $dbh->prepare($sql);
    $sth->execute(@attrs);
    my $stories = $sth->fetchall_arrayref({});

    if (!$stories) {
        $self->disconnect_db();
        return;
    }
    if (@$stories == 0) {
        $self->disconnect_db();
        return $unique ? undef : [];
    }

    # get tags
    $sql = <<"EOSQL";
SELECT tags.*, tagnames.tagname, target.stoid
  FROM (SELECT stories.stoid FROM stories 
        LEFT JOIN firehose
          ON (stories.stoid = firehose.srcid AND firehose.type = "story")
        LEFT JOIN globjs
          ON firehose.globjid = globjs.globjid
        LEFT JOIN globj_types
          ON globjs.gtid = globj_types.gtid
        $where_clause $orderby_clause $limit_clause) AS target
    LEFT JOIN globjs
      ON target.stoid = globjs.target_id
    LEFT JOIN globj_types
      ON globjs.gtid = globj_types.gtid
    LEFT JOIN tags
      ON globjs.globjid = tags.globjid
    LEFT JOIN tagnames
      ON tags.tagnameid = tagnames.tagnameid
    WHERE globjs.gtid = 1
EOSQL

    $sth = $dbh->prepare($sql);
    $sth->execute(@attrs);
    my $tags_table = $sth->fetchall_arrayref({});

    # get topics
    $sql = <<"EOSQL";
SELECT story_topics_rendered.*, story_topics_chosen.weight, topics.*
  FROM (SELECT stories.stoid FROM stories
        LEFT JOIN firehose
          ON (stories.stoid = firehose.srcid AND firehose.type = "story")
        LEFT JOIN globjs
          ON firehose.globjid = globjs.globjid
        LEFT JOIN globj_types
          ON globjs.gtid = globj_types.gtid
        $where_clause $orderby_clause $limit_clause) AS target
    LEFT JOIN story_topics_rendered
      ON target.stoid = story_topics_rendered.stoid
    LEFT JOIN story_topics_chosen 
      ON story_topics_rendered.stoid = story_topics_chosen.stoid
        AND story_topics_rendered.tid = story_topics_chosen.tid
    LEFT JOIN topics
      ON story_topics_rendered.tid = topics.tid
EOSQL

    $sth = $dbh->prepare($sql);
    $sth->execute(@attrs);
    my $topics_table = $sth->fetchall_arrayref({});

    # get params
    $sql = <<"EOSQL";
SELECT story_param.*
  FROM (SELECT stories.stoid FROM stories
        LEFT JOIN firehose
          ON (stories.stoid = firehose.srcid AND firehose.type = "story")
        LEFT JOIN globjs
          ON firehose.globjid = globjs.globjid
        LEFT JOIN globj_types
          ON globjs.gtid = globj_types.gtid
        $where_clause $orderby_clause $limit_clause) AS target
    LEFT JOIN story_param
      ON target.stoid = story_param.stoid
EOSQL

    $sth = $dbh->prepare($sql);
    $sth->execute(@attrs);
    my $params_table = $sth->fetchall_arrayref({});

    # done
    $self->disconnect_db();


    my $tags = {};
    for my $tag (@$tags_table) {
        my $stoid = $tag->{stoid};
        if (!$tags->{$stoid}) {
            $tags->{$stoid} = [];
        }
        push @{$tags->{$stoid}}, $tag;
    }

    my $topics = {};
    for my $topic (@$topics_table) {
        my $stoid = $topic->{stoid};
        if (!$topics->{$stoid}) {
            $topics->{$stoid} = [];
        }
        push @{$topics->{$stoid}}, $topic;
    }

    my $params = {};
    for my $param (@$params_table) {
        my $stoid = $param->{stoid};
        if (!$params->{$stoid}) {
            $params->{$stoid} = [];
        }
        push @{$params->{$stoid}}, $param;
    }

    for my $story (@$stories) {
        my $stoid = $story->{stoid};
        $story->{tags} = $tags->{$stoid} || [];
        $story->{topics} = $topics->{$stoid} || [];
        if ($params->{$stoid}) {
            for my $param (@{$params->{$stoid}}) {
                $story->{$param->{name}} = $param->{value};
            }
        }
        $self->_generalize($story, $params);
    }

    return $stories->[0] if $unique;
    return $stories;
}

sub _check_and_regularize_params {
    my ($self, $params) = @_;
    my $msg;

    if (defined $params->{title}) {
        if (length($params->{title}) > $self->{options}->{Story}->{title_max_byte}) {
            $msg = "title too long. max: $self->{options}->{Story}->{title_max_byte} bytes";
            $self->set_error($msg, -1);
            return;
        }
    }

    $params->{commentstatus} = $params->{commentstatus} || $params->{comment_status} || "enabled";
    if (defined $params->{commentstatus}) {
        if (!grep /\A$params->{commentstatus}\z/, qw(disabled
                                                     enabled
                                                     friends_only
                                                     friends_fof_only
                                                     no_foe
                                                     no_foe_eof 
                                                     logged_in)) {
            $msg = "invalid comment_status";
            $self->set_error($msg, -1);
            return;
        }
    }

    # check timestamp. use ISO8601 style timestamp like: 2006-08-14T02:34:56-0600
    if ($params->{time}) {
        my $rex_timestamp = qr/
                                  ^(\d+)-(\d+)-(\d+)\D+(\d+):(\d+):(\d+(?:\.\d+)?)   # datetime
                                  (?:Z|([+-])(\d+):(\d+))?$                          # tz
                              /xi;
        if ($params->{time} =~ $rex_timestamp) {
            $params->{time} = "$1-$2-$3 $4:$5:$6";
        }
    }

    return 1;
}

sub _set_tags_from_topics {
    my ($self, $user, $stoid, $topics) = @_;

    return if !$stoid;
    return if !$topics;

    my $globjs = $self->new_instance_of("Newslash::Model::Globjs");
    my $globj_id = $globjs->getGlobjidFromTargetIfExists("stories", $stoid);
    if ($globj_id) {
        # set tags
        my $tags = $self->new_instance_of("Tags");
        for my $tid (keys %$topics) {
            my $ret = $tags->set_tag(uid => $user->{uid} || $user->{user_id},
                                     tagname_id => $tid,
                                     globj_id => $globj_id,
                                     private => 0,
                                    );
            #warn "set_tag fault..." if !$ret
        }
    }
    return $stoid;
}

=head2 update

this implementation uses old slash's updateStory($sid, $data),
$sid is takable sid or stoid.

=cut

sub update {
    #my ($self, $params, $user, $extra_params, $opts) = @_;
    my $self = shift;
    my $params = {@_};
    my $user = $params->{user};

    # check id
    my $story;
    my $stoid = $params->{stoid} || $params->{story_id} || $params->{id};
    my $sid = $params->{sid};
    if (!$story) {
        if ($stoid) {
            $story = $self->select(stoid => $stoid);
        }
        elsif ($sid) {
            $story = $self->select(sid => $sid);
        }
    }
    if ($story) {
        $sid = $story->{sid};
        $stoid = $story->{stoid};
    }

    if (!$story) {
        $self->last_error("no story given");
        return;
    }

    # check params
    return if !$self->_check_and_regularize_params($params);

    my $slash_db = Newslash::Model::SlashDB->new($self->{options});

    my $add_related = $params->{add_related};
    delete $params->{add_related} if $add_related;
    my $add_tags = $params->{add_tags};
    delete $params->{add_tags} if $add_tags;

    # add primaryskid
    if (!$params->{primaryskid}) {
	$params->{primaryskid} = 1; # mainpage
    }

    # check tid
    my $topic_nexus = $self->new_instance_of('Topics')->get_topic_nexus();
    if (!$params->{tid} || all { $params->{tid} != $_->{tid} } @$topic_nexus) {
	$params->{tid} = 49; # mainpage
    }

    $stoid = $slash_db->updateStory($stoid, $params);
    return if !$stoid;

    $self->_set_tags_from_topics($params->{user}, $stoid, $params->{topics_chosen});

    # set tags
    if ($add_tags) {
        my $tags = $self->new_instance_of("Tags");
        my $globjs = $self->new_instance_of("Newslash::Model::Globjs");
        my $globj_id = $globjs->getGlobjidFromTargetIfExists("stories", $params->{stoid});
        for my $tag (@$add_tags) {
            my $rs = $tags->add(uid => $user->{uid} || $user->{user_id},
                                globj_id => $globj_id,
                                name => $tag);
            if (!defined $rs) {
                $self->logger->warn("Stories::update: tag $tag set failed: " . $tags->last_error);
            }
        }
    }

    # insert related story
    if ($add_related) {
        for my $related_sid (@$add_related) {
            my $rs = $self->add_related_story(sid => $sid,
                                              related_sid => $related_sid);
            if (!defined $rs) {
                $self->logger->warn("Stories::update: insert related story $related_sid to $sid failed: " . $self->last_error);
            }

            $rs = $self->add_related_story(sid => $related_sid,
                                           related_sid => $sid);
            if (!defined $rs) {
                $self->logger->warn("Stories::update: insert related story $sid to $related_sid failed: " . $self->last_error);
            }
        }
    }

    return wantarray ? ($sid, $stoid) : $stoid;
}

sub update2 {
    my $self = shift;
    my $params = {@_};
    return $self->generic_update(params => $params);
}


=head2 create(\%params, $uid)

create a story.

=over 4

=item Parameters

=over 4

=item \%params

parameters

$params->{fhid}
$params->{subid}

=item $uid

author's uid

=back

=item Return value

stoid

=back

=cut

sub create {
    #my ($self, $params, $user, $extra_params, $opts) = @_;
    my $self = shift;
    return if $self->check_readonly;

    my $params = {@_};
    my $user = $params->{user};

    # check parameters
    my $msg = "";
    $msg = "no_title" if !$params->{title};
    $msg = "no_introtext" if !$params->{introtext} && !$params->{intro_text};
    $msg = "no_topics" if !defined $params->{topics_chosen};
    $msg = "invalid_user" if !defined $user->{uid};

    if (length($params->{title}) > $self->{options}->{Story}->{title_max_byte}) {
        $msg = "title too long. max: $self->{options}->{Story}->{title_max_byte} bytes";
    }

    $params->{commentstatus} = $params->{commentstatus} || $params->{comment_status} || "enabled";
    if (!grep /\A$params->{commentstatus}\z/, qw(disabled enabled friends_only friends_fof_only no_foe no_foe_eof logged_in)) {
        $msg = "invalid comment_status";
    }

    # check timestamp. use ISO8601 style timestamp like: 2006-08-14T02:34:56-0600
    if ($params->{time}) {
        my $rex_timestamp = qr/
                                  ^(\d+)-(\d+)-(\d+)[^ 0-9]+(\d+):(\d+):(\d+(?:\.\d+)?)   # datetime
                                  (?:Z|([+-])(\d+):(\d+))?$                          # tz
                              /xi;
        if ($params->{time} =~ $rex_timestamp) {
            my $dt = DateTime::Format::ISO8601->parse_datetime($params->{time});
            $params->{time} = DateTime::Format::MySQL->format_datetime($dt);
        }
    }

    # check parameters finish
    if (length($msg) > 0) {
        $self->set_error($msg, -1);
        return;
    }
    # if neverdisplay is set, the story not included in RSS...
    #$params->{neverdisplay} ||= 0;
    $params->{uid} = $user->{uid};

    # check submission
    if ($params->{submission_id}) {
	my $submission = $self->new_instance_of("Submissions")->select(id => $params->{submission_id});
	if ($submission) {
	    $params->{submitter} = $submission->{uid};
	    $params->{fhid} = $submission->{fhid};
	}
	else {
	    $self->logger->error("Stories::create - invalid submission_id: $params->{submission_id}");
	}
    }
    if ($params->{journal_id}) {
	my $journal = $self->new_instance_if("Journals")->select(id => $params->{journal_id});
	if ($journal) {
	    $params->{submitter} = $journal->{uid};
	    $params->{fhid} = $journal->{fhid};
	}
	else {
	    $self->logger->error("Stories::create - invalid journal_id: $params->{journal_id}");
	}
    }
    $params->{submitter} ||= $user->{uid};

    # createStory() deletes topics_chosen, so need to save here.
    my $topics_chosen = $params->{topics_chosen};

    # check tid
    my $topic_nexus = $self->new_instance_of('Topics')->get_topic_nexus();
    if (!$params->{tid} || all { $params->{tid} != $_->{tid} } @$topic_nexus) {
	$params->{tid} = 49; # mainpage
    }

    # add primaryskid
    if (!$params->{primaryskid}) {
	$params->{primaryskid} = 1; # mainpage
    }

    my $add_related = $params->{add_related};
    delete $params->{add_related} if $add_related;
    my $add_tags = $params->{add_tags};
    delete $params->{add_tags} if $add_tags;

    my $slash_db = Newslash::Model::SlashDB->new($self->{options});
    my ($sid, $stoid);
    if ($params->{update}) {
        $stoid = $params->{stoid} || $params->{story_id} || $params->{id};
        $sid = $slash_db->updateStory($stoid, $params);
        $self->set_error("updateStory failed");
        return if !$sid;
    }
    else {
        ($sid, $stoid) = $slash_db->createStory($params);
    }

    my $globjs = $self->new_instance_of("Newslash::Model::Globjs");
    my $globj_id = $globjs->getGlobjidFromTargetIfExists("stories", $params->{stoid});

    # set topics
    my $tags = $self->new_instance_of("Tags");
    if ($globj_id) {
        for my $tid (keys %$topics_chosen) {
            my $rs = $tags->add(uid => $user->{uid} || $user->{user_id},
                                tagname_id => $tid,
                                globj_id => $globj_id,
                                private => 0,
                               );
            if (!defined $rs) {
                $self->logger->warn("Stories::create: tag $tid set failed: " . $tags->last_error);
            }
        }
    }

    # set tags
    if ($add_tags) {
        for my $tag (@$add_tags) {
            my $rs = $tags->add(uid => $user->{uid} || $user->{user_id},
                                globj_id => $globj_id,
                                name => $tag);
            if (!defined $rs) {
                $self->logger->warn("Stories::create: tag $tag set failed: " . $tags->last_error);
            }
        }
    }

    # insert related story
    if ($add_related) {
        for my $related_sid (@$add_related) {
            my $rs = $self->add_related_story(sid => $sid,
                                              related_sid => $related_sid);
            if (!defined $rs) {
                $self->logger->warn("insert related story failed: " . $self->last_error);
            }

            $rs = $self->add_related_story(sid => $related_sid,
                                           related_sid => $sid);
            if (!defined $rs) {
                $self->logger->warn("Stories::create: insert related story failed: " . $self->last_error);
            }
        }
    }

    # add firehose item to related_story
    if ($params->{submission_id}) {
	my $submissions = $self->new_instance_of("Submissions");
	my $submission = $submissions->select(id => $params->{submission_id});
	if ($submission) {
            my $rs = $self->add_related_firehose(sid => $sid,
						 fhid => $submission->{fhid});
            if (!defined $rs) {
                $self->logger->error("Stories::create - insert related firehose failed: " . $self->last_error);
            }

	    # change submission status
	    if (!$submissions->update(id => $submission->{subid}, del => 2)) {
		$self->logger->error("Stories::create - cannot change submission status");
	    }
	}
	else {
	    $self->logger->error("Stories::create - invalid submission_id: $params->{submission_id}");
	}
    }

    

    return wantarray ? ($sid, $stoid) : $stoid;
}

sub create2 {
    my $self = shift;
    my $params = {@_};
    return $self->generic_insert(params => $params);
}

sub allocate_sid {
    my ($self, @params) = @_;
    my $params = {@params};
    my $dt = $params->{base_datetime} || DateTime->now;

    # create sid from timestamp
    # my $sid_format = '%02d/%02d/%02d/%02d%0d2%02d';
    my $sid_format = '%y/%m/%d/%H%M%S';
    my $sid = $dt->strftime($sid_format);

    # insert blank story with given sid
    my $dbh = $self->connect_db;
    my $sql = "INSERT INTO stories (sid) VALUES (?)";

    my $n = 100; # retry 100 times
    while (--$n) {
        my $rs = $dbh->do($sql, undef, $sid);
        if (!defined $rs) {
            $self->set_error("sid_insert_error", -1);
            return;
        }
        if ($rs) {
            my $stoid = $dbh->last_insert_id(undef, undef, undef, undef);
            $self->disconnect_db;
            return ($sid, $stoid);
        }

        # allocate failed, so recreate sid
        $dt->subtract( seconds => 1 );
        $sid = $dt->strftime($sid_format);
    }
    $self->set_error("sid_allocate_failed", -1);
    $self->disconnect_db;
    return;
}


# Legacy API
sub createSid {
    my ($self, $bogus_sid) = @_;
    # yes, this format is correct, don't change it :-)
    my $sidformat = '%02d/%02d/%02d/%02d%0d2%02d';
    # Create a sid based on the current time.
    my @lt;
    my $start_time = time;
    if ($bogus_sid) {
        # If we were called being told that there's at
        # least one sid that is invalid (already taken),
        # then look backwards in time until we find it,
        # then go one second further.
        my $loops = 1000;
        while (--$loops) {
            $start_time--;
            @lt = localtime($start_time);
            $lt[5] %= 100; $lt[4]++; # year and month
            last if $bogus_sid eq sprintf($sidformat, @lt[reverse 0..5]);
        }
        if ($loops) {
            # Found the bogus sid by looking
            # backwards.  Go one second further.
            $start_time--;
        } else {
            # Something's wrong.  Skip ahead in
            # time instead of back (not sure what
            # else to do).
            $start_time = time + 1;
        }
    }
    @lt = localtime($start_time);
    $lt[5] %= 100; $lt[4]++; # year and month
    return sprintf($sidformat, @lt[reverse 0..5]);
}


=head2 get_histories

=cut

sub get_histories {
}

=head2 get_related_items($stoid)

get related links.

=over 4

=item Parameters

=over 4

=item $stoid

story id

=back

=item Return value

ARRAY of related links

=back

=cut

sub get_related_items {
    my $self = shift;
    my $params = {@_};
    my $stoid = $params->{stoid} || $params->{story_id} || $params->{id};
    return if !$stoid;

    my $dbh = $self->connect_db;

    my $sql = <<"EOSQL";
SELECT related.*, 
       story_text.title as title2,
       firehose.*,
       stories.*,
       topics.*
  FROM (
    SELECT * FROM related_stories
      WHERE stoid = ?
      ORDER BY ordernum ASC
    ) AS related
  LEFT JOIN story_text ON story_text.stoid = related.rel_stoid
  LEFT JOIN firehose ON firehose.id = related.fhid
  LEFT JOIN stories ON stories.sid = related.rel_sid
  LEFT JOIN topics ON topics.tid = stories.tid
EOSQL

    my $sth = $dbh->prepare($sql);
    $sth->execute($stoid);
    my $related = $sth->fetchall_arrayref({});
    $self->disconnect_db();

    for my $r (@$related) {
        $r->{create_time} = $r->{time};
        $r->{title} = $r->{title2} unless $r->{title};
        if ($r->{rel_sid}) {
            $r->{type} = "story";
            $r->{key_id} = $r->{rel_sid};
        } else {
            $r->{type} = "submission";
            $r->{key_id} = $r->{srcid};
        }
        $r->{primary_topic} = {};
        $r->{primary_topic}->{tid} = $r->{tid};
        for my $k (qw{keyword textname series image width height
                      submittable searchable storypickable usesprite}) {
            next if !$r->{$k};
            $r->{primary_topic}->{$k} = $r->{$k};
            delete $r->{$k};
        }
    }

    return $related;
}

sub add_related_story {
    my $self = shift;
    my $args = {@_};

    my $story = $args->{story};
    if (!$story) {
        if ($args->{stoid}) {
            $story = $self->select(stoid => $args->{stoid});
        } elsif ($args->{sid}) {
            $story = $self->select(sid => $args->{sid});
        }
    }

    if (!$story || !$story->{stoid} || !$story->{sid}) {
        $self->last_error("cannot get target story");
        return;
    }

    my $related = $args->{related};
    if (!$related) {
        if ($args->{related_stoid}) {
            $related = $self->select(stoid => $args->{related_stoid});
        } elsif ($args->{related_sid}) {
            $related = $self->select(sid => $args->{related_sid});
        }
    }

    if (!$related || !$related->{stoid} || !$related->{sid} ) {
        $self->last_error("cannot get related story");
        return;
    }

    # get related item count
    my $dbh = $self->start_transaction;
    my $sql = "SELECT rel_stoid FROM related_stories WHERE stoid = ?";
    my $stoids = $dbh->do($sql, undef, $story->{stoid});
    $stoids = [] if ref($stoids) ne "ARRAY";

    # check if related story is already set
    if (any { $_ eq $related->{stoid} } @$stoids) {
        $self->commit;
        return 0;
    }

    # insert
    my $count = @$stoids;
    my $rs = $self->related_stories->insert(stoid => $story->{stoid},
                                            rel_stoid => $related->{stoid},
                                            rel_sid => $related->{sid},
                                            ordernum => $count + 1);

    if (!defined $rs) {
        $self->last_error("insert failed: " . $self->related_stories->last_error);
        $self->rollback;
        return;
    }
    $self->commit;
    return 1;
}

sub add_related_firehose {
    my $self = shift;
    my $args = {@_};

    my $story = $args->{story};
    if (!$story) {
        if ($args->{stoid}) {
            $story = $self->select(stoid => $args->{stoid});
        } elsif ($args->{sid}) {
            $story = $self->select(sid => $args->{sid});
        }
    }

    if (!$story || !$story->{stoid} || !$story->{sid}) {
        $self->last_error("cannot get target story");
        return;
    }

    my $fh_item = $args->{firehose_item};
    if (!$fh_item) {
        if ($args->{fhid}) {
            $fh_item = $self->new_instance_of("Firehose")->select(id => $args->{fhid});
        }
    }

    if (!$fh_item) {
        $self->last_error("cannot get related firehose item");
        return;
    }

    # get related item count
    my $dbh = $self->start_transaction;
    my $sql = "SELECT rel_stoid FROM related_stories WHERE stoid = ?";
    my $stoids = $dbh->do($sql, undef, $story->{stoid});
    $stoids = [] if ref($stoids) ne "ARRAY";

    # insert
    my $count = @$stoids;
    my $rs = $self->related_stories->insert(stoid => $story->{stoid},
                                            rel_stoid => 0,
                                            rel_sid => 0,
					    title => "Firehose: $fh_item->{title}",
					    fhid => $fh_item->{id},
                                            ordernum => $count + 1);

    if (!defined $rs) {
        $self->last_error("insert failed: " . $self->related_stories->last_error);
        $self->rollback;
        return;
    }
    $self->commit;
    return 1;
}


=head2 parameters($stoid)

get story parameters.

=over 4

=item Parameters

=over 4

=item $stoid

story id

=back

=item Return value

HASH of parameters

=back

=cut

sub parameters {
    my ($self, $stoid) = @_;

    my $dbh = $self->connect_db;

    my $sql = <<"EOSQL";
SELECT * FROM story_param WHERE stoid = ?
EOSQL

    my $sth = $dbh->prepare($sql);
    $sth->execute($stoid);
    my $params = $sth->fetchall_hashref('name');
    $sth->finish;

    # get next/prev story info
    if ($params->{next_stoid} || $params->{prev_stoid}) {
        $sql = <<"EOSQL";
SELECT stories.*, story_text.* FROM stories JOIN story_text ON stories.stoid = story_text.stoid WHERE stories.stoid = ? OR stories.stoid = ?
EOSQL
        $sth = $dbh->prepare($sql);
        my $next = $params->{next_stoid}->{value} || 0;
        my $prev = $params->{prev_stoid}->{value} || 0;
        $sth->execute($next, $prev);
        my $sids = $sth->fetchall_hashref('stoid');
        $params->{next_stoid}->{story} = $sids->{$next} if $sids->{$next};
        $params->{prev_stoid}->{story} = $sids->{$prev} if $sids->{$prev};
        $sth->finish;
    }

    $self->disconnect_db();
    return $params;
}

#========================================================================

=head2 set_dirty($key, $id)

set writestatus dirty for the story.

=over 4

=item Parameters

=over 4

=item $key

'stoid'

=item $id

id of the story

=back

=item Return value

1/0

=back

=cut

sub set_dirty {
    my ($self, $key, $id) = @_;
    return if $self->check_readonly;

    my $stoid;
    if ($key eq 'stoid') {
        $stoid = $id;
    }
    else {
        return;
    }

    my $dbh = $self->connect_db;
    my $sql = <<"EOSQL";
INSERT INTO story_dirty (stoid) VALUES (?)
EOSQL

    my $rs = $dbh->do($sql, undef, $stoid);
    if (!$rs) {
        return;
    }
    return 1;
}

sub _generalize {
    my ($self, $story, $params) = @_;
    $params ||= {};

    # NTO-nized
    $story->{id} = $story->{stoid};
    $story->{story_id} = $story->{stoid};
    $story->{create_time} = $story->{time};
    $story->{update_time} = $story->{last_update};

    for my $t (@{$story->{topics}}) {
        if ($t->{tid} && $t->{tid} == $story->{tid}) {
            $story->{primary_topic} = $t;
        }
    }

    $story->{content_type} = "story";
    $story->{intro_text} = $story->{introtext};
    $story->{bodytext} ||= "";
    $story->{body_text} = $story->{bodytext};
    if ($story->{body_text}) {
        $story->{full_text} = join("\n", $story->{intro_text}, $story->{body_text});
    }
    else {
        $story->{full_text} = $story->{intro_text};
    }
    $story->{fulltext} = $story->{full_text};

    $story->{discussion_id} = $story->{discussion};

    # no public flag given, public is 'yes'
    $story->{public} = 'yes' if !$story->{public};

}

# delete story from database
# this method is for test purpose only.
sub hard_delete {
    my $self = shift;
    return if $self->check_readonly;
    my $params = {@_};

    my $stoid = $params->{story_id};
    return if !$stoid;

    my $error = 0;
    my $sql;

    #my $dbh = $self->connect_db({AutoCommit => 0,});
    my $dbh = $self->start_transaction;
    for my $table (qw(stories story_param story_text story_topics_chosen story_topics_rendered)) {
        my $sql = "DELETE FROM $table WHERE stoid = ?";
        my $rs = $dbh->do($sql, undef, $stoid);
        if (!defined $rs || $rs == 0) {
            Mojo::Log->new->warn("DELETE FROM $table failed. stoid is $stoid.");
            $error = 1;
        }
    }

    # delete from firehose
    my $firehose = $self->new_instance_of("Firehose");
    $firehose->hard_delete("story", $stoid);

    $self->commit;
    return !$error;

    # delete globjs
    # delete tags

    return;
}

1;
