package Newslash::Model::Submissions;
use Newslash::Model::Base -base;

use Data::Dumper;
use Mojo::Log;

use Mail::Address;
use Digest::MD5 qw(md5_hex);
use Encode;

use Newslash::Model::Globjs;
use Newslash::Model::Discussions;
use Newslash::Model::Urls;

use constant FACULTIES => { 1000 => [qw(ipid subnetid note comment)] };

sub count {
    my $self = shift;
    my $params = {@_};
    my $join = 'LEFT JOIN firehose ON (submissions.subid = firehose.srcid AND firehose.type = "submission")';

    my @clauses;
    if ($params->{hide_future}) {
	push @clauses,  'submission.time < NOW()';
    }
    if ($params->{public_only}) {
	push @clauses, 'firehose.public != "no"';
    }
    my $where = join " AND ", @clauses;

    return $self->generic_count(table => "submissions",
                                target => "subid",
                                timestamp => "time",
                                join => $join,
                                where => $where,
                                @_);
}

sub count_by_author {
    my $self = shift;
    my $args = {@_};
    my $limit = $args->{limit} || 10;
    my $years = $args->{years};

    my @attr;
    my $where_clause = "";
    if ($years =~ m/^\d+$/ ) {
        $where_clause = "WHERE time >= DATE_ADD(NOW(), INTERVAL - ? YEAR)";
        push @attr, $years;
    }

    my $sql = <<"EOSQL";
SELECT users.*, COUNT(subid) AS count
  FROM submissions
  JOIN users USING(uid)
  $where_clause
  GROUP BY uid
  ORDER BY COUNT(subid) DESC
  LIMIT ?
EOSQL
    push @attr, $limit;

    $self->_last_query($sql, \@attr);

    my $dbh = $self->connect_db;
    my $sth = $dbh->prepare($sql);
    $sth->execute(@attr);
    my $rs = $sth->fetchall_arrayref({});
    if (!defined $rs) {
        $self->set_error("select failed", $dbh->errstr, $dbh->err);
        $self->disconnect_db;
        return;
    }
    $self->disconnect_db;
    return $rs;
}

#========================================================================

=head2 select($query_type, $value)

get a story.

=over 4

=item Parameters

=over 4

=item $query_type

query key, "subid"

=item $value

value for query

=back

=item Return value

HASH of story contents

=back

=cut

sub select {
    my $self = shift;
    my $params = {@_};

    my $unique_keys = { id => "submissions.subid",
                        submission_id => "submissions.subid",
                        subid => "submissions.subid",
                        signature => "submissions.signature",
                        discussion_id => "firehose.discussion",
                        discussion => "firehose.discussion",
                      };
    my $keys = { user_id => "submissions.uid",
                 uid => "submissions.uid",
                 topic_id => "submissions.tid",
                 tid => "submissions.tid",
                 create_time => "submissions.time",
                 note => "submissions.note",
                 del => "submissions.del",
                 public => "firehose.public",
                 popularity => "firehose.popularity",
               };
    my $datetime_keys = { create_time => "submissions.time" };
    my $timestamp = "submissions.time";

    my ($where_clause, $where_values, $unique) = $self->build_where_clause(unique_keys => $unique_keys,
                                                                           keys => $keys,
                                                                           datetime_keys => $datetime_keys,
                                                                           timestamp => $timestamp,
                                                                           params => $params);
    my ($limit_clause, $limit_values) = $self->build_limit_clause(params => $params);
    my ($orderby_clause, $orderby_values) = $self->build_order_by_clause(unique_keys => $unique_keys,
                                                                         keys => $keys,
                                                                         params => $params);

    # TODO: give reasonable LIMIT Value...
    $limit_clause = "LIMIT 50" if !$limit_clause;

    # show future story?
    my @where_clauses;
    if (!$params->{show_future}) {
        push @where_clauses, "submissions.time <= NOW()";
    }

    # show non-public story?
    if (!$params->{show_nonpublic} && !$params->{public}) {
        push @where_clauses, "firehose.public != 'no'";
    }

    if (@where_clauses) {
        if ($where_clause) {
            $where_clause = $where_clause . " AND ";
        }
        else {
            $where_clause = "WHERE ";
        }
        $where_clause = $where_clause . join(" AND ", @where_clauses);
    }

    my @attrs;
    push @attrs, @$where_values, @$limit_values, @$orderby_values;

    my $sql = <<"EOSQL";
SELECT submissions.*,
       users.nickname AS author,
       firehose.toptags,
       firehose.discussion,
       firehose.public,
       firehose.popularity,
       firehose.editorpop,
       firehose.neediness,
       firehose.activity,
       firehose.id AS fhid,
       topics.*,
       urls.url,
       discussions.type AS discussion_type,
       discussions.commentcount AS comment_count
  FROM submissions
    LEFT JOIN users ON submissions.uid = users.uid
    LEFT JOIN firehose
      ON (submissions.subid = firehose.srcid AND firehose.type = "submission")
    LEFT JOIN discussions ON firehose.discussion = discussions.id
    LEFT JOIN topics ON submissions.tid = topics.tid
    LEFT JOIN urls ON firehose.url_id = urls.url_id
  $where_clause
  $orderby_clause
  $limit_clause
EOSQL


    my $dbh = $self->connect_db;
    my $sth = $dbh->prepare($sql);
    $sth->execute(@attrs);
    my $items = $sth->fetchall_arrayref({});


    if (!$items) {
        $self->disconnect_db();
        return;
    }
    if (@$items == 0) {
        $self->disconnect_db();
        return $unique ? undef : [];
    }

    # get tags
    my $tags_model = $self->new_instance_of("Tags");
    my $tags = $tags_model->select_for_items("submission", "subid", $items) || {};

    my $sec_lv = $params->{security_level} || 0;

    for my $item (@$items) {
        my $subid = $item->{subid};
        $item->{tags} = $tags->{$subid} if $tags->{$subid};
        $self->_generalize($item);
    }
    return $unique ? $items->[0] : $items;
}

sub _generalize {
    my ($self, $submission) = @_;

    return if !$submission;
    return {} if (keys %$submission == 0);

    # NTO-nized
    $submission->{title} = $submission->{subj};
    $submission->{intro_text} = $submission->{story};
    $submission->{introtext} = $submission->{story}; # for compatible
    $submission->{full_text} = $submission->{intro_text};
    $submission->{fulltext} = $submission->{full_text};

    $submission->{drafttext} = $self->create_draft_text($submission);  # for compatible
    $submission->{draft_text} = $self->create_draft_text($submission);

    $submission->{id} = $submission->{subid};
    $submission->{submission_id} = $submission->{subid};
    $submission->{discussion_id} = $submission->{discussion};

    $submission->{create_time} = $submission->{time};
    $submission->{update_time} = $submission->{time};
    $submission->{content_type} = 'submission';

    if ($submission->{toptags}) {
        my @tags = split /\s+/, $submission->{toptags};
        $submission->{toptags} = \@tags;
    }

    my $primary_topic = {};
    for my $item (qw(tid keyword textname series image width height submittable searchable storypickable usesprite)) {
        $primary_topic->{$item} = $submission->{$item};
    }
    $submission->{primary_topic} = $primary_topic;
    $submission->{topics} = [$primary_topic,];

    # no public flag given, public is 'yes'
    $submission->{public} = 'yes' if !$submission->{public};

    return $submission;
}

sub _email_to_domain {
    my ($email) = @_;
    if (!$email) {
        return '';
    }
    my $addr = Mail::Address->new('', $email);
    return '' if !$addr;
    my $host = $addr->host();
    return '' if !$host;
    return fullhost_to_domain($host);
}

sub fullhost_to_domain {
    my($fullhost) = @_;
    my $info = lc $fullhost;
    if ($info =~ m/^([\d.]+)\.in-addr\.arpa$/) {
        $info = join('.', reverse split /\./, $1);
    }
    if ($info =~ m/^(\d{1,3}\.){3}\d{1,3}$/) {
        # leave a numeric IP address alone
    } elsif ($info =~ m/([\w-]+\.[a-z]{3,4})$/) {
        # a.b.c.d.com -> d.com
        $info = $1;
    } elsif ($info =~ m/([\w-]+\.[a-z]{2,4}\.[a-z]{2})$/) {
        # a.b.c.d.co.uk -> d.co.uk
        $info = $1;
    } elsif ($info =~ m/([\w-]+\.[a-z]{2})$/) {
        # a.b.c.realdomain.gr -> realdomain.gr
        $info = $1;
    } else {
        # any other a.b.c.d.e -> c.d.e
        my @info = split /\./, $info;
        my $num_levels = scalar @info;
        if ($num_levels >= 3) {
            $info = join('.', @info[-3..-1]);
        }
    }
    return $info;
}

sub generate_signature {
    my ($self, $introtext, $title) = @_;
    return md5_hex(encode_utf8($introtext . $title));
}

#========================================================================

=head2 create(\%params, \%options)

insert submission

=over 4

=item Parameters

=over 4

=item \%options

submission parameters

 * title
 * url
 * email
 * introtext
 * tags

=item \%options

options for submission

=back

=item Return value

submission id

=back

=cut

sub create {
    #my ($self, $params, $user, $extra_params, $opts) = @_;
    my $self = shift;
    my $params = {@_};
    return if $self->check_readonly;

    my $extra_params = $params->{extra_params} || {};

    my $user = $params->{user};
    return if !$user;

    # TODO: save headers to submission_param is needed?

    my $dbh = $self->start_transaction;
    my $sql = <<"EOSQL";
INSERT INTO submissions
    (email, emaildomain, name, time,  subj, story, tid, primaryskid, uid, ipid, subnetid, signature)
   VALUES
    (?,     ?,           ?,    NOW(), ?,    ?,     ?,   ?,           ?,   ?,    ?,        ?)
EOSQL

    $params->{email} ||= '';

    # set URL
    if ($params->{url}) {
        my $urls = $self->new_instance_of("Urls");
        my $url_data = {url => $params->{url}};
        my $url_id = $urls->getUrlCreate($url_data);
        $params->{url_id} = $url_id;
        $extra_params->{url_id} = $url_id;
    } else {
        $params->{url_id} = 0;
    }

    # check signature
    my $sig = $self->generate_signature($params->{introtext} || $params->{intro_text}, $params->{title});
    my $rs = $self->select(signature => $sig);
    if ($rs) {
        $self->set_error("duplicated signature");
        $self->rollback;
        return 0;
    }

    # do insert
    $dbh->do($sql, undef,
             $params->{email}, #email
             _email_to_domain($params->{email}), #emaildomain
             $user->{nickname}, #name
             $params->{title}, #subj
             $params->{introtext} || $params->{intro_text} || "", #story
             $params->{tid} || $params->{topic_id}, #tid
             1, #primaryskid (1 == 'mainpage')
             $user->{uid} || $user->{user_id}, #uid,
             $user->{ipid} || $user->{ip_id}, #ipid,
             $user->{subnetid} || $user->{subnet_id}, #subnetid,
             $sig #signature
            );
    my $sub_id = $dbh->last_insert_id(undef, undef, undef, undef);

    if (!$sub_id) {
        # error occured
        $self->set_error($dbh->errstr, $dbh->err);
        $self->rollback;
        return;
    }

    $sql = "INSERT INTO submission_param (subid, name, value) VALUES (?, ?, ?)";
    my @param_ids;
    if ($extra_params) {
        for my $k (keys %$extra_params) {
            my $param_id = $dbh->do($sql, undef, $sub_id, $k, $extra_params->{$k});
            if (!$param_id) {
                # error occured
                $self->set_error($dbh->errstr, $dbh->err);
                $self->rollback;
                return;
            }
            push @param_ids, $param_id;
        }
    }

    # get creationtime
    my $sth = $dbh->prepare("SELECT time FROM submissions WHERE subid = ?");
    $sth->execute($sub_id);
    $rs = $sth->fetchrow_hashref;
    if (!$rs) {
        # error occured
        $self->set_error($dbh->errstr, $dbh->err);
        $self->rollback;
        return;
    }
    $sth->finish;
    my $create_time = $rs->{time};

    # create firehose entry
    my $globjs = Newslash::Model::Globjs->new($self->{options});
    my $globj_id = $globjs->getGlobjidCreate("submissions", $sub_id);
    if (!$globj_id) {
        # error occured
        warn "invalid globjid";
        $self->set_error("invalid globjid");
        $self->rollback;
        return;
    }

    # TODO: sprite_info column don't have default value...
    $sql = <<"EOSQL";
INSERT INTO firehose
    (uid, globjid, type, createtime, popularity, editorpop, public, attention_needed, primaryskid, tid, srcid,
     url_id, email, emaildomain, name, ipid, subnetid, sprite_info)
  VALUES
    (?,  ?,        ?,    ?,          ?,          ?,         ?,      ?,                ?,           ?,   ?,
     ?,      ?,     ?,           ?,    ?,    ?,        "")
EOSQL

    # my $midpop = $self->getEntryPopularityForColorLevel(5);
    # argument is 5, then value is "102, 93". entry is 102, min is 93
    # '290,240 220,200 185,175 155,138 102,93 30,25 0,-20 -60,-999999'
    my $midpop = 102;

    $rs = $dbh->do($sql, undef,
                   $user->{uid} || $user->{user_id}, #uid
                   $globj_id, #globjid
                   "submission", #type
                   $create_time, #createtime
                   $midpop, #popularity
                   $midpop, #editorpop
                   "yes", #public
                   "yes", #attention_needed
                   1, #primaryskid, 1 == mainpage
                   $params->{tid} || $params->{topic_id}, #tid
                   $sub_id, #srcid
                   $params->{url_id}, #url_id
                   $params->{email}, #email
                   _email_to_domain($params->{email}), #emaildomain
                   $user->{nickname}, #name
                   $user->{ipid} || $user->{ip_id}, #ipid
                   $user->{subnetid} || $user->{subnet_id} #subnetid
                  );
    if (!$rs) {
        # error occured
        $self->set_error($dbh->errstr, $dbh->err);
        $self->rollback;
        return;
    }
    my $firehose_id = $dbh->last_insert_id(undef, undef, undef, undef);

    $sql = "INSERT INTO firehose_text (title, introtext) VALUES (?, ?)";
    $rs = $dbh->do($sql, undef, $params->{title}, $params->{introtext} || $params->{intro_text});
    if (!$rs) {
        # error occured
        $self->set_error($dbh->errstr, $dbh->err);
        $self->rollback;
        return;
    }
    my $firehose_text_id = $dbh->last_insert_id(undef, undef, undef, undef);

    # set tags
    my $tags = $self->new_instance_of("Tags");
    my @tag_names = split(/\s+/, $params->{tags_string});
    for my $tag_name (@tag_names) {
        next if $tags->is_reserved_name($tag_name);
        my $new_id = $tags->set_tag(globj_id => $globj_id,
                                    uid => $user->{uid},
                                    name => $tag_name);
        if (!$new_id) {
            $self->logger->warn("setting tag failed: $tag_name to $globj_id");
        }
    }

    # add "submission" tag
    my $new_id = $tags->set_tag(globj_id => $globj_id,
                                uid => $user->{uid},
                                name => "submission");
    if (!$new_id) {
        $self->logger->warn("setting tag failed: submission to $globj_id");
    }

    # in old slash, add 'nod' tag when author is not anonymous, but ommit them.

    # create discussion
    #my $discuss = Newslash::Model::Discussions->new($self->{options});
    my $discuss = $self->new_instance_of("Discussions");
    my $default_comment_status = 'logged_in';
    #my $defualt_url = "$constants->{rootdir}/firehose.pl?op=view&id=$firehose_id";
    my $default_url = "/firehose.pl?op=view&id=$firehose_id";

    my $discussion_id = $discuss->create(uid           => 0,
                                         kind          => 'submission',
                                         title         => $params->{title},
                                         topic         => $params->{tid} || $params->{topic_id},
                                         primaryskid   => 1, #primaryskid, 1 == mainpage
                                         commentstatus => $default_comment_status,
                                         url           => $default_url,
                                         ts            => $create_time,
                                        );
    if (!$discussion_id) {
        # error occured
        $self->set_error($dbh->errstr, $dbh->err);
        $self->rollback;
        return;
    }
    $sql = "UPDATE firehose SET discussion = ? WHERE id = ?";
    $rs = $dbh->do($sql, undef, $discussion_id, $firehose_id);
    if (!$rs) {
        # error occured
        $self->set_error($dbh->errstr, $dbh->err);
        return;
    }

    $self->commit;
    return $sub_id;
}


#depricated
sub _upsert_param {
    my $self = shift;
    return if $self->check_readonly;
    my %params = @_;

    my $subid = $params{subid};
    my $values = $params{params};
    if (!$subid || !%params) {
        return;
    }

    my $dbh = $self->connect_db;
    my $sql = <<"EOSQL";
INSERT INTO submission_param (subid, name, value) VALUES (?, ?, ?)
  ON DUPLICATE KEY UPDATE value = ?
EOSQL
    my $sth = $dbh->prepare($sql);

    for my $k (keys %$values) {
        my $rs = $sth->execute($subid, $k, $values->{$k}, $values->{$k});
        if (!$rs) {
            # TODO
        }
    }
    $self->disconnect_db;
    return 1;
};

sub update {
    my $self = shift;
    return if $self->check_readonly;
    my $params = {@_};

    my $subid = $params->{subid}
      || $params->{submission_id}
      || $params->{id}
      || $params->{sub_id};
    return if !$subid;

    my $dbh = $self->start_transaction;

    my $total_update = 0;
    if (defined $params->{del}) {
        my $sql = "UPDATE submissions SET del = ? WHERE subid = ?";
        my $rs = $dbh->do($sql, undef, $params->{del}, $subid);
        if (!defined $rs) {
            $self->rollback;
            return;
        }
        $total_update += $rs;
    }

    if (defined $params->{public}) {
        my $sql = "UPDATE firehose SET public = ? WHERE type = 'submission' AND srcid = ?";
        my $rs = $dbh->do($sql, undef, $params->{public}, $subid);
        if (!defined $rs) {
            $self->rollback;
            return;
        }
        $total_update += $rs;
    }

    $self->commit;
    return 1;
}

sub create_draft_text {
    my ($self, $submission) = @_;
    # TODO
    my $author = $submission->{author};
    my $users = $self->new_instance_of("Users");
    if ($users->is_anonymous_uid($submission->{uid})) {
        $author = "ある" . $author;
    }
    my $text = $submission->{story};
    my $draft = "<p>${author}曰く、</p><blockquote\n>$text\n</blockquote>";
    return $draft;
}

# delete story from database
# this method is for test purpose only.
sub hard_delete {
    my $self = shift;
    return if $self->check_readonly;
    my $params = {@_};

    my $subid = $params->{submission_id} || $params->{id};
    return if !$subid;

    my $dbh = $self->start_transaction;

    # delete submissions
    my $sql = "DELETE FROM submissions WHERE subid = ?";
    my $rs = $dbh->do($sql, undef, $subid);
    if (!defined $rs || $rs == 0) {
        Mojo::Log->new->warn("DELETE FROM submissions failed. stoid is $subid.");
    }

    # delete submission_param
    $sql = "DELETE FROM submission_param WHERE subid = ?";
     $rs = $dbh->do($sql, undef, $subid);
    if (!defined $rs || $rs == 0) {
        Mojo::Log->new->warn("DELETE FROM submission_param failed. stoid is $subid.");
    }

    # delete from firehose
    my $firehose = $self->new_instance_of("Firehose");
    $firehose->hard_delete("submission", $subid);

    $self->commit;
    return 1;
}

1;
