package Newslash::Model::Users;
use Newslash::Model::Base -base;

use Newslash::Model::Users::Friends;
use Newslash::Model::Users::Util;
use Newslash::Model::Users::ACL2;
use Newslash::Model::Users::Prefs;
use Newslash::Model::Users::Configs;
use Newslash::Model::Users::Achievements;
use Newslash::Model::Users::Passwords;
use Newslash::Model::Users::Info;
use Newslash::Model::Users::ACL;
use Newslash::Model::Users::Comments;
use Newslash::Model::Users::User;
use Newslash::Model::Users::Sidebar;
use Newslash::Model::Users::Messages;
use Newslash::Model::Users::Index;
use Newslash::Model::Users::Class;
use Newslash::Model::Users::Param;
use Newslash::Model::Users::Permissions;

use Newslash::Util::Timezones;

use Email::Valid;
use Data::Dumper;
use Mojo::JSON qw(to_json from_json);

use Encode qw(encode_utf8 decode_utf8);
use Mojo::Log;

# nickname, matchname, and realemail columns doesn't have UNIQUE limitation,
# but practically they are unique
sub key_definition {
    return {
            table => "users",
            primary => "uid",
            unique => [qw(nickname realemail matchname)],
            datetime => [qw(newpasswd_ts journal_last_entry_date)],
            other => [qw(fakeemail homepage
                         passwd sig seclev newpasswd
                         author shill_id)],
            aliases => { user_id => "uid",
                         id => "uid",
                       }
           };
}

sub on_start_up {
    my $self = shift;

    # create tables
    return 1 if $self->check_readonly;

    # sidebar
    $self->sidebar->on_start_up;

    # create `ns_users_config` table
    $self->configs->_check_and_create_table;

    # check anonymous user exists
    my $anon = $self->anonymous_user;
    if (!$anon) {
        $self->hard_delete(1);
        $self->create("Anonymous Coward", 'AnonymousCoward@example.com', {uid => 1});
        $self->update(uid => 1, seclev => 0);
        $self->info->update(uid => 1,
                            realname => "名無しのゴンベエ",
                            karma => -25);
        $self->warn("create anonymous_user");
    }

    # convert AC (uid==1) user's data
    my $converted = $self->prefs->convert_old_prefs(1);
    if ($converted) {
        my $rs = $self->configs->update(id => 1, config => $converted);
        if (!$rs) {
            $self->warn("update config failed. uid: 1");
        }
    }
    return 1;
}

##### sub models
sub friends      { return shift->_get_model('::Users::Friends'); }
sub prefs        { return shift->_get_model('::Users::Prefs'); }
sub configs      { return shift->_get_model('::Users::Configs'); }
sub achievements { return shift->_get_model('::Users::Achievements'); }
sub passwords    { return shift->_get_model('::Users::Passwords'); }
sub info         { return shift->_get_model('::Users::Info'); }
sub acl          { return shift->_get_model('::Users::ACL'); }
sub acl2         { return shift->_get_model('::Users::ACL2'); }
sub comments     { return shift->_get_model('::Users::Comments'); }
sub sidebar      { return shift->_get_model('::Users::Sidebar'); }
sub messages     { return shift->_get_model('::Users::Messages'); }
sub index        { return shift->_get_model('::Users::Index'); }
sub class        { return shift->_get_model('::Users::Class'); }
sub param        { return shift->_get_model('::Users::Param'); }
sub permissions  { return shift->_get_model('::Users::Permissions'); }

sub _get_model {
    my ($self, $class) = @_;
    $self->{sub_models} ||= {};
    if (!$self->{sub_models}->{$class}) {
        $self->{sub_models}->{$class} = $self->new_instance_of($class);
    }
    else {
        if ($self->transaction_mode && !$self->{sub_models}->{$class}->transaction_mode) {
            $self->{sub_models}->{$class}->use_transaction($self->{_tr_dbh});
        }
    }
    return $self->{sub_models}->{$class};
}

##### utility functions

# auth related
sub authentification {
    return shift->passwords->authentification(@_);
}

sub is_anonymous_uid {
    my ($self, $uid) = @_;
    return ($uid == 1);
}

sub anonymous_user {
    my $self = shift;
    return  $self->select(uid => 1);
}

#========================================================================

=head2 nickname_to_matchname

Convert nickname to matchname.

nickname consists of almost all alphabet, number, symbol.

matchname consists of lower alphabet and number.

=cut

sub nickname_to_matchname {
    my $self = shift;
    my $nick = shift;
    $nick = lc $nick;
    $nick =~ s/[^a-zA-Z0-9]//g;
    return $nick;
}

sub update_karma {
    my $self = shift;
    my $params = {@_};

    my $uid = $params->{uid};
    return if !$uid;

    my $add = $params->{add};
    my $karma => $params->{karma};
    return if (!defined $karma && !defined $add);

    my $rs = 0;
    if (defined $karma) {
        $rs = $self->info->update(uid => $uid, karma => $karma);
        return if !defined $rs;
    }

    if (defined $add) {
        my $rs1 = $self->info->update(uid => $uid,
                                      karma => {add => $add});
        return if !defined $rs1;
        $rs += $rs1;
    }

    return $rs;
}

##### select/create/update/delete
sub select {
    my $self = shift;
    my $params = {@_};
    my $rs = $self->generic_select(params => $params);
    return if (!defined $rs);
    return $self->_create_user_object($rs);
}

sub get_user_config {
    my ($self, $users) = @_;
    my $target = (ref($users) eq "ARRAY") ? $users : [$users,];

    for my $u (@$target) {
        $u->{config} = $self->configs->select(id => $u->{uid});
        $u->{configJSON} = to_json($u->{config});
    }
}

sub _create_user_object {
    my ($self, $user) = @_;

    my $_fn = sub {
        my $user = shift;
        my $result = {%$user};
        #for my $k (qw{uid nickname realemail fakeemail homepage sig seclev matchname author}) {
        #    $result->{$k} = $user->{$k};
        #}
        # if seclev is greater than 10000, the user is admin
        # if seclev is 0, the user is Anonymous user
        $result->{is_admin} = $result->{seclev} >= 10000 ? 1 : 0;
        $result->{is_login} = $result->{seclev} != 0 ? 1 : 0;

        #$result->{config} = $self->configs->select(id => $user->{uid});
        #$result->{configJSON} = to_json($result->{config});

        return $result;
    };

    if (ref($user) eq 'ARRAY') {
        my @ret = map { $_fn->($_) } @$user;
        return \@ret;
    }
    else {
        return $_fn->($user);
    }
}

sub update {
    my $self = shift;
    my $params = {@_};

    # check password
    if ($params->{passwd}) {
        $params->{passwd} = $self->passwords->encrypt_password($params->{passwd});
    }

    return $self->generic_update(params => $params);
}

sub hard_delete {
    my ($self, $uid) = @_;
    return if $self->check_readonly;
    my $error = 0;
    my $dbh = $self->start_transaction;

    for my $table (qw(users users_info users_prefs
                      users_comments users_hits users_class
                      users_index users_param)) {
        my $sql = "DELETE FROM $table WHERE uid = ?";
        my $rs = $dbh->do($sql, undef, $uid);
        if (!defined $rs || $rs == 0 && $table eq "users_param") {
            Mojo::Log->new->warn("DELETE FROM $table failed. uid is $uid.");
            $error = 1;
        }
    }

    $self->commit;
    return !$error;
}

#========================================================================

=head2 create

Create new user.

=over 4

=item Return value

HASH

=back

=cut

sub create {
    my ($self, $nickname, $email, $passwd, $opts) = @_;
    return if $self->check_readonly;
    $opts ||= {};

    # check: nickname is given?
    if ($nickname =~ m/\A\s+\z/) {
        $self->set_error("INVALID_ID");
        return;
    }

    my $matchname = $self->nickname_to_matchname($nickname);

    # check: email is valid?
    if (!Email::Valid->address($email)) {
        $self->set_error("INVALID_EMAIL");
        return;
    }

    # check: nickname already exists?
    my $user = $self->select(nickname => $nickname);
    if ($user) {
        $self->set_error("ID_EXISTS");
        return;
    }

    # check: mail already registered?
    $user = $self->select(realemail => $email);
    if ($user) {
        $self->set_error("EMAIL_EXISTS");
        return;
    }

    my $dbh = $self->start_transaction;
    my $sql;
    my @params;
    my $sth;

    # create user and set temporary password
    if (!$passwd) {
        $passwd = $self->passwords->generate_random_password;
    }
    my $enc_passwd = $self->passwords->encrypt_password($passwd);
    my $newpasswd = $opts->{password_token};

    # build SQL and parameters
    if ($opts->{password_token}) {
        $sql = <<"EOSQL";
INSERT INTO users (uid, realemail, nickname, matchname,  seclev,  passwd, newpasswd, newpasswd_ts)
  VALUES          (?,   ?,         ?,        ?,          ?,       ?,      ?,         NOW())
EOSQL
        @params = ($opts->{uid},
                   $email,
                   $nickname,
                   $matchname,
                   $opts->{seclev} // 1,
                   $enc_passwd,
                   $opts->{password_token},
                  );
    }
    else {
        $sql = <<"EOSQL";
INSERT INTO users (uid, realemail, nickname, matchname,  seclev,  passwd)
  VALUES          (?,   ?,         ?,        ?,          ?,       ?)
EOSQL
        @params = ($opts->{uid}, $email, $nickname, $matchname, 1, $enc_passwd);
    }

    $dbh->do($sql, undef, @params);
    my $uid = $dbh->last_insert_id(undef, undef, undef, undef);

    if (!$uid) {
        $self->rollback;
        $self->set_error($dbh->errstr, $dbh->err);
        return;
    }

    # create users_info, etc.
    my $result;
    $sql = <<"EOSQL";
INSERT INTO users_info (uid, lastaccess, created_at, bio)
  VALUES               (?,   NOW(),      NOW(),      ?)
EOSQL
    $result = $dbh->do($sql, undef, $uid, '');
    if (!$result) {
        $self->rollback;
        $self->set_error($dbh->errstr, $dbh->err);
        return;
    }

    for my $table (qw{users_prefs users_comments users_hits users_class}) {
        $sql = "INSERT INTO $table (uid) VALUES (?)";
        $result = $dbh->do($sql, undef, $uid);
        if (!$result) {
            $self->rollback;
            $self->set_error($dbh->errstr, $dbh->err);
            return;
        }
    }

    $sql = <<"EOSQL";
INSERT INTO users_index (uid, story_never_topic, slashboxes, story_always_topic)
  VALUES                (?,   ?,                 ?,          ?)
EOSQL
    $result = $dbh->do($sql, undef, $uid, '', '', '');
    if (!$result) {
        $self->rollback;
        $self->set_error($dbh->errstr, $dbh->err);
        return;
    }

    # finish
    $self->commit;
    return $uid;
}


1;
