package Newslash::Model::Users::Passwords;
use Newslash::Model::Base -base;

use Digest::MD5 qw(md5_hex md5_base64);
use Crypt::Eksblowfish::Bcrypt qw(bcrypt en_base64 de_base64);

#========================================================================

=head2 authentification

  my $result = $users->passwords->authentification($nickname, $password);

Return user hash when nickname and password pair is correct.

=cut

sub authentification {
    my ($self, $nickname, $passwd) = @_;
    my $users = $self->new_instance_of("Users");

    my $u = $users->select(nickname => $nickname);
    return if !$u;

    if ($self->compare_password($passwd, $u->{passwd})
        || $self->_compare_password_compat($passwd, $u->{passwd}, $users->{uid}, 0, 1)) {
        return $u;
    }
    return;
}

sub _compare_password_compat {
    # depracated. exists for compatibility.
    my ($self, $passwd, $hashed, $uid, $is_plain, $is_enc) = @_;

    if ($is_plain) {
        return 1 if $passwd eq $hashed;
    }

    if (!$is_enc) {
        return 1 if md5_hex($passwd) eq $hashed;

        my @salts = ('',);
        for my $salt (@salts) {
            return 1 if md5_hex("$salt:$uid:$passwd") eq $hashed;
            return 1 if length($salt) && md5_hex("$salt$passwd") eq $hashed;
        }
    }
    return;
}

#========================================================================

=head2 nickname_to_matchname

  my $matchname = $users->passwords->nickname_to_matchname($nickname);

Return matchname converted from nickname.

'nickname' consists of almost all alphabet, number, symbol.
'matchname'consists of only lower alphabet and number.

=cut

sub nickname_to_matchname {
    my ($self, $nick) = @_;
    $nick = lc $nick;
    $nick =~ s/[^a-zA-Z0-9]//g;
    return $nick;
}

#========================================================================

=head2 random_password

  my $passwd = $users->passwords->random_password;

Return new random 8-character password composed of 0..9, A..Z, a..z
(but not including possibly hard-to-read characters [0O1Iil]).

=cut

sub generate_random_password {
    my $self = shift;
    my @chars = grep !/[0O1Iil]/, 0..9, 'A'..'Z', 'a'..'z';
    return join '', map { $chars[rand @chars] } 0 .. 7;
}

#========================================================================

=head2 compare_password

  my $result = $users->passwords->compare_password($password, $hashed_password);

Return 1 when hashed password correspod to password.

=cut

sub compare_password {
    my ($self, $passwd, $hashed) = @_;

    # check password hash with Bcrypt
    if ($hashed =~ m/\$2a?\$\d{2}\$.+/) {
        return 1 if $hashed eq bcrypt($passwd, $hashed);
    }
    return;
}

sub verify_password {
    my $self = shift;
    my $args = {@_};
    my $users = $self->new_instance_of("Users");
    my $user = $args->{user};

    if ($args->{uid}) {
        $user = $users->select(uid => $args->{uid});
    }
    elsif ($args->{nickname}) {
        $user = $users->select(nickname => $args->{nickname});
    }

    my $hashed = $user ? $user->{passwd} : $args->{hashed};
    if (!$hashed) {
        $self->last_error("INVALID_USER");
        return;
    }

    my $password = $args->{password};
    if (!$password) {
        $self->last_error("NO_PASSWORD_GIVEN");
        return;
    }

    my $rs = $self->compare_password($password, $hashed);
    return 1 if $rs;

    if ($user && $self->_compare_password_compat($password,
                                                 $hashed,
                                                 $users->{uid}, 0, 1)) {
        return 1;
    }

    $self->last_error("INVALID_PASSWORD");
    return;
}

#========================================================================

=head2 encrypt_password

  my $hashed_password = $users->passwords->encrypt_password($passwd);

Return encrypted (hashed) password.

=cut

sub encrypt_password {
    my ($self, $passwd) = @_;

    # newslash's default hash algorythm is bcrypt
    my $salt = substr(en_base64(md5_hex(rand())), 0, 16);
    my $setting = '$2a$10$' . en_base64($salt);
    my $hashed = bcrypt($passwd, $setting);

    return $hashed;
}



1;
