package Newslash::Plugin::AddPageProperty;
use Mojo::Base 'Mojolicious::Plugin';
use Data::Dumper;
use List::Util qw(any);

use constant CANONICAL_TABLE => {
                                 qr{^/~([^/]*)/journal/?} => '/~$1/journals',
                                };
use constant EXCLUDE_PATH => [ qr{^/api/}, ];

sub register {
    my ($self, $app, $conf) = @_;

    $app->hook(before_render => sub {
                   my ($c, $args) = @_;
                   my $path = $c->url_for->to_abs->path;
                   return if any { $path =~ $_ } @{EXCLUDE_PATH()};

                   my $page = $args->{page} || $c->stash("page") || {};

                   set_properties($c, $args, $page);
                   $page->{structured_data} = generate_structured_data($c, $args);
               });
}

sub set_properties {
    my ($c, $args, $page) = @_;
    my $cfg = $c->app->config->{Site} || {};
    my $base_url = $cfg->{base_url};
    my $icon_baseurl = $base_url . $cfg->{topic_icon_base_url} . "/";

    # retrieve values from stash/argument
    for my $key (qw{description thumbnail canonical
                    author create_time update_time title}) {
        $page->{$key} = $args->{$key} || $c->stash($key);
    }

    my $item = $c->stash("item") || $args->{item};
    if ($item) {
        # set title
        if (!$page->{title} && $item->{title}) {
            $page->{title} = $c->strip_html_tag($item->{title});
        }

        # set description
        if (!$page->{description} && $item->{intro_text}) {
            $page->{description} = $c->extract_description($item->{intro_text});
        }

        # set thumbnail
        if (!$page->{thumbnail} && ref($item->{primary_topic}) eq "HASH") {
            if ($item->{primary_topic}->{image}) {
                $page->{thumbnail} = $icon_baseurl . $item->{primary_topic}->{image};
            }
        }
        # set author
        if (!$page->{author} && $item->{uid} || $item->{user_id}) {
            my $u = $c->model('users')->select(uid => $item->{uid});
            $page->{author} = $u if $u;
        }

        # set create time
        if (!$page->{create_time} && $item->{create_time}) {
            $page->{create_time} = $item->{create_time};
        }

        # set update time
        if (!$page->{update_time} && $item->{update_time}) {
            $page->{update_time} = $item->{update_time};
        }

    }

    # set canonical url
    if (!$page->{canonical}) {
        my $base_url = $cfg->{base_url};
        my $url = Mojo::URL->new($c->url_for);
        my $path = get_canonical_path($url->path);

        if ($base_url) {
            $url = Mojo::URL->new($base_url);
            $url->path($path);
            $page->{canonical} = $url->to_string;
        } else {
            $page->{canonical} = $url->to_string;
        }
    }
    # fallback
    if (!$page->{thumbnail} && $cfg->{default_thumbnail}) {
        $page->{thumbnail} = $base_url . $cfg->{default_thumbnail};
    }
}

sub generate_structured_data {
    my ($c, $args) = @_;
    my $cfg = $c->app->config->{Site} || {};
    my $page = $args->{page} || $c->stash("page");

    my $data = { '@context' => "http://schema.org",
               };

    # set content type
    if (!$page) {
        $c->app->log->debug("AddPageProperty: no page property.");
        return;
    }

    if (!$page->{type}) {
        $c->app->log->debug("AddPageProperty: no page type given. page: " . Dumper($page));
        return;
    }

    if ($page->{type} eq "single") {
        # single content page
        if ($page->{content_type} eq "story") {
            $data->{'@type'} = "NewsArticle";
        }
        elsif ($page->{content_type} eq "journal") {
            $data->{'@type'} = "BlogPosting";
        }
    }
    elsif ($page->{type} eq "user") {
        if ($page->{content_type} eq "recent") {
            $data->{'@type'} = "Person";
            my $the_user = $args->{the_user} || $c->stash("the_user");
            if ($the_user) {
                $data->{name} = $the_user->{nickname};
                my $info = $the_user->{config}->{info};
                $data->{alternateName} = $info->{realname} if $info->{realname};
                $data->{url} = $info->{url} if $info->{url};
                $data->{description} = $c->strip_html_tag($info->{biography}) if $info->{biography};
                return $data;
            }
        }
        return;
    }
    elsif ($page->{type} eq "timeline") {
        $data->{'@type'} = "WebSite";
        $data->{name} = $cfg->{name};
        $data->{alternateName} = $cfg->{alternate_name} if $cfg->{alternate_name};
        $data->{description} = $cfg->{description};
        $data->{url} = $cfg->{base_url};
        $data->{potentialAction} = { '@type' => "SearchAction",
                                     target => "$cfg->{base_url}/search?q={search_term_string}",
                                     "query-input" => "required name=search_term_string",
                                   };
        $data->{sameAs} = $cfg->{same_as} if $cfg->{same_as};
    }
    else {
        # no generate structured data
        return;
    }

    $data->{headline} = $page->{title} if $page->{title};
    $data->{datePublished} = $c->format_timestamp(mysql => $page->{create_time},
                                                  format => 'iso8601') if $page->{create_time};
    $data->{dateModified} = $c->format_timestamp(mysql => $page->{update_time},
                                                  format => 'iso8601') if $page->{update_time};
    $data->{articleBody} = $page->{description} if $page->{description};
    $data->{image} = $page->{thumbnail} if $page->{thumbnail};

    $data->{mainEntityOfPage} = { '@type' => "WebPage",
                                  '@id' => $page->{canonical} };

    # add author data
    if ($page->{author}) {
        my $u = $page->{author};
        my $i = $u->{config}->{info};
        $data->{author} = { '@type' => "Person",
                            name => $u->{nickname} };
        $data->{author}->{alternateName} = $i->{realname} if $i->{realname};
        $data->{author}->{url} = $i->{homepage} if $i->{homepage};
    }

    # add publisher data
    $data->{publisher} = { '@type' => "Organization",
                           name => $cfg->{publisher},
                           url => $cfg->{publisher_url},
                           logo => { '@type' => "ImageObject",
                                     url => $cfg->{publisher_logo},
                                   },
                         };

    return $data;
}

sub get_canonical_path {
    my ($path) = @_;
    for my $rex (keys %{CANONICAL_TABLE()}) {
        my $to_path = CANONICAL_TABLE->{$rex};
        if ($path =~ $rex) {
            my $rep = $1;
            $path = $to_path;
            $path =~ s/\$1/$rep/;
        }
    }
    if (substr($path, -1, 1) ne "/") {
        $path = "$path/";
    }
    return $path;
}




1;
