package Newslash::Plugin::Preprocessor;
use Mojo::Base 'Mojolicious::Plugin';

use File::Spec;
use File::Basename qw(fileparse);
use Digest::MD5 qw(md5_hex);
use IPC::Open3;
use Symbol qw(gensym);
use IO::Select;
use POSIX ":sys_wait_h";
use Mojo::Util qw(dumper);
use Encode;
use IO::Compress::Gzip qw(gzip);
use Data::Dumper;

sub _local_store {
    my $self = shift;
    $self->{_local_store} //= {};
    return $self->{_local_store};
};

use constant CACHE_KEY => "Preprocessor";

sub _get_absolute_path {
    my ($self, $rel_path) = @_;
    my $base_dir = $self->{conf}->{base_directory};
    if (!$base_dir) {
        $base_dir = $self->{app}->home;
    }
    if (!$base_dir) { # fallback
        $base_dir = "./";
    }
    my $path = File::Spec->catdir($base_dir, $rel_path);
    $path = File::Spec->rel2abs($path);
    return $path;
}

sub _strip_md5_from_path {
    my ($self, $pathname) = @_;
    if ($pathname =~ m/^(.+)_[0-9a-f]+(\.\w+)$/) {
        return $1 . $2;
    }
    return $pathname;
}

sub _add_md5_to_path {
    my ($self, $pathname, $md5) = @_;
    my ($base, $dir, $ext) = fileparse($pathname, qw/\.[^.]*/);
    return "${dir}${base}_$md5$ext";
}

sub _generate_content {
    my ($self, $pathname) = @_;
    $pathname = $self->_strip_md5_from_path($pathname);

    my $targets = $self->{conf}->{targets} || {};
    my $params = $targets->{$pathname};

    if (!$params) {
        $self->{app}->log->error("Preprocessor: cannot generate '$pathname' - target not defined");
        return;
    }

    my $gen_conf = $self->{conf}->{$params->{type}};
    if (!defined $gen_conf) {
        $self->{app}->log->error("Preprocessor: cannot generate '$pathname' - generate config not found for '$params->{type}'");
        return;
    }

    my $cmd = $gen_conf->{command};
    if (!$cmd) {
        $self->{app}->log->error("Preprocessor: cannot generate '$pathname' - command '$gen_conf->{command}' not found");
        return;
    }

    my $content = '';
    if ($params->{source}) {
        my $abs_pathname = $self->_get_absolute_path($params->{source});
        if (! -f $abs_pathname) {
            $self->{app}->log->error("Preprocessor: cannot generate '$pathname' - '$abs_pathname' not found");
            return;
        }
        my $fh = FileHandle->new;
        if ($fh->open($abs_pathname, "<:utf8")) {
            $content = do { local $/; <$fh> };
            $fh->close;
        }
    }

    my $opts = $self->_build_options($gen_conf->{options});
    my $content_body = $self->_execute_cmd($cmd, $content, $opts);

    if ($self->{last_execute_status}) {
        # error occured
        return;
    }
    return $self->_build_content_object($content_body, $pathname, $gen_conf->{content_type});
}

sub _build_content_object {
    my ($self, $content_body, $pathname, $content_type) = @_;

    my $content_body_bin;
    if (Encode::is_utf8($content_body)) {
        $content_body_bin = Encode::encode_utf8($content_body);
    }
    else {
        $content_body_bin = $content_body;
        # $content_body = Encode::decode_utf8($content_body);
    }

    my $rs = { md5 => md5_hex($content_body_bin),
               content => $content_body,
               path => $pathname,
               type => $content_type };

    if ($self->{compress}) {
        $rs->{gz_content} = "";
        if (!gzip(\$content_body_bin, \$rs->{gz_content})) {
            delete $rs->{gz_content};
            $self->{app}->log->error("Preprocessor::_build_content_object: gzip failed");
        }
    }

    return $rs;
}

sub _cache_content {
    my ($self, $content) = @_;

    my $md5_url = $self->_add_md5_to_path($content->{path}, $content->{md5});

    if ($self->{use_local_store}) {
        $self->_local_store->{CACHE_KEY . $content->{path}} = $content;
        $self->_local_store->{CACHE_KEY . $md5_url} = $content;
    }
    else {
        # encode content body
        my $encoded_content = { %$content };
        if (Encode::is_utf8($encoded_content->{content})) {
            $encoded_content->{content} = Encode::encode_utf8($encoded_content->{content});
        }

        $self->{app}->ccache->hash_set(CACHE_KEY, $content->{path}, $encoded_content);
        $self->{app}->ccache->hash_set(CACHE_KEY, $md5_url, $encoded_content);
    }
}

sub expire_cache {
    my ($self, $key, $target) = @_;
    if ($self->{use_local_store}) {
        del $self->_local_store->{CACHE_KEY . $key};
    }
    else {
        $self->{app}->ccache->del(CACHE_KEY, $key, $target);
    }
}

sub _hash_to_options {
    my ($self, $hash) = @_;
    my $opts;

    for my $name (keys %$hash) {
        my $value = $hash->{$name};
        if ($value =~ m/^\.\//) {
            $value = $self->_get_absolute_path($value);
        }
        my $opt = $name;
        if ($value) {
            if ($name =~ m/^--/) {
                $opt = "$name=$value";
            }
            else {
                $opt = "$name $value";
            }
        }
        push @$opts, $opt;
    }
    return $opts;
}

sub _build_options {
    my ($self, $options) = @_;
    return [] if !$options;

    my $opts = [];
    if (ref($options) eq "HASH") {
        my $rs = $self->_hash_to_options($options);
        push @$opts, @$rs;
        return $opts;
    }
    if (ref($options) eq "ARRAY") {
        for my $item (@$options) {
            if (!ref($item)) {
		if ($item =~ m/^([^=]+)=(.*)$/) {
		    my $h = {$1 => $2};
		    my $rs = $self->_hash_to_options($h);
		    push @$opts, @$rs;
		}
		else {
		    push @$opts, $item;
		}
            }
            elsif (ref($item) eq "HASH") {
                my $rs = $self->_hash_to_options($item);
                push @$opts, @$rs;
            }
        }
    }
    return $opts;
}

sub generate_all {
    my $self = shift;
    my $targets = $self->{conf}->{targets} || {};
    my $err = 0;
    my @err_msgs;
    for my $target (keys %$targets) {
        my $content = $self->_generate_content($target);
        if (!$content) {
            $err = 1;
            push @err_msgs, $self->{last_error_msg};
            push @err_msgs, $self->{last_result};
        }
        $self->_cache_content($content);
    }
    if ($err) {
        $self->{last_error_msg} = join("\n", @err_msgs);
    }
    return !$err;
}

sub rebuild {
    my $self = shift;
    return $self->generate_all;
}

sub get_content {
    my ($self, $path, $options) = @_;
    $options ||= {};
    my $content;

    if (!$options->{reload}) {
        if ($self->{use_local_store}) {
            $content = $self->_local_store->{CACHE_KEY . $path};
        }
        else {
            $content = $self->{app}->ccache->hash_get(CACHE_KEY, $path);
            if ($content) {
                # decode content body
                $content->{content} = Encode::decode_utf8($content->{content});
            }
        }
    }

    if (!$content) {
        # regenerate contents
        $content = $self->_generate_content($path);
        if ($content) {
            $self->_cache_content($content);
        }
    }
    return $content;
}

sub register {
    my ($self, $app, $conf) = @_;

    # set default config value
    my $cnf = $app->config->{Preprocessor} ||= {};
    $cnf->{base_url} ||= "/css";
    $cnf->{targets} ||= { "css/newslash.css" => { source => "css/newslash.less",
						type => "less" },
			};
    $cnf->{less} ||= { command => "/usr/bin/nodejs",
		       "content_type" => "text/css",
		       options => [ "--no-deprecation",
				    "/usr/bin/lessc",
				    "--include-path=./css",
				    "--no-color",
				    "-",
				  ],
		     };
    $cnf->{use_compression} ||= 0;
    $cnf->{build_mode} //= "cached";

    $self->{conf} = {%$conf, %$cnf};
    $self->{app} = $app;
    $self->{last_error_msg} = "";
    $self->{last_execute_status} = 0;
    $self->{compress} = $cnf->{use_compression};

    if ($self->{conf}->{build_mode}) {
        $self->{use_local_store} = $self->{conf}->{build_mode} eq "start_up";
    }

    my $r = $app->routes;
    my $base_url = $self->{conf}->{base_url};
    if (substr($base_url, -1, 1) ne "/") {
        $base_url = $base_url . "/";
    }

    $r->any("${base_url}*content_path" => sub {
                my $c = shift;
                my $path = $base_url . $c->stash('content_path');
                if ($path) {
                    my $content = $self->get_content($path);
                    if ($content) {
                        # check Etag
                        if ($c->req->headers->if_none_match
                            && $content->{md5}
                            && $c->req->headers->if_none_match eq "\"$content->{md5}\"") {
                            $c->rendered(304);
                            return;
                        }
                        $c->res->headers->content_type($content->{type});
                        my $output = $content->{content};

                        # check compressable
                        if ($self->{compress}
                            && $content->{type} =~ m{^(text/|application/json|application/javascript)}
                            && ($c->req->headers->accept_encoding // '') =~ /gzip/i
                            && $content->{gz_content}
                           ) {
                            # use gzip compression
                            # Add header
                            $c->res->headers->append(Vary => 'Accept-Encoding');
                            $c->res->headers->content_encoding('gzip');
                            $output = $content->{gz_content};
                        }

                        # add Etag
                        if ($content->{md5}) {
                            $c->res->headers->etag("\"$content->{md5}\"");
                        }

                        # send result
                        # $c->res->body() needs binary data
                        if (Encode::is_utf8($output)) {
                            $output = Encode::encode_utf8($output);
                        }
                        $c->res->body($output);
                        $c->rendered(200);
                        return;
                    }
                }
                $c->rendered(404);
            });

    $app->helper(preprocessor => sub { state $preprocessor = $self; });
}

sub get_md5_path {
    my ($self, $path) = @_;
    my $content = $self->get_content($path);

    if ($content && $content->{path}) {
        return $self->_add_md5_to_path($content->{path}, $content->{md5});
    }

    $self->{app}->log->error("Preprocessor: '$path' not found.");
    return;
}

sub _execute_cmd {
    my ($self, $command, $content, $args) = @_;
    my ($stdin, $stdout, $stderr);

    if (!defined $content) {
        $stdin = gensym;
    }
    my $args_str = join(" ", @$args);
    $self->{app}->log->debug("Preprocessor: execute $command $args_str");

    my $pid = open3($stdin, $stdout, $stderr, $command, @$args);
    if ($pid =~ /^open3:/) {
        $self->{app}->log->error("Preprocessor: cannot execute $command");
        return;
    }

    my $sel = IO::Select->new($stdout, $stderr);
    my $result = '';
    my $err_msg = '';

    if (defined $content) {
        print $stdin $content;
        close $stdin;
    }

    while (1) {
        while (my @ready = $sel->can_read) {
            foreach my $fh (@ready) {
                if ($fh == $stdout) {
                    if (sysread($fh, my $buf_o, 1024) > 0) {
                        $result .= $buf_o;
                    }
                    else {
                        $sel->remove($fh);
                        next;
                    }
                }
                elsif ($fh == $stderr) {
                    if (sysread($fh, my $buf_e, 1024) > 0) {
                        $err_msg .= $buf_e;
                    }
                    else {
                        $sel->remove($fh);
                        next;
                    }
                }
            }
        }
        last if (waitpid($pid, WNOHANG) > 0);
    }

    my $status = WIFEXITED($?) ? WEXITSTATUS($?) : WTERMSIG($?);
    $self->{last_error_msg} = $err_msg;
    $self->{last_execute_status} = $status;
    $self->{last_result} = $result;
    if ($err_msg) {
        $self->{app}->log->error("Preprocessor: command $command execute failed - $err_msg");
    }
    if ($status != 0) {
        $self->{app}->log->error("Preprocessor: command $command returns non-zero code.");
        if ($result) {
            $self->{app}->log->error($result);
        }
        if ($err_msg) {
            $self->{app}->log->error($err_msg);
        }
        return;
    }
    # decode result to Perl's internal format
    $result = Encode::decode('utf8', $result);

    return wantarray ? ($result, 1) : $result;
}



1;
