package Newslash::Plugin::ReCaptcha;
use Mojo::Base 'Mojolicious::Plugin';
use Mojo::Util qw(dumper);
use Mojo::UserAgent;
use Mojo::JSON qw(decode_json);

use Compress::Zlib;

use constant RECAPTCHA_URL => 'https://www.google.com/recaptcha/api/siteverify';
use constant RECAPTCHA_SCRIPT => '<script src="https://www.google.com/recaptcha/api.js"></script>';
use constant RECAPTCHA_CODE => <<'EOC';
<div class="g-recaptcha"
     data-badge="inline"
     data-sitekey="%s"
     data-callback="recaptchaDone"
     data-size="invisible">
</div>
EOC

sub register {
    my ($self, $app, $conf) = @_;

    # set default config value
    my $cnf = $app->config->{ReCaptcha} ||= {};
    $cnf->{site_key} ||= "";
    $cnf->{secret_key} ||= "";
    $cnf->{enable} = 0 if !$cnf->{enable};

    $app->helper(insert_recaptcha => sub {
                     my $c = shift;
                     my $mesg = "<!-- captcha disabled -->";
                     return $mesg if !$c->stash('use_captcha');

                     my $cf = $c->config->{ReCaptcha} || {};
                     return $mesg if !$cf->{enable};

                     my $user = $c->stash('user') || {};
                     return $mesg if $user->{is_login};

                     if (!$c->stash('load_recaptcha')) {
                         $c->insert_code(begin_footer => RECAPTCHA_SCRIPT);
                         $c->stash(load_recaptcha => 1);
                     }

                     return sprintf(RECAPTCHA_CODE, $cf->{site_key});
                 });

    $app->hook(around_action => sub {
                   my ($next, $c, $action, $last) = @_;
                   return $next->() if $c->req->method ne 'POST';
                   return $next->() if !$c->stash('captcha_check');
                   return $next->() if $c->req->json('/action') eq 'preview';

                   my $config = $c->config->{ReCaptcha} || {};
                   return $next->() if !$config->{enable};

                   my $user = $c->stash('user');
                   return $next->() if $user->{is_login};

                   my $token = $c->req->json->{recaptcha_token};
                   if (!$token) {
                       on_auth_failed($c, $token);
                       return;
                   }

                   # validate token
                   my $ua  = Mojo::UserAgent->new;
                   my $param = {
                                secret => $c->app->config->{ReCaptcha}->{secret_key},
                                response => $token,
                                #remoteip => foo,
                               };
                   my $url = RECAPTCHA_URL;

                   my $tx = $ua->post($url => form => $param);
                   if ($tx->error) {
                       on_server_error($c, $token);
                       return;
                   }

                   if ($tx->res->code == 200) {
                       my $encoding = $tx->res->headers->header('Content-Encoding');
                       my $resp_body = $tx->res->body;
                       my $json = decode_json($resp_body);
                       if ($encoding && $encoding =~ /gzip/) {
                           # response is gziped
                           $c->app->log->debug("recaptcha response is gziped: $resp_body");
                           if (!$json) {
                               $resp_body = Compress::Zlib::memGunzip($resp_body);
                               $json = decode_json($resp_body);
                           }
                       }
                       my $success = $json->{success} if $json;
                       if($success) {
                           return $next->();
                       }
                   }
                   on_auth_failed($c, $token);
                   return;

                   # $ua->post($url => form => $param
                   #           => sub {
                   #               my ($ua, $tx) = @_;
                   #               warn(dumper($tx));
                   #               if ($tx->error) {
                   #                   on_server_error($c, $token);
                   #                   return;
                   #               }
                   #               if ($tx->res->code == 200) {
                   #                   my $encoding = $tx->res->headers->header('Content-Encoding');
                   #                   my $resp_body = $tx->res->body;
                   #                   warn($resp_body);
                   #                   #if ($encoding && $encoding =~ /gzip/) {
                   #                   #    # response is gziped
                   #                   #    $c->app->log->debug("responce is gziped");
                   #                   #    $resp_body = Compress::Zlib::memGunzip($resp_body);
                   #                   #}
                   #                   my $json = decode_json($resp_body);
                   #                   my $success = $json->{success} if $json;
                   #                   if ($success) {
                   #                       return $next->();
                   #                   }
                   #               }
                   #               on_auth_failed($c, $token);
                   #               return;
                   #           });
               });

}

sub on_auth_failed {
    my ($c, $token) = @_;
    # validation error
    my $url = $c->url_for();
    $c->app->log->debug("invalid reCaptcha_token: $token for: $url");
    $c->render(json => { error => 1, reason => "invalid_recaptcha_token", message => "ReCaptcha: validation error" });
    $c->rendered(400);
    return;
}

sub on_server_error {
    my ($c, $token) = @_;
    # validation error
    my $url = $c->url_for();
    $c->app->log->debug("reCaptcha server error: $token for $url");
    $c->render(json => { error => 1, reason => "recaptcha_server_error", message => "ReCaptcha: server error" });
    $c->rendered(400);
    return;
}

1;
