package Newslash::Plugin::TT2Renderer;
use Mojo::Base 'Mojolicious::Plugin';

use strict;
use warnings;
use utf8;
use feature ':5.10';

use Template;

use File::Basename;
use File::Spec;
use FileHandle;

use Mojo::JSON qw(to_json from_json);
use Mojo::Util qw(dumper);

#use Data::Dumper;
#use Mojo::Log;

# subclass for customizing template-provider
{
  package Newslash::Plugin::TT2Renderer::Provider;
  use base qw(Template::Provider);

  my $EXT = ["", ".js.tt2", ".html.tt2", ".xml.tt2", ".rss.tt2", ".tt2"];

  sub _find_template {
    my ($self, $fname) = @_;
    foreach my $ext (@$EXT) {
      my $path = $fname . $ext;
      if (-f $path) {
	return $path;
      }
    }
    return undef;
  }

  sub _template_modified {
    my ($self, $fname) = @_;
    my $path = $self->_find_template($fname);
    if (!$path) {
      return undef;
    }

    return (stat($path))[9];
  }

  sub _template_content {
    my ($self, $fname) = @_;
    my $path = $self->_find_template($fname);
    if (!$path) {
      return undef;
    }
    my $fh = FileHandle->new;
    if ($fh->open($path, "r")) {
      my $tmpl = do { local $/; <$fh> };
      $fh->close;
      return $tmpl;
    }
    return undef;
  }
}


# subclass for TT2 plugin
{
  package Newslash::Plugin::TT2Renderer::TT2Plugin;
  use base qw(Template::Plugin);
}

sub define_vmethod {
  my ($self, $type, $name, $code) = @_;
  $self->{tt}->context->define_vmethod($type, $name, $code);
}

sub add_object {
    my ($self, $name, $object) = @_;
    $self->{extra_objects} = {} if !$self->{extra_objects};
    $self->{extra_objects}->{$name} = $object;
}

sub _render {
    my ($self, $template, $vars, $output_ref, $paths, $options) = @_;
    $options ||= {};
    $options->{INCLUDE_PATH} = $paths;
    $self->{tt_provider}->include_path($paths);

    return $self->{tt}->process($template, $vars, $output_ref, $options);
}

sub render_content {
    my ($self, $template, $vars, $output_ref, $paths, $options) = @_;
    $options ||= {};
    $options->{INCLUDE_PATH} = $paths;
    $self->{tt_provider}->include_path($paths);

    return $self->{tt}->process($template, $vars, $output_ref, $options);
}

sub register {
  my ($self, $app, $conf) = @_;

  my $tt_options = $conf || {};
  $tt_options->{ENCODING} = 'utf8';
  $self->{tt_provider} = Newslash::Plugin::TT2Renderer::Provider->new($tt_options);
  $self->{app} = $app;

  $tt_options->{LOAD_TEMPLATES} = [$self->{tt_provider},];
  #$tt_options->{STRICT} = 1;
  $self->{tt} = Template->new($tt_options);

  #$app->config->{TT2Renderer} = { 'self' => $self, };
  #$self->{NS_plugin} = Newslash::Plugin::TT2Renderer::TT2Plugin->new;
  $app->helper(tt2renderer => sub { state $tt2renderer = $self; });

  # add .json vmethod
  my $json_encoder = sub {
      my $val = shift;
      return to_json($val);
  };
  $self->define_vmethod('hash', 'json', $json_encoder);
  $self->define_vmethod('list', 'json', $json_encoder);

  # add .rand vmethod
  my $rand_vmethod = sub {
      my $val = shift;
      my $rand = int(rand(@$val));
      return $val->[$rand];
  };
  $self->define_vmethod('list', 'rand', $rand_vmethod);

  $app->renderer->add_handler('tt2' => sub {
				 my ($renderer, $c, $output, $options) = @_;

				 my $inline_template = $options->{inline};
				 my $template_path = $renderer->template_path($options);
				 my $data_template = $renderer->get_data_template($options);
				 my $template_name = $renderer->template_name($options);
				 my $paths = $renderer->paths;
                                 my $user = $c->stash('user');

				 my $template = '';
				 if ($inline_template) {
				   # use string as a template
				   $template = $inline_template;
				 } elsif ($data_template) {
				   # use template in DATA sections
				   $template = $data_template;
				 } else {
				   # read from file
				   $template = $template_name;
				 }

                                 # prepare variables
				 my $vars = $c->stash;

                                 # register helpers
                                 $vars->{helpers} = {};
                                 my $helpers = $renderer->helpers;
                                 for my $helper (keys %$helpers) {
                                     my $f = $helpers->{$helper};
                                     if (ref($f) eq 'CODE') {
                                         $vars->{helpers}->{$helper} = sub { $helpers->{$helper}->($c, @_); };
                                     }
                                     else {
                                         $vars->{helpers}->{$helper} = $helpers->{$helper};
                                     }
                                 }

                                 # Config variables
                                 # deprecated! 
                                 for my $key (keys %{$app->config}) {
                                     $vars->{$key} = $app->config->{$key};
                                 }
                                 # use "config" variable to access configs.
                                 $vars->{config} = $app->config;

                                 # add extra objects
                                 if ($self->{extra_objects}) {
                                     for my $k (keys %{$self->{extra_objects}}) {
                                         $vars->{$k} = $self->{extra_objects}->{$k};
                                     }
                                 }

				 my $process_opts = {};
                                 my $result = $self->_render($template, $vars, $output, $paths, $process_opts);
				 if (!$result) {
                                     my $e = $self->last_error;
                                     my $type = $e->type;
                                     my $info = $e->info;
                                     die "template error: $type - $info";
				 }
                                 return $$output;
			       });
}

sub render {
    my ($self, $template_name, $vars) = @_;
    my $output = "";
    my $home = $self->{app}->home;
    my $paths = [ File::Spec->join($home, 'templates'), ];
    my $process_opts = {};

    my $updated_vars = {};
    for my $k (keys %$vars) {
        $updated_vars->{$k} = $vars->{$k};
    }
    for my $key (keys %{$self->{app}->config}) {
        $updated_vars->{$key} = $self->{app}->config->{$key};
    }

    my $result = $self->_render($template_name, $updated_vars, \$output, $paths, $process_opts);
    if (!$result) {
        $self->{app}->log->debug($self->last_error);
    }
    return $output;
}

sub last_error {
    my $self = shift;
    return $self->{tt}->error;
}


1;
