package Newslash::Plugin::CSSCompile;
use Mojo::Base 'Mojolicious::Plugin';
use Mojo::Home;
use Mojo::Util qw(slurp spurt);

use File::Spec;
use Digest::MD5 qw(md5_hex);
use File::Basename;
use IPC::Open3;
use Symbol qw(gensym);
use IO::Select;
use POSIX ":sys_wait_h";

#use Data::Dumper;
use Mojo::Log;
my $log = Mojo::Log->new;

my $LESSC = '/usr/bin/lessc';

sub register {
  my ($self, $app, $conf) = @_;

  $conf ||= $app->config->CSSCompile;
  $conf->{source_dir} ||= 'css';
  $conf->{dest_dir} ||= 'public/css';
  $conf->{lessc} ||= $LESSC;

  $self->{lessc} = $conf->{lessc};

  for my $target (qw(source_dir dest_dir)) {
    my $dir = $conf->{$target};
    if ($dir =~ m/^\//) {
      $self->{$target} = $dir;
    } else {
      $self->{$target} = $app->home->rel_dir($dir);
    }
  }
  $self->{css_path} = $app->config->{Site}->{css_path};

  my $files = {};
  $self->{files} = $files;

  $self->compile;

  if ($app->config->{TT2Renderer}) {
    my $tt2r = $app->config->{TT2Renderer}->{self};
    $tt2r->add_NS_function('css_path', sub {
			     my ($c, $name) = @_;
			     my $css = $files->{$name};
			     if ($css) {
			       return $css;
			     }
			     return '';
			   });
  }
  $app->config->{CSSCompile} = { 'self' => $self, };

}

sub _load_css {
  my ($self, $pathname) = @_;

  if ($pathname =~ m/^(.*)\.less$/) {
    my $less = $self->less_compile_file($pathname);
    return $less;
  }

  if ($pathname =~ m/^(.*)\.css$/) {
    my $fh = FileHandle->new;
    if ($fh->open($pathname, "r")) {
      my $css = do { local $/; <$fh> };
      $fh->close;
      return $css;
    }
  }
  return undef;
}

sub compile {
  my ($self) = @_;

  opendir(my $dh, $self->{source_dir});

  while(my $fname = readdir $dh) {
    if ($fname =~ m/^\./) {
      next;
    }
    my $pathname = File::Spec->catfile($self->{source_dir}, $fname);

    if (-d $pathname) {
      next;
    }

    if ($fname =~ m/^(.*)\.less$/) {
      my $basename = $1;
      my $less = $self->less_compile_file($pathname);
      my $hash = md5_hex($less);
      my $dest = File::Spec->catfile($self->{dest_dir}, "${basename}_$hash.css");

      $self->store($less, $dest, $basename);
      next;
    }

    if ($fname =~ m/^(.*)\.css$/) {
      my $basename = $1;
      my $fh = FileHandle->new;
      if ($fh->open($pathname, "r")) {
	my $css = do { local $/; <$fh> };
	$fh->close;

	my $hash = md5_hex($css);
	my $dest = File::Spec->catfile($self->{dest_dir}, "${basename}_$hash.css");
	$self->store($css, $dest, $basename);
      }
    }
  }
}

sub css_url {
  my ($self, $basename) = @_;
  return $self->{files}->{$basename};
}

sub store {
  my ($self, $contents, $dest, $basename) = @_;
  my $fh = FileHandle->new;
  if ($fh->open($dest, "w")) {
    $fh->print($contents);
    $fh->close;
    my $dest_basename = basename($dest);
    $self->{files}->{$basename} = "$self->{css_path}/$dest_basename";
    return 1;
  }
  return 0;
}

sub less_compile_file {
  my ($self, $pathname) = @_;
  my ($stdout, $stderr) = (gensym, gensym);

  $log->debug("CSSCompile: compile $pathname...");

  my $pid = open3(undef, $stdout, $stderr, $self->{lessc}, $pathname);
  if ($pid =~ /^open3:/) {
    $log->debug("CSSCompile: cannot exec $self->{lessc} $pathname.");
    return undef;
  }

  my $sel = IO::Select->new($stdout, $stderr);
  my $result = '';
  my $err_msg = '';

  while (1) {
    while (my @ready = $sel->can_read) {
      foreach my $fh (@ready) {
	if ($fh == $stdout) {
	  if (sysread($fh, my $buf, 1024) > 0) {
	    $result .= $buf;
	  } else {
	    $sel->remove($fh);
	    next;
	  }
	} elsif ($fh == $stderr) {
	  if (sysread($fh, my $buf, 1024) > 0) {
	    $err_msg .= $buf;
	  } else {
	    $sel->remove($fh);
	    next;
	  }
	}
      }
    }

    if (waitpid($pid, WNOHANG) > 0) {
      last;
    }
  }

  my $status = WIFEXITED($?) ? WEXITSTATUS($?) : WTERMSIG($?);
  if ($err_msg) {
    $log->debug("CSSCompile: $LESSC: $err_msg");
  }
  if ($status != 0) {
    $log->debug("CSSCompile: '$LESSC $pathname' returns non-zero code.");
  }

  return $result;
}


1;
