package Newslash::Plugin::ModelCache;
use Mojo::Base 'Mojolicious::Plugin';
use Mojo::Log;
use Mojo::Util qw(md5_sum dumper);
use Storable;

sub register {
    my ($self, $app, $conf) = @_;

    my $kvs = $app->kvs;
    $self->{app} = $app;
    $self->{kvs} = $kvs;
    $self->{inmemory} = $app->cache;

    if (!$self->{app}) {
        Mojo::Log->new->warn('ModelCache: no app given');
    }
    if (!$self->{kvs}) {
        Mojo::Log->new->warn('ModelCache: no kvs given');
    }

    $app->helper(model_cache => sub { state $model_cache = $self; });
}

sub _param_to_hash {
    my $self = shift;
    my @params = @_;

    my $serialized = Storable::freeze([@params]);
    my $sha1sum = md5_sum($serialized);
}



#========================================================================

=head2 select($model_name, $expire, $params)

execute select query for given model with cache.

=item Parameters

=over 4

=item $model_name

model name

=item $expire

cache expire time (seconds)

=item $params

query params to give model's select method

=back

=item Return value

query result

=cut

sub select {
    my $self = shift;
    my $model_name = shift;
    my $expire = shift;
    my @params = @_;

    return if !$model_name;

    my $model = $self->{app}->model($model_name);
    if (!$model) {
        $self->{app}->log->warn("ModelCache: invalid model name: $model_name");
        return;
    }

    my $result;
    if (!$self->{immemory} && !$self->{kvs}) {
        return $model->select(@params);
    }
    my $hashed_key = "$model_name:" . $self->_param_to_hash(@params);

    # at first, check in-memory cache
    if ($self->{inmemory}) {
        $result = $self->{inmemory}->get($hashed_key);
        if ($result) {
            $self->{app}->log->debug("use inmemory cache: $hashed_key");
            return $result;
        }
    }

    # next, check KVS cache
    if ($self->{kvs}) {
        $result = $self->{kvs}->get($hashed_key);
        if ($result) {
            $self->{app}->log->debug("use kvs cache: $hashed_key");
            my $new_expire = $self->{kvs}->ttl($hashed_key);
            if ($new_expire > 0) {
                $self->{app}->log->debug("update inmemory cache: $hashed_key (TTL:$new_expire)");
                $self->{inmemory}->set($hashed_key, $result, $new_expire);
            }
            return $result;
        }
    }

    # finally, select from DB
    $result = $model->select(@params);
    if ($result) {
        $self->{kvs}->set($hashed_key, $result, $expire);
        $self->{inmemory}->set($hashed_key, $result, $expire);
    }

    return $result;
}



1;
