package Newslash::Util::DOMTestRunner;
use Mojo::Base -base;
use Carp 'croak';

use Data::Dumper;
use Mojo::JSON qw(decode_json);
use Test::More;
use Mojo::Server::Daemon;
use Mojo::UserAgent;

use constant SIGINT => 2;

has [qw(success tx)];


sub new {
    my $self = shift->SUPER::new;

    $self->{app_class} = shift;
    $self->{options} = {@_};

    if (!$self->{app_class}) {
        croak("DOMTestRunnner->new() needs app class");
    }

    $SIG{CHLD} = sub {
        if ($self->{daemon_pid} && !kill(0, $self->{daemon_pid})) {
            croak("daemon process is terminated unexpectedly. exit.");
        }
        if ($self->{nodejs_pid} && !kill(0, $self->{nodejs_pid})) {
            croak("nodejs process is terminated unexpectedly. exit.");
        }
    };

    $self->start_daemon;
    $self->start_domtest_server;
    return $self;
}

sub DESTROY {
    my $self = shift;
    $SIG{CHLD} = 'DEFAULT';
    $self->stop_daemon;
    $self->stop_domtest_server;
    $? = 0;
}

sub _path_to_url {
    my ($self, $path) = @_;
    my $scheme = $self->{options}->{mojo_scheme} || "http";
    my $host = $self->{options}->{mojo_host} || "127.0.0.1";
    my $port = $self->{options}->{mojo_port} || "3000";
    return "$scheme://$host:$port$path";
}

sub _domtest_url {
    my $self = shift;
    my $host = $self->{options}->{domtest_host} || "127.0.0.1";
    my $port = $self->{options}->{domtest_port} || "4000";
    return "http://$host:$port/";
}

sub test {
    my ($self, $path, $test_name) = @_;
    $test_name ||= "test for $path";
    my $ua = Mojo::UserAgent->new;

    my $url = $self->_path_to_url($path);
    my $domtest_url = $self->_domtest_url;

    my $message = { url => $url };
    if ($self->{options}->{ignore_messages}) {
        $message->{ignoreMessages} = $self->{options}->{ignore_messages};
    }
    if ($self->{options}->{domtest_finish_timeout}) {
        $message->{finishTimeout} = $self->{options}->{domtest_finish_timeout};
    }

    my $tx = $ua->post($domtest_url => json => $message);

    $self->tx($tx);
    if ($tx->success) {
        my $res = decode_json($tx->success->body);
        if ($res->{runtimeErrorCount}) {
            diag Dumper $res->{runtimeErrors};
            return $self->success(!!Test::More::ok(0, $test_name));
        }
        else {
            return $self->success(!!Test::More::ok(1, $test_name));
        }
    }
    else {
        my $err = $tx->error;
        diag "$err->{code} response: $err->{message}\n" if $err->{code};
        diag "Connection error: $err->{message}\n";
        return $self->success(!!Test::More::ok(0, $test_name));
    }
}

sub start_domtest_server {
    my $self = shift;

    my $nodejs_cmd = $self->{options}->{nodejs_cmd} || "nodejs";
    my $domtest_script = $self->{options}->{domtest_script} || "./dom_test/app.js";

    my $domtest_host = $self->{options}->{domtest_host} || "127.0.0.1";
    my $domtest_port = $self->{options}->{domtest_port} || "4000";

    my $domtest_cmd = "$nodejs_cmd $domtest_script -H $domtest_host -p $domtest_port";
    my $domtest_url = "http://$domtest_host:$domtest_port/";
    if (my $pid = fork()) {
        # main process
        $self->{nodejs_pid} = $pid;

        # wait for starting server
        my $ua = Mojo::UserAgent->new;
        my $tx = $ua->head($domtest_url);
        my $retry = 0;
        my $max_retry = 5;
        while(!$tx->success && !$tx->error->{code}) {
            sleep(1);
            $tx = $ua->head($domtest_url);
            $retry++;
            if ($retry == $max_retry) {
                croak("cannot access domtest server: $domtest_url");
            }
        }
        return $pid;
    }
    else {
        # sub process
        exec($domtest_cmd);
    }
}

sub stop_domtest_server {
    my $self = shift;
    return if !$self->{nodejs_pid};

    kill(SIGINT, $self->{nodejs_pid});
    waitpid($self->{nodejs_pid}, 0);
    $self->{nodejs_pid} = 0;
}

sub stop_daemon {
    my $self = shift;
    return if !$self->{daemon_pid};

    kill(SIGINT, $self->{daemon_pid});
    waitpid($self->{daemon_pid}, 0);
    $self->{daemon_pid} = 0;
}

sub start_daemon {
    my $self = shift;

    # prepare to run
    my $daemon = Mojo::Server::Daemon->new;
    my $opts = { subcommand => "daemon" };
    if ($self->{options}->{mode}) {
        $opts->{options}->{mode} = $self->{mode};
    }
    my $app = $self->{app_class}->new($opts);
    $daemon->app($app);

    my $listen = [];

    if ($self->{options}->{mojo_listen}) {
        push @$listen, $self->{options}->{mojo_listen};
    }
    else {
        push @$listen, 'http://*:3000';
    }
    $daemon->listen($listen);

    # start daemon by fork
    if (my $pid = fork()) {
        # main process
        $self->{daemon_pid} = $pid;
        return $pid;
    }
    else {
        # sub process - start daemon
        # redirect stderr to stdout
        open(STDERR, '>&', *STDOUT) or croak('cannot redirect STDERR to STDOUT');

        # start daemon
        $daemon->run;
        $daemon->stop;
        exit(0);
    }
}

1;

=encoding utf8

=head1 NAME

Newslash::Util::DOMTestRunner - Test runner for DOM test with dom_test.js

=head1 SYNOPSIS

  use Newslash::Util::DOMTestRunner;
  
  my $dom_test = Newslash::Util::DOMTestRunner->new('Newslash::Web',
                                                  ignore_messages => ['adsbygoogle'],
                                                 );
  
  $dom_test->test("/");
  
  done_testing;

=head1 DESCRIPTION

L<Newslash::Util::DOMTestRunner> is test runnter using dom_test.js.

=head1 METHODS

=head2 new

  my $dom_test = Newslash::Util::DOMTestRunner->new(<app class name>,
                                                    <option1> => <value>,
                                                    <option2> => <value>, ...);

  options:

 * ignore_messages
 * nodejs_cmd
 * domtest_script
 * domtest_host
 * domtest_port
 * mojo_listen
 * mojo_host
 * mojo_port
 * mojo_scheme
 * domtest_finish_timeout

=head2 test

  $dom_test->test("/");

Get DOM and render it virtualy.

=head1 SEE ALSO

=cut
