package Newslash::Util::Test;

use strict;
use warnings;
use utf8;
use feature ':5.10';

use DateTime;
use DateTime::Format::MySQL;

use Exporter 'import';
use Newslash::Plugin::UserAuth;
use Newslash::Model::Users::Util;


our @EXPORT_OK = qw(create_user delete_user
                    create_story delete_story
                    create_admin_user
                    create_comments
                    create_comment delete_comment
                    create_users
                    create_journal delete_journal
                    update_user_passwd
                    get_anonymous_user
                    srcid
                  );

use Data::Dumper;

my $fake_ipid = "0123456789abcdef0123456789abcdef";
my $fake_subnetid = "0123456789abcdef0123456789abcdef";


sub srcid {
    my ($ip, $mask) = @_;
    return Newslash::Plugin::UserAuth::srcid($ip, $mask);
}

=head2 create_user($app, $user_name)

create user for test.

=over 4

=item Parameters

=over 4

=item $app

application / test object

=back

=item $user_name

user name to create

=back

=item Return value

user hashref

=back

=cut

sub create_user {
    my ($app, $user_name, $passwd) = @_;
    my $users = $app->model('users');
    my $new_passwd = $passwd || Newslash::Model::Users::Util::generate_random_password;

    my $user = $users->select(nickname => $user_name);
    if ($user) {
        warn("Test::create_user: user '$user_name' already exists. delete before test.");
        delete_user($app, $user->{nickname});
    }
    my $uid = $users->create($user_name, $user_name . '@example.com', $new_passwd);
    if ($uid) {
        $user = $users->select(uid => $uid);
    }
    else {
        $user = $users->select(nickname => $user_name);
    }
    if (!$user) {
        warn "Newslash::Util::Test:create_user: create user failed. uid: " . ($uid ? $uid : "(no uid)");
        return;
    }
    if ($passwd) {
        my $rs = Newslash::Util::Test::update_user_passwd($app, $user->{uid}, $passwd);
        if (!$rs) {
            warn "passwd update error!";
        }
    }

    $user->{ipid} = $fake_ipid;
    $user->{subnetid} = $fake_subnetid;

    return $user;
}

sub create_admin_user {
    my ($app, $user_name, $passwd) = @_;
    my $user = create_user($app, $user_name, $passwd);
    $app->model('users')->update(uid => $user->{uid}, seclev => 10000, author => 1);
    $user = $app->model('users')->select(uid => $user->{uid});
    return if !$user;

    $user->{ipid} = $fake_ipid;
    $user->{subnetid} = $fake_subnetid;
    return $user;
}

sub update_user_passwd {
    my ($app, $uid, $passwd) = @_;
    my $users = $app->model('users');
    return $users->update(uid => $uid, passwd => $passwd);
}

sub create_users {
    my ($app, $prefix, $users_count) = @_;

    my $test_users = {};
    for my $n (1..$users_count) {
        my $name = $prefix . $n;
        my $user = create_user($app, $name);
        $test_users->{$name} = $user;
    }
    return $test_users;
}


=head2 delete_user($app, $user_name)

delete user for test.

=over 4

=item Parameters

=over 4

=item $app

application / test object

=back

=item $user_name

user name to delete

=back

=item Return value

user hashref

=back

=cut

sub delete_user {
    my ($app, $user_name) = @_;
    my $users = $app->model('users');
    my $user = $users->select(nickname => $user_name);
    return if !$user;

    $users->hard_delete($user->{uid});
    return $user;
}

=head2 create_story($app, $user, @params)

create user for test.

=over 4

=item Parameters

=over 4

=item $app

application / test object

=back

=over 4

=item $user

user hashref

=back

=over 4

=item @params

foo => bar style parameters

=back

=item Return value

story hashref

=back

=cut

sub create_story {
    my $app = shift;
    my $user = shift;
    my $args = {@_};

    my $stories = $app->model('stories');
    if ($args->{params}) {
        my $p = $args->{params};
        delete $args->{params};
        $args = {%$args, %$p};
    }

    if (!$app) {
        warn "invalid app!";
        return;
    }
    if (!$user) {
        warn "invalid user!";
        return;
    }

    if (!$args->{time}) {
        my $dt = DateTime->now();
        $args->{time} = DateTime::Format::MySQL->format_datetime($dt);
    }
    my $params = {
                  uid => $user->{uid},
                  title => $args->{title} || "test story",
                  introtext => $args->{introtext} || '<p>foo bar hoge</p>',
                  bodytext => $args->{bodytext} || '',
                  topics_chosen => $args->{topics} || {49 => 10, 2271 => 20},
                  commentstatus => '',
                  tid => $args->{tid} || 49,
                  dept => $args->{dept} || "",
                  time => $args->{time},
                 };
    my $stoid = $stories->create(user => $user, %$params);
    if (!$stoid) {
        warn("create_story error: " . $stories->last_error);
        return;
    }
    my $story = $stories->select(stoid => $stoid, show_future => 1, show_nonpublic => 1);
    return $story;
}

=head2 delete_story($app, $stoid)

delete story for test.

=over 4

=item Parameters

=over 4

=item $app

application / test object

=back

=item $stoid

story id to delete

=back

=item Return value

undef / 1

=back

=cut

sub delete_story{
    my ($app, $stoid) = @_;
    my $stories = $app->model('stories');
    $stories->hard_delete(stoid => $stoid);

    return 1;
}

sub create_journal {
    my $app = shift;
    my $user = shift;
    my $args = {@_};

    my $journals = $app->model('journals');
    if ($args->{params}) {
        my $p = $args->{params};
        delete $args->{params};
        $args = {%$args, %$p};
    }

    return if !$app;
    return if !$user;
    my $id = $journals->create(user => $user,
                               description => $args->{title} || "test title",
                               article => $args->{intro_text} || $args->{introtext} || "test journal",
                               commentstatus => $args->{commentstatus} || $args->{comment_status} || "enabled",
                               posttype => 2);
    if (!$id) {
        warn($journals->last_error);
        return;
    }
    my $journal = $journals->select(journal_id => $id);
    return $journal;
}

sub delete_journal {
    my ($app, $id) = @_;
    my $rs = $app->model('journals')->hard_delete(id => $id);
    return $rs;
}

sub create_submission {
    my $app = shift;
    my $user = shift;
    my $args = {@_};

    my $submissions = $app->model('submissions');
    if ($args->{params}) {
        my $p = $args->{params};
        delete $args->{params};
        $args = {%$args, %$p};
    }

    return if !$app;
    return if !$user;
    my $id = $submissions->create(user => $user,
                                  title => $args->{title} || "test title",
                                  introtext => $args->{intro_text} || $args->{introtext} || "test submission",
                                  tid => 1913,
                                 );
    if (!$id) {
        warn($submissions->last_error);
        return;
    }
    my $submission = $submissions->select(submission_id => $id);
    return $submission;
}

sub delete_submission {
    my ($app, $id) = @_;
    my $rs = $app->model('submissions')->hard_delete(id => $id);
    return $rs;
}

sub create_comment {
    my ($app, $user, $discussion_id, $title, $comment, $pid) = @_;
    my $params = {
                  discussion_id => $discussion_id,
                  title => $title,
                  comment => $comment,
                  pid => $pid,
                 };
    my $cid = $app->model('comments')->create($params, $user);
    return if (!$cid);
    my $comm = $app->model('comments')->select(cid => $cid);
    return $comm;
}

sub delete_comment {
    my ($app, $cid) = @_;
    my $rs = $app->model('comments')->hard_delete(cid => $cid);
    return $rs;
}

sub create_comments {
    my ($app, $discussion_id, $comms, $pid) = @_;
    my $rs = {};
    for my $c (@$comms) {
        my $comment = create_comment($app, $c->{author}, $discussion_id,
                                 $c->{title}, $c->{comment}, $pid);
        $rs->{$c->{key}} = $comment;
        if ($c->{children}) {
            my $rs2 = create_comments($app, $discussion_id, $c->{children}, $c->{cid});
            $rs = { %$rs, %$rs2 };
        }
    }
    return $rs;
}

sub get_anonymous_user {
    my $app = shift;
    return if !$app;
    my $anon = $app->model('users')->anonymous_user;

    $anon->{ipid} = $fake_ipid;
    $anon->{subnetid} = $fake_subnetid;
    return $anon;

}

1;
