package Newslash::Web::Controller::API::Story;
use Mojo::Base 'Mojolicious::Controller';
use Mojo::Util qw(dumper);
use Mojo::JSON qw(decode_json encode_json);

use DateTime::Format::ISO8601;
use DateTime::Format::MySQL;
use Data::Dumper;

sub _submission_to_story {
    my ($c, $submission) = @_;
    my $story = {};
    for my $k (qw{title tags}) {
        $story->{$k} = $submission->{$k};
    }
    # convert intro_text
    $story->{intro_text} = $c->tt2renderer->render("system/submission/submission_formatter", $submission);
    $story->{submitter} = $submission->{uid};
    return $story;
}

sub _journal_to_story {
    my ($c, $journal) = @_;
    my $story = {};
    for my $k (qw{title tags}) {
        $story->{$k} = $journal->{$k};
    }
    $story->{intro_text} = $c->tt2renderer->render('system/subsmission/journal2submission',
                                                   { text => $journal->{full_text},
                                                     author => $journal->{author},
                                                   });
    if (!$story->{intro_text}) {
        $story->{intro_text} = $journal->{full_text};
    }
    $story->{submitter} = $journal->{uid};
    return $story;
}

sub get {
    my $c = shift;
    my $sid = $c->param('sid');
    my $stoid = $c->param('stoid');
    my $subid = $c->param('subid') || $c->param('submission_id');
    my $journal_id = $c->param('journal_id');
    my $stories = $c->model('stories');
    my $story;
    my $user = $c->stash('user');

    if ($stoid) {
        $story = $stories->select(story_id => $stoid,
                                  hide_future => !$user->{is_admin},
                                  public_only => !$user->{is_admin},
                                 );
    }
    elsif ($sid) {
        $story = $stories->select(sid => $sid,
                                  hide_future => !$user->{is_admin},
                                  public_only => !$user->{is_admin},
                                 );
    }
    elsif ($subid) {
        my $item = $c->model('submissions')->select(subid => $subid);
        if ($item) {
            $story = $c->_submission_to_story($item);
        }
    }
    elsif ($journal_id) {
        my $item = $c->model('journals')->select(id => $journal_id);
        if ($item) {
            $story = $c->_journal_to_story($item);
        }
    }
    else {
        $c->render(json => { error => 1, message => "invalid_request" });
        $c->rendered(400);
        return;
    }

    if (!$story) {
        $c->render(json => { error => 1, message => "not_found" });
        $c->rendered(404);
        return;
    }
    my $related = $stories->get_related_items(stoid => $story->{stoid});
    $story->{related} = $related;
    $c->render(json => { item => $story });
}

sub post {
    my $c = shift;
    my $params = {};

    my $user = $c->stash('user');
    my $stories = $c->model('stories');
    my $users = $c->model('users');
    my $topics = $c->model('topics');
    my $submissions = $c->model('submissions');

    my $data = $c->req->json;
    my $item = $data->{item};
    my $message = "";

    if (!$user->{is_admin} && !$user->{editor}) {
        $c->render(json => { err => 1, message => "no_permission" });
        $c->rendered(403);
        return;
    }

    # client send data like this:
    # {"item":{"title":"ほげ","introtext":"ほんぶん","bodytext":"","createtime":"2017-06-30T21:50:04","author":"hylom","dept":"テスト","commentstatus":"enabled","submissioncopy":0,"url":"","email":"","tags_string":"news","related_urls":"http://sdtest.osdn.co.jp:3000/story/16/01/28/0610215/"},"action":"preview"}

    $params->{title} = $c->format_htmltext($item->{title}, "title");
    $message = "no_title" if !$params->{title};

    # convert time
    my $dt = DateTime::Format::ISO8601->parse_datetime($item->{createtime} || $item->{create_time});
    if (!$dt) {
        $message = "invalid_createtime";
    }
    else {
        # convert localtime to GMT
        if ($user && $user->{config}->{ui}->{offset_sec}) {
            my $offset_sec = $user->{config}->{ui}->{offset_sec} || 0;
            if ($offset_sec) {
                $dt->subtract(seconds => $offset_sec);
            }
        }
        $params->{time} = DateTime::Format::MySQL->format_datetime($dt);
    }

    # check: is author valid ?
    my $author = $users->select(nickname => $item->{author});
    if (!defined $author || !$author->{is_admin}) {
        $message = "invalid_author";
    }
    $params->{uid} = $author->{uid};

    # check: dept exists ?
    $params->{dept} = $item->{dept};
    if (!$params->{dept}) {
        $message = "department_not_given";
    }

    # check: introtext exists ?
    my $intro_text = $item->{intro_text} || $item->{introtext} || "";
    $params->{introtext} = $c->format_htmltext($intro_text, "story");
    $params->{intro_text} = $params->{introtext};
    if (!$params->{introtext}) {
        $message = "introtext_not_given";
    }

    # bodytext (omittable)
    my $body_text = $item->{body_text} || $item->{bodytext} || "";
    $params->{bodytext} = $c->format_htmltext($body_text, "story");

    # related urls
    my $related_string = $item->{add_related} || $item->{related_urls} || "";
    my @relateds = split(/\s+/, $related_string);
    my $related_sids = [];
    for my $related (@relateds) {
        if ($related =~ m|^.*(\d\d/\d\d/\d\d/\d+)/?|) {
            push @$related_sids, $1;
        }
    }
    $params->{add_related} = $related_sids;

    # tags
    # tags_strings is like: "mainpage japan business transport"
    my @tags = split(/\s+/, $item->{tags_string} || "");
    $params->{add_tags} = \@tags;

    # add topics
    my $primary_topic = $topics->get_primary_topic_from_tags(\@tags);
    if ($primary_topic) {
        $params->{topics_chosen} = { $primary_topic->{tid} => 10 };
    }
    else {
        $params->{topics_chosen} = { 49 => 10, }; # tid => weight
    }

    # TODO: media related
    #$item->{mediaurl};
    #$item->{mediatype};
    #$item->{media};

    if ($item->{media}) {
	$params->{media} = $item->{media};
    }

    # source
    if ($item->{source_id}) {
	if ($item->{source_type} eq "submission") {
	    $params->{submission_id} = $item->{source_id};
	}
	elsif ($item->{source_type} eq "journal") {
	    $params->{journal_id} = $item->{source_id};
	}
    }

    $params->{commentstatus} = $item->{commentstatus} || $item->{comment_status} || "enabled";
    if (!$item->{display}) {
	$params->{neverdisplay} = 1;
    }

    $params->{action} = $data->{action} || 'preview';

    if (length($message) > 0) {
        $c->render(json => { err => 1, message => $message });
        $c->rendered(400);
        return;
    }

    if ($data->{action} eq 'preview') {
        $c->render(json => {item => $params});
        return;
    }

    # invalid action
    if ($params->{action} ne 'post') {
        $c->render(json => { err => 1, message => "invalid_action" });
        $c->rendered(400);
        return;
    }

    # post action
    my ($sid, $stoid);
    if ($item->{stoid} || $item->{id}) {
        # update story
        $params->{stoid} = $item->{stoid} || $item->{id};
        ($sid, $stoid) = $stories->update(user => $user, %$params);
    }
    else {
        # create story
        ($sid, $stoid) = $stories->create(user => $user, %$params);
        #$stoid = $stories->create2(user => $user,
        #                           config => $c->config->{Story},
        #                           params => $params);
    }

    # error check
    if (!$stoid) {
        if ($stories->last_errorno && $stories->last_errorno == 1062) { #ER_DUP_ENTRY
            $c->render(json => { error => 1, message => "duplicated_entry" });
            $c->rendered(409);
            return;
        }
        $c->render(json => { error => 1, message => $stories->last_error });
        $c->rendered(500);
        return;
    }

    # post succeeded, return result
    $c->render(json => {type => "story", id => $stoid, sid => $sid, url_id => $sid});
    $c->event_que->emit("story", "post", $user->{uid}, $stoid);
    $c->stats->add_event_counter("story_create");
    return;
}

1;
