package Newslash::Web::Controller::API::Timeline;
use Mojo::Base 'Mojolicious::Controller';
use Data::Dumper;
use List::Util qw(any);

sub _content_url {
    my ($c, $item) = @_;
    return if !$item || !$item->{content_type};

    if ($item->{content_type} eq "journal") {
        if (!$item->{author}) {
            $c->app->log->error("Timeline::_add_url: invalid author for id: $item->{id}");
            return;
        }
        return "/~$item->{author}/journal/$item->{id}/";
    }
    elsif ($item->{content_type} eq "story") {
        return "/story/$item->{sid}/";
    }
    else {
        return "/$item->{content_type}/$item->{id}/";
    }
    return;
}

sub _get_heatmap {
    my $c = shift;
    my $cfg = $c->config("Timeline");
    my $heatmap = $cfg->{heatmap};

    if (!$heatmap) {
        $c->app->log->error("Timeline: no heatmap defined.");
        return;
    }

    my @keys = keys %$heatmap;
    @keys = sort { $heatmap->{$a} <=> $heatmap->{$b} } @keys;
    my $rs = [];
    for my $k (@keys) {
        push @$rs, { $k => $heatmap->{$k} };
    }
    return $rs;
}

sub _get_primary_topic_icon_url {
    my ($c, $item) = @_;
    my $cfg = $c->config("Site") || {};
    my $base_url = $cfg->{topic_icon_base_url};

    if (!$base_url) {
        $c->app->log->error("Timeline: Site.topic_icon_base_url is not defined");
        return;
    }
    my $t = $item->{primary_topic} || {};
    my $image = $t->{image};

    if ($image) {
        return "$base_url/$image";
    }
    return;
}

sub _moderation_points_to_score {
    my ($c, $item) = @_;
    my $point = $item->{points};
    return if !defined $point;

    my $heatmap = _get_heatmap($c);
    if (!$heatmap) {
        return;
    }

    my $color = $heatmap->[$point + 1];
    return -50 if !$color;

    my @scores = values %$color;
    return $scores[0] + 1;
}

sub _score_to_heatmap_color {
    my ($c, $item) = @_;
    my $heatmap = _get_heatmap($c);
    if (!$heatmap) {
        return;
    }

    my $last_color;
    for my $i (reverse @$heatmap) {
        my @k = keys %$i;
        my $color = $last_color = $k[0];
        my $threshold = $i->{$color};

        if (defined $item->{popularity} && $item->{popularity} > $threshold) {
            return $color;
        }
    }
    return $last_color;
}

sub _threshold_to_popularity {
    my ($c, $threshold) = @_;
    my $heatmap = _get_heatmap($c);
    if (!$heatmap) {
        $c->app->log->error("Timeline::_threshold_to_popularity: no heatmap defined.");
        return;
    }

    my $color = $heatmap->[$threshold];
    if (!$color) {
        $c->app->log->error("Timeline::_threshold_to_popularity: no color for threshold $threshold.");
        return;
    }

    my @k = keys %$color;
    return $color->{$k[0]};
}


sub get {
    my $c = shift;
    my $user = $c->stash('user');
    my $params = $c->req->query_params->to_hash;
    my $target = $params->{target} || "all";
    my $cfg = $c->config("Timeline");

    my $hide_future = !$user->{is_admin} && !$user->{editor};
    my $public_only = !$user->{is_admin} && !$user->{editor};

    my $limit = $params->{limit} || $cfg->{item_per_page} || 10;

    my $max_limit = $cfg->{item_per_page_limit} || 1000;
    $limit = $max_limit if $limit > $max_limit;

    my $skip = $params->{skip} || 0;
    my $min_popularity;
    if ($params->{threshold}) {
        $min_popularity = _threshold_to_popularity($c, $params->{threshold});
        # $c->app->log->debug("Timeline::_threshold_to_popularity: use min_pop $min_popularity.");
    }

    # check date-range parameter
    my ($term, $range);
    for my $k (qw{years months weeks days hours minutes}) {
        if ($params->{$k}) {
            $term = $k;
            $range = $params->{$k};
            last;
        }
    }

    my $result;
    my $model;

    if ($target eq "story") {
        $model = $c->ccache->model('stories');
    }
    elsif ($target eq "journal") {
        $model = $c->ccache->model('journals');
    }
    elsif ($target eq "comment") {
        $model = $c->ccache->model('comments');
        if ($params->{threshold} && $params->{threshold} > 6) {
            # comment's moderation point is lesser than 7, no item found
            $c->render(json => { error => { code => -1, message => "not_found" }});
            $c->rendered(404);
            return;
        }
    }
    elsif ($target eq "poll") {
        $model = $c->ccache->model('polls');
    }
    elsif ($target eq "submission") {
        $model = $c->ccache->model('submissions');
    }
    elsif ($target eq "all") {
        $model = $c->ccache->model('timeline');
    }
    elsif ($target eq "recent") {
        $model = $c->ccache->model('timeline');
    }
    elsif ($target eq "popular") {
        $model = $c->ccache->model('timeline');
    }
    else {
        $c->render(json => { error => { code => -1, message => "invalid_request" }});
        $c->rendered(400);
        return;
    }

    # build query
    my @q_params = (hide_future => $hide_future,
                    public_only => $public_only,
                    limit => $limit + 1,
                    skip => $skip,
                   );
    if ($target eq "popular") {
        push @q_params, order_by => [ popularity => 'desc',
                                      create_time => 'desc' ];
        if (!$term) {
            # set default date range
            my $cfg = $c->config("Timeline") || {};
            my $popular_period = $cfg->{popular_period} || {days => 1};
            for my $k (keys %$popular_period) {
                if ($k =~ /^(minutes|hours|days|weeks|months|years)$/) {
                    $term = $k;
                    $range = $popular_period->{$k};
                    last;
                }
            }
        }

        if (!$term) {
            #fallback
            $term = "days";
            $range = 1;
        }
    }
    else {
        push @q_params, order_by => {create_time => 'desc'};
    }

    push @q_params, $term => $range if $term;

    if ($target eq "comment") {
        # TODO: comment score is independent with comment score...
        push @q_params, points => { ge => $params->{threshold} - 1 } if defined $params->{threshold};
    }
    else {
        push @q_params, popularity => { ge => $min_popularity } if defined $min_popularity;
    }

    if ($user->{is_login}) {
        $result = $model->select_nocache(@q_params);
    }
    else {
        $result = $model->select(@q_params);
    }

    if (!$result) {
        $c->render(json => { error => { code => -1, message => "internal_server_error" }});
        $c->rendered(500);
        return;
    }

    if (!@$result) {
        $c->render(json => { error => { code => -1, message => "not_found" }});
        $c->rendered(404);
        return;
    }

    my $prev_item;
    if (@$result == $limit + 1) {
        my $i = pop @$result;

        my $dt = $c->format_timestamp(user => $user,
                                      mysql => $i->{create_time},
                                      format => "datetime");
        $prev_item = { id => $i->{id},
                       year => $dt->year,
                       month => $dt->month,
                       day => $dt->day,
                       content_type => $i->{content_type},
                     };
    }

    # add headmap info and topic icon url
    my $item_for = {};
    for my $item (@$result) {
        # convert comment moderation point to popularity for comments
        if ($item->{content_type} eq "comment") {
            $item->{popularity} = _moderation_points_to_score($c, $item);
        }

        $item->{color} = _score_to_heatmap_color($c, $item);
        $item->{icon_url} = _get_primary_topic_icon_url($c, $item);
        $item->{content_url} = _content_url($c, $item);
        $item->{tagmoderation} = "";
        $item->{formated_timestamp} = $c->format_timestamp(user => $user,
                                                           mysql => $item->{create_time});

        # create index
        if ($item->{globjid}) {
            $item_for->{$item->{globjid}} = $item;
        }

        $c->blockquote_submission($item);
}

    # add usertag
    if ($user->{is_login}) {
        my @globjids = map { $_->{globjid} } @$result;
        my $tags = $c->model('tags');
        my $rs = $tags->select_usertag_for_globjids($user->{uid}, \@globjids);
        if ($rs) {
            for my $tag (@$rs) {
                if (my $i = $item_for->{$tag->{globjid}}) {
                    $i->{user_tags} //= [];
                    push @{$i->{user_tags}}, $tag;
                    if ($tag->{tagname} eq "nod" || $tag->{tagname} eq "nix") {
                        $i->{tagmoderation} = $tag->{tagname};
                    }
                }
            }
        }
        else {
            $c->app->log->error("Timeline::get: cannot get usertags. error: " . $tags->last_error);
        }
    }

    $c->render(json => { result => { items => $result,
                                     prev_info => $prev_item, }
                       });
}


1;
