/* newslash.js */

var Newslash = function (options) {
  options = options || {};
  this.baseUrl = options.baseUrl || '/api/v1';
};

function _initNewslash() {
  // fake Promise for old browser
  function FakePromise(f) {
    this.then = function(resolve, reject) {
      f(resolve, reject);
    };
  }

  Newslash.prototype.fail = function fail(param) {
    return new FakePromise(function (resolve, reject) {
      return reject(param);
    });
  };

  Newslash.prototype.urlParse = function (url) {
    var parser = document.createElement('a');
    parser.href = url;
    return parser;
  };

  /* Utility API */
  Newslash.prototype.pathFromDate = function (year, month, day) {
    try {
      year = String(year);
      month = String(month);
      day = String(day);
    }
    catch (e) { return ""; }
    if (!year || !month || !day) { return ""; }
    var path = "/";
    path += year.length > 2 ? year.substring(2) : year;
    path += month.length > 1 ? ("/" + month) : ("/0" + month);
    path += day.length > 1 ? ("/" + day) : ("/0" + day);
    return path + "/";
  };

  /* CSRF securiry token */
  Newslash.getCsrfToken = function getCsrfToken(resourceID) {
    if (!resourceID) { return this.fail(); }
    var url = "/api/v1/token?rc=" + resourceID;
    return this.get(url);
  };

  /* base get.post API */
  Newslash.prototype.get = function (path) {
    var url = this.baseUrl + path;
    return new FakePromise(function (resolve, reject) {
      var req = new XMLHttpRequest();
      req.open('GET', url);
      req.responseType = "json";
      req.onreadystatechange = function onReadyStateChange () {
        if (req.readyState == XMLHttpRequest.DONE) {
          // hack for IE/Edge support (req.responseType == "json" is not supported)
          var respType = req.responseType;
          var resp = req.response;

          if (respType === "") {
            try {
              resp = JSON.parse(req.response);
              respType = "json";
            } catch(e) {
              console.log("Newslash.post: cannot parse JSON");
            }
          }

          // request done
          if (req.status == 200 && respType == "json") {
            return resolve(resp);
          }

          // request failed
          if (respType == "json") {
            return reject(resp);
          }
          return reject(req);
        }
        return 1;
      };
      req.ontimeout = function onTimeout () {
        return reject(req);
      };
      req.send();
    });
  };

  Newslash.prototype.post = function (path, data) {
    var url = this.baseUrl + path;

    return new FakePromise((resolve, reject) => {
      _doPost(path, data, 0, resolve, reject);
    });

    function _doPost(path, data, retryCount, resolve, reject) {
      var req = new XMLHttpRequest();
      req.open('POST', url);
      req.responseType = "json";
      req.onreadystatechange = () => {
        if (req.readyState == XMLHttpRequest.DONE) {
          // hack for IE/Edge support (req.responseType == "json" is not supported)
          var respType = req.responseType;
          var resp = req.response;

          if (respType === "") {
            try {
              resp = JSON.parse(req.response);
              respType = "json";
            } catch(e) {
              console.log("Newslash.post: cannot parse JSON");
            }
          }

          // request done
          if (req.status == 200 && respType == "json") {
            return resolve(resp);
          }
          
          // request failed
          if (respType == "json") {
            if (resp && resp.reason == "invalid_csrf_token") {
              if (retryCount >= 3) {
                return reject({error: 1, reason: "exceeded retry count"});
              }
              var newToken = resp.csrf_token;
              if (newToken) {
                data.csrf_token = newToken;
                // retry with new token
                _doPost(path, data, retryCount + 1, resolve, reject);
                return 1;
              }
            }
            return reject(resp);
          }
          return reject(req);
        }
        return 1;
      };
      req.ontimeout = function onTimeout () {
        return reject(req);
      };
      req.setRequestHeader("Content-Type", "application/json");
      req.send(JSON.stringify(data));
    }
  };

  /* login API */
  Newslash.prototype.login = function (username, password) {
    var url = "/login";
    var data = {nickname: username, password: password};
    return this.post(url, data);
  };

  /* newUser API */
  Newslash.prototype.validateNewUser = function (username, email) {
    var url = "/newuser/validate";
    var data = {nickname: username, email: email};
    return this.post(url, data);
  };

  Newslash.prototype.createNewUser = function (username, email, options) {
    options = options || {};
    var url = "/newuser/create";
    var data = { nickname: username,
                 email: email,
                 options: options,
               };
    return this.post(url, data);
  };

  /* user related API */
  Newslash.prototype.updatePasswordByToken = function (username, token, password, options) {
    options = options || {};
    var url = "/newuser/password";
    var data = { nickname: username,
                 token: token,
                 password: password,
                 options: options,
               };
    return this.post(url, data);
  };

  Newslash.prototype.requestEmailChange = function (newAddress, options) {
    options = options || {};
    var url = "/user";
    var data = { type: "email", address: newAddress, };
    return this.post("/user", data);
  };

  Newslash.prototype.updatePassword = function (oldPassword, newPassword, options) {
    options = options || {};
    var url = "/user";
    var data = { type: "password",
                 current_password: oldPassword,
                 new_password: newPassword };
    return this.post("/user", data);
  };

  Newslash.prototype.resetPassword = function (email) {
    var url = "/login";
    var data = { type: "reset_password",
                 email: email };
    return this.post(url, data);
  };

  Newslash.prototype.getMessageSettings = function getMessageSettings () {
    return this.get("/user?type=message_config");
  };

  Newslash.prototype.updateSettings = function updateSettings (settings) {
    if (!settings) return this.fail();
    var data = { type: "config", config: settings };
    return this.post("/user", data);
  };

  Newslash.prototype.updateMessageSettings = function updateMessageSettings (items) {
    if (!items) return this.fail();
    var data = { type: "message", items: items };
    return this.post("/user", data);
  };

  /* comments API */
  Newslash.prototype.getComments = function getComments (discussionID, parentID) {
    if (!discussionID) return this.fail();

    var url = "/comment?discussion_id=" + discussionID;
    if (parentID) {
      url = url + "&parent_id=" + parentID;
    }
    return this.get(url);
  };

  /* journals API */
  Newslash.prototype.getJournal = function getJournal (journalID) {
    if (!journalID) return this.fail();

    var url = "/journal?id=" + journalID;
    return this.get(url);
  };

  Newslash.prototype.postJournal = function createJournal (data) {
    data.action = "post";
    return this.post("/journal", data);
  };

  Newslash.prototype.previewJournal = function previewJournal (data) {
    data.action = "preview";
    return this.post("/journal", data);
  };

  /* submissions API */
  Newslash.prototype.postSubmission = function createSubmission (data) {
    data.action = "post";
    return this.post("/submission", data);
  };

  Newslash.prototype.previewSubmission = function previewSubmission (data) {
    data.action = "preview";
    return this.post("/submission", data);
  };

  /* timeline related API */
  Newslash.prototype.getTimeline = function getTimeline (target, options) {
    if (!target) { target = "all"; }
    options = options || {};

    var url = "/timeline?target=" + target;
    if (options.threshold !== undefined) {
      url = url + "&threshold=" + options.threshold;
    }

    if (options.dateRange && options.dateRange != "all") {
      if (!/^(\d+)_([a-z]+)$/.exec(options.dateRange)) {
        return this.fail("invalid_date_range");
      }
      var ret = options.dateRange.split("_");
      var term = ret[0];
      var range = ret[1];
      
      url = url + "&" + range + "=" + term;
    }
    
    return this.get(url);
  };

  Newslash.prototype.vote = function vote (globjId, voteType) {
    var url = "/tagmoderation";
    if (voteType != "nod" && voteType != "nix") {
      return this.fail("invalid_voteType");
    }
    var data = { action: "vote",
                 params: { name: voteType,
                           globjid: globjId, },
               };
    return this.post(url, data);
  };

  Newslash.prototype.cancelVote = function cancelvote (globjId, voteType) {
    var url = "/tagmoderation";
    if (voteType != "nod" && voteType != "nix") {
      return this.fail("invalid_voteType");
    }
    var data = { action: "unset",
                 params: { name: voteType,
                           globjid: globjId, },
               };
    return this.post(url, data);
  };
}

_initNewslash();
var newslash = new Newslash();
