# -*-Perl-*-
# destructive database-related  tests
use Mojo::Base -strict;
use Mojo::Date;
use Test::More;
use Test::Mojo;
use Mojo::Util qw(dumper);
use POSIX qw(strftime);
use DateTime::Format::MySQL;

use Newslash::Util::TestMan;
use Newslash::Eventd::Handlers;
use Data::Dumper;

my $t = Test::Mojo->new('Newslash::Web');
my $test_man = Newslash::Util::TestMan->new($t);
my $test_nickname = "usertest";
my $default_passwd = "foobar_usertest";
my $user;

# prepare for test
if ($t->app->mode eq 'test') {
    # at first, create test users.
    $user = $test_man->create_user($test_nickname, $default_passwd);
}

subtest 'user query' => sub {
    $t->get_ok("/api/v1/user?uid=$user->{uid}")
      ->status_is(200)
      ->content_type_like(qr|application/json|)
      ->json_has('/item')
      ->json_is('/item/uid', $user->{uid})
      ->or(sub {diag "message: " . dumper($t->tx->res->json);});

    $t->get_ok("/api/v1/user?nickname=$user->{nickname}")
      ->status_is(200)
      ->content_type_like(qr|application/json|)
      ->json_has('/item')
      ->json_is('/item/uid', $user->{uid})
      ->or(sub {diag "message: " . dumper($t->tx->res->json);});
};

subtest 'get users info' => sub {
    $test_man->login($user);

    # config
    $t->get_ok('/api/v1/user?type=config')
      ->status_is(200)
      ->content_type_like(qr|application/json|)
      ->json_has('/config')
      ->or(sub {diag "message: " . dumper($t->tx->res->json);});

    # message_config
    $t->get_ok('/api/v1/user?type=message_config')
      ->status_is(200)
      ->content_type_like(qr|application/json|)
      ->json_has('/config')
      ->or(sub {diag "message: " . dumper($t->tx->res->json);});

    # sidebar
    $t->get_ok('/api/v1/user?type=sidebar')
      ->status_is(200)
      ->content_type_like(qr|application/json|)
      ->json_has('/items')
      ->or(sub {diag "message: " . dumper($t->tx->res->json);});

    # all
    # type=all need admin privileges
    $t->get_ok('/api/v1/user?type=all')
      ->status_is(403);

    $test_man->logout;
};

subtest 'validate new users' => sub {
    my $new_email = 'hjkhkjadsf@example.com';

    # nickname exists
    $t->post_ok('/api/v1/newuser/validate' => {Accept => '*/*'} => json => {nickname => $test_nickname, email => $new_email})
      ->status_is(400)
      ->content_type_like(qr|application/json|)
      ->json_has('/error')
      ->json_is('/id_error', "ID_EXISTS")
      ->or(sub {diag "message: " . dumper($t->tx->res->json);});

    # invalid nickname
    $t->post_ok('/api/v1/newuser/validate' => {Accept => '*/*'} => json => {nickname => 'a&~foo', email => $new_email})
      ->status_is(400)
      ->content_type_like(qr|application/json|)
      ->json_has('/error')
      ->json_is('/id_error', "INVALID_ID")
      ->or(sub {diag "message: " . dumper($t->tx->res->json);});

    # exists email
    $t->post_ok('/api/v1/newuser/validate' => {Accept => '*/*'} => json => {nickname => 'hoge', email => $user->{realemail}})
      ->status_is(400)
      ->content_type_like(qr|application/json|)
      ->json_has('/error')
      ->json_is('/email_error', "ADDRESS_EXISTS")
      ->or(sub {diag "message: " . dumper($t->tx->res->json);});

    # invalid email
    $t->post_ok('/api/v1/newuser/validate' => {Accept => '*/*'} => json => {nickname => 'foobar', email => 'test01example.com'})
      ->status_is(400)
      ->content_type_like(qr|application/json|)
      ->json_has('/error')
      ->json_is('/email_error', "INVALID_ADDRESS")
      ->or(sub {diag "message: " . dumper($t->tx->res->json);});

    # ok
    $t->post_ok('/api/v1/newuser/validate' => {Accept => '*/*'} => json => {nickname => 'hoge', email => $new_email})
      ->status_is(200)
      ->content_type_like(qr|application/json|)
      ->json_hasnt('/error');
};

subtest 'update email' => sub {
    $test_man->login($user);

    my $new_address = 'foobarhogehoge@example.com';
    my $users = $t->app->model('users');

    # send invalid address
    $t->post_ok('/api/v1/user' => {Accept => '*/*'} => json => { type => "email",
                                                                 address => "foobarhogehoge" })
      ->status_is(400)
      ->content_type_like(qr|application/json|)
      ->json_is('/error', 1)
      ->json_is('/message', "INVALID_ADDRESS");
    diag dumper $t->tx->res->json if !$t->success;

    # send
    $t->post_ok('/api/v1/user' => {Accept => '*/*'} => json => { type => "email",
                                                                 address => $new_address })
      ->status_is(200)
      ->content_type_like(qr|application/json|)
      ->json_hasnt('/error')
      ->json_is('/uid', $user->{uid})
      ->json_is('/email', $new_address);
    diag dumper $t->tx->res->json if !$t->success;


    # check if event emited
    my $events = $t->app->model('events');
    my $ev = $events->select(event_type => "user",
                             event_subtype => "change_email",
                             source_user_id => $user->{uid},
                             active => 1,
                            );
    ok($ev && @$ev > 0, "event emited");

    # check if event processed correctly
    my $handlers = Newslash::Eventd::Handlers->new($t->app, { test => 1 });
    my $rs = $handlers->execute($ev->[0]);
    ok($rs, "event processed correctly");

    # check email que
    my $user_handler = $handlers->{_handlers}->{user};
    ok($user_handler, "retrieve user event handler");
    my $que = $user_handler->{_test_mailque};

    # check validate mail
    my $validate_mail = pop @$que;
    ok($validate_mail, "retrieve validate mail");
    is($validate_mail->{to}, $new_address, "validate mail address is correct");
    my $url_regex = qr|/my/change_email\?token=(\S+)|;
    ok($validate_mail->{body} =~ $url_regex, "validate mail has correct url");
    my $token = $1;
    ok($token, "extract token from mail");
    diag Dumper $validate_mail if !$1;

    # verify token
    # check with invalid user
    $test_man->logout;
    $t->get_ok("/my/change_email?token=$token")
      ->status_is(307); # redirect

    # check with invalid token
    $test_man->login($user);
    $t->get_ok("/my/change_email?token=hogehoge")
      ->status_is(400);

    # check with valid token
    $t->get_ok("/my/change_email?token=$token")
      ->status_is(200);
    diag $t->tx->res->body if !$t->success;

    # check if update done correctly
    my $u = $users->select(uid => $user->{uid});
    is($u->{realemail}, $new_address, "realemail updated correctly");

    # request and cancel
    $new_address = 'test2test@example.com';
    ok($users->param->set(uid => $user->{uid},
                          name => "new_email",
                          value => $new_address),
       "set new address");
    $token = $t->app->users->generate_email_change_token($u->{uid});
    $t->get_ok("/my/change_email?cancel=1")
      ->status_is(200);

    # check if update done correctly
    $u = $users->select(uid => $user->{uid});
    isnt($u->{realemail}, $new_address, "realemail didn't update");

    # restore email for next test
    ok($users->param->set(uid => $user->{uid},
                          name => "new_email",
                          value => $user->{realemail}),
       "set old address");
    $token = $t->app->users->generate_email_change_token($u->{uid});
    $t->get_ok("/my/change_email?token=$token")
      ->status_is(200);

    # done
    $test_man->logout;
};

subtest 'update password' => sub {
    $test_man->login($user);
    my $new_password = "foobarhogehoge";

    $t->post_ok('/api/v1/user' => {Accept => '*/*'} => json => { type => "password",
                                                                 current_password => $default_passwd,
                                                                 new_password => $new_password })
      ->status_is(200)
      ->content_type_like(qr|application/json|)
      ->json_hasnt('/error')
      ->json_is('/uid', $user->{uid});
    diag dumper $t->tx->res->json if !$t->success;

    # check if database updated
    my $users = $t->app->model('users');
    my $updated = $users->select(uid => $user->{uid});
    my $rs = $users->passwords->compare_password($new_password,
                                                 $updated->{passwd});
    ok($rs, "password updated");

    # check if event emited
    my $events = $t->app->model('events');
    my $ev = $events->select(event_type => "user",
                             event_subtype => "update_password",
                             source_user_id => $user->{uid},
                             active => 1);
    ok($ev && @$ev > 0, "event emited");

    # check if event processed correctly
    my $handlers = Newslash::Eventd::Handlers->new($t->app, { test => 1 });
    $rs = $handlers->execute($ev->[0]);
    ok($rs, "event processed correctly");

    # check email que
    my $user_handler = $handlers->{_handlers}->{user};
    ok($user_handler, "retrieve user event handler");
    my $que = $user_handler->{_test_mailque};

    # check notify mail
    my $notify_mail = pop @$que;
    ok($notify_mail, "retrieve notify mail");
    is($notify_mail->{to}, $user->{realemail}, "notify mail address is correct");
    my $regex = qr|パスワード|;
    $rs = $notify_mail->{title} =~ $regex;
    ok($rs, "notify mail has correct title");
    diag Dumper $notify_mail if !$rs;

    #done
    $test_man->logout;
};

subtest 'reset password' => sub {
    $t->post_ok('/api/v1/login' => {Accept => '*/*'} => json => { type => "reset_password",
                                                                  email => $user->{realemail} })
      ->status_is(200)
      ->content_type_like(qr|application/json|)
      ->json_hasnt('/error');
    diag dumper $t->tx->res->json if !$t->success;

    # check if event emited
    ok($test_man->process_event(event_type => "user",
                                event_subtype => "reset_password",
                                source_user_id => $user->{uid}),
       "process event");

    # check validate mail
    my $mail = $test_man->pop_mailque()
      ->to_is($user->{realemail})
      ->body_matches(qr|/my/activation\?nickname=(?<nickname>\S+)\&token=(?<token>\S+)|);
    my $nickname = $mail->last_match->{nickname};
    my $token = $mail->last_match->{token};
    ok($nickname, "extract nickname from mail") or diag $mail->body;
    ok($token, "extract token from mail") or diag $mail->body;

    # check with invalid token
    $t->get_ok("/my/activation?nickname=$nickname&token=hogehoge&cancel=1")
      ->status_is(400);
    diag $t->tx->res->body if !$t->success;

    # cancel with valid token
    $t->get_ok("/my/activation?nickname=$nickname&token=$token&cancel=1")
      ->status_is(200);
    diag $t->tx->res->body if !$t->success;

    # cancel succeeded?
    $t->get_ok("/my/activation?nickname=$nickname&token=$token")
      ->status_is(400);
    diag $t->tx->res->body if !$t->success;
};

$test_man->cleanup;
done_testing();
