# -*-Perl-*-
# Newslash::Model::Stories related  tests
use Mojo::Base -strict;

use FileHandle;
use Test::More;
use Test::Mojo;
use JSON;
use File::Temp;

use Data::Dumper;
use DateTime;

use Newslash::Model::Cowrapper;

my $t = Test::Mojo->new('Newslash::Web');
my $options = $t->app->config;
my $c = Newslash::Model::Cowrapper->new($options);


subtest 'check build_where_clause' => sub {
    my $params = { oerder_by => { points => "DESC" },
                   months => 12 };
    my $unique_keys = { id => "comments.cid",
                        comment_id => "comments.cid",
                        cid => "comments.cid",
                      };
    my $keys = { pid => "comments.pid",
                 discussion_id => "comments.sid",
                 pid => "comments.pid",
                 uid => "comments.uid",
                 user_id => "comments.uid",
                 points => "comments.points",
                 create_time => "comments.date",
                 update_time => "comments.last_update",
               };
    my $datetime_keys = { create_time => "comments.date",
                          update_time => "comments.last_update",
                        };
    my $timestamp = "comments.date";

    my ($where_clause, $where_values, $unique) = $c->build_where_clause(unique_keys => $unique_keys,
                                                                        keys => $keys,
                                                                        datetime_keys => $datetime_keys,
                                                                        timestamp => $timestamp,
                                                                        params => $params);
    ok($where_clause, "valid result");
    is($where_clause, "WHERE comments.date >= ? AND comments.date < ?", "valid where clause");
};

subtest 'export_json / import_json' => sub {
    my $dbh = $c->connect_db;

    # create test table
    my $create_sql = <<"EOSQL";
CREATE TABLE IF NOT EXISTS cow_test (
  id             int(8) unsigned AUTO_INCREMENT PRIMARY KEY ,
  foo             varchar(128)    NOT NULL UNIQUE,
  bar             smallint        NOT NULL DEFAULT 0,
  hoge            varchar(128)    NOT NULL DEFAULT ""
) DEFAULT CHARSET=utf8mb4
EOSQL
    my $rs = $dbh->do($create_sql, undef);
    ok($rs, "create test table");

    $rs = $dbh->do("DELETE FROM cow_test", undef);
    ok($rs, "initialize test table");

    # insert data
    my $test_data = [ { foo => "abcdefg", bar => 200, hoge => "hogehoge" },
                      { foo => "hijklmn", bar => 300, hoge => "hogehoge" },
                      { foo => "opqrstu", bar => 100, hoge => "hogehoge" },
                    ];
    my $insert_sql = <<"EOSQL";
INSERT INTO cow_test (foo, bar, hoge) VALUES (?, ?, ?)
EOSQL
    for my $d (@$test_data) {
        #my $rs = $dbh->do($insert_sql, undef, $d->{foo}, $d->{bar}, $d->{hoge});
        my $rs = $c->generic_insert(table => "cow_test",
                           keys => [qw(foo bar hoge)],
                           params => $d,
                          );
        ok($rs, "insert $d->{foo}");
    }

    # export test
    my $fname = File::Temp::tmpnam();
    $rs = $c->export_json(file => $fname,
                          table => "cow_test",
                          exclude => [qw(id hoge)],
                          sort_key => "bar",
                            );
    ok($rs, "execute export_json");
    my $json_ok = <<'EOL';
[
   {
      "bar": 100,
      "foo": "opqrstu"
   },
   {
      "bar": 200,
      "foo": "abcdefg"
   },
   {
      "bar": 300,
      "foo": "hijklmn"
   }
]
EOL
    # check file
    my $fh = FileHandle->new($fname, "r");
    ok($fh, "open test file");
    my $content = do { local $/; <$fh> };
    $fh->close;

    ok($content, "read from test file");
    is($json_ok, $content, "check test file's conent");
    ok(unlink($fname), "delete test file");

    # import test
    my $test_data2 = [ { foo => "abcdefg", bar => 200, hoge => "hogehoge" },
                       { foo => "hijklmn", bar => 300, hoge => "arege" },
                       { foo => "vwxyz",   bar => 400, hoge => "mogmog" },
                     ];
    my $ok_data = [ { foo => "abcdefg", bar => 200, hoge => "hogehoge" },
                    { foo => "hijklmn", bar => 300, hoge => "arege" },
                    { foo => "opqrstu", bar => 100, hoge => "hogehoge" },
                    { foo => "vwxyz",   bar => 400, hoge => "mogmog" },
                  ];
    my $test_json = JSON::to_json($test_data2);
    ok($test_json, "create test json data");
    $rs = $c->import_json(json => $test_json,
                          table => "cow_test",
                          unique_key => "foo",
                         );
    ok($rs, "execute import_json");
    my $updated_datas = $c->generic_select(table => "cow_test");
    ok($updated_datas, "select updated data");
    for my $d (@$updated_datas) {
        delete $d->{id};
    }
    is_deeply($updated_datas, $ok_data, "check updated data");

    # drop test table
    my $drop_sql = "DROP TABLE cow_test";
    $rs = $dbh->do($drop_sql, undef);
    ok($rs, "drop test table");

    $c->disconnect_db;
};

subtest 'calculate_time_range()' => sub {
    my ($from, $to);
    ($from, $to) = $c->calculate_time_range({year => 2010,
                                              month => 6,
                                              day => 25});
    is($from, "2010-06-25 00:00:00", "ymd1");
    is($to, "2010-06-26 00:00:00", "ymd2");

    ($from, $to) = $c->calculate_time_range({year => 2010,
                                              month => 6,
                                              day => 25,
                                              months => 1,
                                             });
    is($from, "2010-06-25 00:00:00", "ymd+m1");
    is($to, "2010-07-25 00:00:00", "ymd+m2");

    ($from, $to) = $c->calculate_time_range({date => "today"});
    ok($from, "today1");
    ok($to, "today2");

    ($from, $to) = $c->calculate_time_range({date => "2016-06-25"});
    is($from, "2016-06-25 00:00:00", "date1");
    is($to, "2016-06-26 00:00:00", "date2");

    ($from, $to) = $c->calculate_time_range({date => "2016-06-25",
                                              days => 3,
                                             });
    is($from, "2016-06-25 00:00:00", "date+d1");
    is($to, "2016-06-28 00:00:00", "date+d2");


    ($from, $to) = $c->calculate_time_range({until => "2016-03-02"});
    ok(!$from, "until1");
    is($to, "2016-03-02 00:00:00", "until2");

    ($from, $to) = $c->calculate_time_range({since => "2016-03-02"});
    is($from, "2016-03-02 00:00:00", "since1");
    ok(!$to, "since2");

    ($from, $to) = $c->calculate_time_range({since => "2016-03-02",
                                              until => "2016-04-01",
                                             });
    is($from, "2016-03-02 00:00:00", "since+until1");
    is($to, "2016-04-01 00:00:00", "since+until2");

    ($from, $to) = $c->calculate_time_range({month => 12,
                                              day => 1,
                                             });
    ok(!$from, "month+day");
    ok(!$to, "month*day");
};


done_testing();
