# -*-Perl-*-
# Newslash::Model::Journals related  tests
use Mojo::Base -strict;

use Test::More;
use Test::Mojo;
use Data::Dumper;
use DateTime;

use Newslash::Util::TestMan;

my $t = Test::Mojo->new('Newslash::Web');
my $journals = $t->app->model('journals');
ok($journals, "get Journals instance");

my ($admin, $user);
my $test_man = Newslash::Util::TestMan->new($t);

# prepare for test
if ($t->app->mode eq 'test') {
    #$admin = $test_man->create_admin("polltestadmin");
    #ok($admin, "create admin user");
    $user = $test_man->create_user("journaltest");
    ok($user, "create test user");
}

subtest 'create/select/update/latest journals' => sub {
    plan skip_all => "mode is not 'test'" if ($t->app->mode ne 'test');

    # create
    my $params = { description => "タイトル",
                   article => "本文",
                   commentstatus => "enabled",
                   posttype => 2,
                   formatter => "modern",
                 };
    my $journal_id = $journals->create(user => $user, %$params);
    ok($journal_id, "create journal");

    # select
    my $journal = $journals->select(journal_id => $journal_id);
    ok($journal, "select journal by id");
    is($params->{description}, $journal->{title}, "valid title");

    # update title
    ok($journals->update(description => 'タイトル2',
                         journal_id => $journal_id,
                         user => $user,
                        ), "update title");
    $journal = $journals->select(journal_id => $journal_id);
    is($journal->{description}, 'タイトル2', "title updated");

    # update body
    ok($journals->update(journal_id => $journal_id,
                         user => $user,
                         article => "ほげほげ",
                         posttype => 2,
                         formatter => "modern",
                        ), "update title");
    $journal = $journals->select(journal_id => $journal_id);
    is($journal->{description}, 'タイトル2', "title updated");

    # latest
    my $items = $journals->select(limit => 10);
    ok($items, "select returns valid value");
    ok(@$items, "select returns valid items");

    my $dt = DateTime->now;
    #my $offset = 9 * 60 * 60;
    my $offset = 0;

    # multiple select
    $items = $journals->select(year => $dt->year,
                               month => $dt->month,
                               day => $dt->day,
                               offset_sec => $offset,
                               order_by => { time => "DESC" }
                              );
    ok($items, "select multiple journals");
    ok(@$items, "select one more journals");

    # count
    my $c = $journals->count(year => $dt->year,
                         month => $dt->month,
                         day => $dt->day,
                         offset_sec => $offset);
    ok($c, "get count of the day");
    ok($c > 0, "valid count of the day");

    $c = $journals->count(year => $dt->year,
                         month => $dt->month,
                         offset_sec => $offset);
    ok($c, "get count of the month");
    ok($c->{$dt->day}, "valid count of the month");


    $c = $journals->count(year => $dt->year,
                         offset_sec => $offset);
    ok($c, "get count of the year");
    ok($c->{$dt->month}, "valid count of the year");


    # delete
    ok($journals->hard_delete(journal_id => $journal_id), "delete story");
};

subtest 'convert functions' => sub {
# posttype: 1: PLAINTEXT, 2: HTML, 3: EXTRANS, 4: CODE, 77: FULLHTML
    my $posttype = 1;

    my $article = "日記テスト本文\nてすとてすと";

    my $introtext = $journals->article_to_introtext($article, $posttype);
    unlike($introtext, qr/\A\s*\z/, "article_to_introtext returns valid");
    #diag("introtext is $introtext");

    my $bodytext = $journals->article_to_bodytext($article, $posttype);
    is(length $bodytext, 0, "article_to_bodytext returns valid");
    #diag("bodytext is $bodytext");

    my $fulltext = $journals->article_to_fulltext($article, $posttype);
    unlike($fulltext, qr/\A\s*\z/, "article_to_fulltext returns valid");
    #diag("fulltext is $fulltext");


    $article = <<"EOT";
東芝が2013年7月から2015年8月までに発売されたノートPC45機種で使われているバッテリーの一部について、使い方によっては過熱してショートし発火する恐れがあるとし、新品への無償交換を行うことを発表した（<a href="http://www3.nhk.or.jp/news/html/20161110/k10010763661000.html">NHK</a>、<a href="http://pc.watch.impress.co.jp/docs/news/1029272.html">PC Watch</a>）。

東芝のノートPCについては、<a href="http://hardware.srad.jp/story/14/11/14/2013216/">2014年</a>および<a href="https://mobile.srad.jp/story/16/01/29/2321242/">今年1月</a>にバッテリーの過熱・発煙・発火の恐れがあるとしてバッテリーのリコールが行われているが、発見されたのはこれとは別の不具合とのこと（<a href="https://batterycheck.toshiba.com/BatteryUpdate/InformationJapan?region=TJPN&amp;country=JP&amp;lang=ja">東芝の発表</a>）。そのため、以前問題ないとされたPCについても、再度確認が必要とのこと。

EOT

    $introtext = $journals->article_to_introtext($article, $posttype);
    unlike($introtext, qr/\A\s*\z/, "article_to_introtext returns valid 2");
    #diag("introtext is $introtext");

    $bodytext = $journals->article_to_bodytext($article, $posttype);
    unlike($bodytext, qr/\A\s*\z/, "article_to_bodytext returns valid 2");
    #diag("bodytext is $bodytext");

    $fulltext = $journals->article_to_fulltext($article, $posttype);
    unlike($fulltext, qr/\A\s*\z/, "article_to_fulltext returns valid 2");
    #diag("fulltext is $fulltext");
};

$test_man->cleanup;
done_testing();
