# -*-Perl-*-
# Newslash::Model::Stories related  tests
use Mojo::Base -strict;

use Test::More;
use Test::Mojo;
use Data::Dumper;
use DateTime;

use Newslash::Util::TestMan;

my $t = Test::Mojo->new('Newslash::Web');
my $stories = $t->app->model('stories');
ok($stories, "get Stories instance");

my ($admin, $user);
my $test_man = Newslash::Util::TestMan->new($t);

# prepare for test
if ($t->app->mode eq 'test') {
    # at first, create test users.
    $admin = $test_man->create_admin("storytestadmin");
    ok($admin, "create admin user");
}

subtest 'create/select/update/latest stories' => sub {
    plan skip_all => "mode is not 'test'" if ($t->app->mode ne 'test');

    my $slash_db = Newslash::Model::SlashDB->new($stories->{options});

    my $params = {
                  user => $admin,
                  title => "test story",
                  introtext => '<p>foo bar hoge</p>',
                  bodytext => '',
                  topics_chosen => {49 => 10, 2271 => 20},
                  commentstatus => '',
                  tid => 49,
                  time => "2010-11-11 10:00:00"
                 };

    #warn Dumper $slash_db->getSkins;

    my $stoid = $stories->create(%$params);
    ok($stoid, "create story");
    diag($stories->last_error) if !$stoid;

    #warn Dumper $slash_db->getSkins;

    my $new_text = "<p>foo bar hoge hoge</p>";
    $params = {};
    $params->{introtext} = $new_text;
    $params->{bodytext} = "";
    $params->{topics_chosen} = {49 => 10, 2271 => 20};
    $params->{story_id} = $stoid;

    #warn Dumper $slash_db->getSkins;

    my $rs = $stories->update(%$params);

    #warn Dumper $slash_db->getSkins;

    ok($rs, "update story");
    diag $stories->last_error if !$rs;

    my $story = $stories->select(stoid => $stoid);
    ok($story, "select story by stoid");
    is($story->{introtext}, $new_text, "update story correctly");


    # set/delete related
    $rs = $stories->add_related_story(stoid => $stoid,
                                      related_stoid => $stoid);
    ok($rs, "add related story");
    my $related = $stories->get_related_items(stoid => $stoid);
    ok($related && @$related, "get related items");
    is($related->[0]->{stoid}, $stoid, "get correct related items");

    my $latest = $stories->select(limit => 10, order_by => {create_time => "DESC"});
    ok($latest, "retrive latest stories");
    ok(@$latest > 0, "retrive 1 or more stories");

    my $dt = DateTime->now();
    my $c;
    my $year = 2010;
    my $month = 11;
    my $day = 11;
    my $offset = $admin->{config}->{offset_sec} || 0;

    $c = $stories->count(year => $year,
                         month => $month,
                         day => $day,
                         offset_sec => $offset);
    ok($c, "get count of the day");
    ok($c > 0, "valid count of the day");

    $c = $stories->count(year => $year,
                         month => $month,
                         offset_sec => $offset);
    ok($c, "get count of the month");
    ok($c->{$day}, "valid count of the month");

    $c = $stories->count(year => $year,
                         offset_sec => $offset);
    ok($c, "get count of the year");
    ok($c->{$month}, "valid count of the year");

    # test update2
    my $hits = $story->{hits};
    $rs = $stories->update2(stoid => $story->{stoid}, hits => { add => 1 });
    ok($rs, "update hits by update2");
    $story = $stories->select(stoid => $story->{stoid});
    is($story->{hits}, $hits + 1, "update hits succeeded");

    ok($stories->hard_delete(story_id => $stoid), "delete story");

};

subtest 'create stories from submission' => sub {
    my $submission = {};
    my $params = {
	title => "test story 2",
	url => "http://example.com/",
	email => 'foo@example.com',
	introtext => '<p>foo bar hoge</p>',
	tags_string => "internet foo bar", # tid is 1913
	tid => 1913,
    };
    my $subm = $test_man->create_submission(%$params);
    
    $params->{submission_id} = $subm->{subid};
    $params->{user} = $admin;
    $params->{topics_chosen} = {49 => 10, 2271 => 20};
    $params->{time} = "2010-11-11 10:00:00";

    my ($sid, $stoid) = $stories->create(user => $user, %$params);
    ok($stoid, "create story from submission");
    diag $stories->last_error if !$stoid;

    my $subm2 = $t->app->model('submissions')->select(id => $subm->{subid});
    ok($subm2, "get updated submission");
    is($subm2->{del}, 2, "submission accepted");

    my $story2 = $stories->select(stoid => $stoid);
    ok($story2, "get created story");
    is($story2->{submitter}, 1, "story has valid submittor");

    my $related = $stories->get_related_items(stoid => $stoid);
    ok($related && @$related, "get related items");

    my $fh_related;
    for my $item (@$related) {
	if (!$item->{rel_stoid} && !$item->{rel_sid}) {
	    $fh_related = $item;
	    last;
	}
    }
    ok($fh_related, "correct fh_item inserted");
    is($fh_related->{title}, "Firehose: " . $params->{title}, "fh_item title ok");

    ok($stories->hard_delete(story_id => $stoid), "delete story");
};

$test_man->cleanup;

done_testing();
