# -*-Perl-*-
# Model::Users tests
use Mojo::Base -strict;

use Test::More;
use Test::Mojo;
use Data::Dumper;

use Newslash::Util::TestMan;

my $t = Test::Mojo->new('Newslash::Web');
my $users = $t->app->model('users');
ok($users, "get Users instance");

my $test_man = Newslash::Util::TestMan->new($t);
my $uid;
my $admin;

subtest 'create users' => sub {
    plan skip_all => "mode is not 'test'" if ($t->app->mode ne 'test');

    my $user = $test_man->create_user("testuser");
    ok($user, "create test user");
    $uid = $user->{uid};

    $admin = $test_man->create_user("testadmin");
    ok($user, "create admin user");

};

subtest 'select/update users' => sub {
    plan skip_all => "mode is not 'test'" if ($t->app->mode ne 'test');

    my $user = $users->select(uid => $uid);
    ok($user, "select method returns valid value");

    my $rs = $users->update(uid => $user->{uid},
                            seclev => 10000,
                            author => 1);
    ok($rs, "update users table");

    diag $users->last_error if (!$rs);

    $user = $users->select(uid => $uid);
    ok($user->{uid} == $uid, "select method returns valid uid");
    ok($user->{seclev} == 10000, "update seclv");
    ok($user->{author} ==1, "update author");

    $user = $users->select(nickname => "testuser");
    ok($user, "select by nickname returns valid value");
    ok($user->{uid} == $uid, "select by nickname method returns valid uid");

    my $anon = $users->anonymous_user;
    ok($anon, "anonymous_user method not returns undefined");
    ok($anon->{uid} == 1, "anonmous_user's uid is 1");
    ok($anon->{seclev} == 0, "anonymous_user's sevlev is 0");

};

subtest 'acl2' => sub {
    plan skip_all => "mode is not 'test'" if ($t->app->mode ne 'test');

    my $acl2_types = $users->acl2->types;
    ok($acl2_types, "acl2_types");
    is($acl2_types->{spammer}->{al2tid}, 11, "valid tid of acl2_types");

    my $rs = $users->acl2->update(uid => $uid,
                                  admin_uid => $admin->{uid},
                                  types => [qw(nopost norss)],
                                  comment => "testing...");
    ok($rs, "update_acl2");

    $rs = $users->acl2->select(uid => $uid);
    ok($rs, "select: target => acl2");
    is(@{$rs->{values}}, 2, "valid numbers of acl2 set");
    is($rs->{value}, ((1 << 2) + (1 << 4)), "valid value of acl2 set");

    $rs = $users->acl2->update(uid => $uid,
                               admin_uid => $admin->{uid},
                               types => [qw(spammer)],
                               comment => "testing 2");
    ok($rs, "update_acl2 (2)");

    $rs = $users->acl2->select(uid => $uid);
    ok($rs, "select: target => acl2 (2)");
    is(@{$rs->{values}}, 1, "valid numbers of acl2 set (2)");
    is($rs->{value}, (1 << 9), "valid value of acl2 set (2)");

    $rs = $users->acl2->update(uid => $uid,
                               admin_uid => $admin->{uid},
                               types => [],
                               comment => "testing 3");
    ok($rs, "update_acl2 (3)");

    $rs = $users->acl2->select(uid => $uid);
    ok($rs, "select: target => acl2 (3)");
    is(@{$rs->{values}}, 0, "valid numbers of acl2 set (3)");
    is($rs->{value}, 0, "valid value of acl2 set (3)");

    $rs = $users->acl2->select_log(uid => $uid);
    ok($rs, "select: target => acl2_log");
    ok(@$rs > 0, "valid number of acl2_log");
    my $log = pop @$rs;
    is($log->{comment}, "testing 3", "valid comment of acl2_log");

    my $user = $users->select(uid => $uid);

    my $srcid24 = $test_man->srcid("192.168.1.1", 24);
    ok($srcid24, "get srcid24");

    $rs = $users->acl2->update(src_id => $srcid24,
                               admin_uid => $admin->{uid},
                               types => [qw(spammer)],
                               comment => "testing 4");
    ok($rs, "update_acl2_by_srcid");

    $rs = $users->acl2->select(srcid => $srcid24);
    ok($rs, "select by srcid => acl2");
    is($rs->{value}, (1 << 9), "valid value of acl2 set by srcid");

    $rs = $users->acl2->update(src_id => $srcid24,
                               admin_uid => $admin->{uid},
                               type => [],
                               comment => "testing 5");
    ok($rs, "update_acl2_by_srcid (2)");

    $rs = $users->acl2->hard_delete(src_id => $srcid24);
    ok($rs, "delete acl2 by srcid");

    $rs = $users->acl2->hard_delete(uid => $uid);
    ok($rs, "delete acl2 by uid");
};

subtest 'sidebar' => sub {
    my $rs = $users->sidebar->select(uid => $uid);
    ok(defined $rs, "select sidebar");
    my $boxes = $users->sidebar->get_sidebar_contents(uid => $uid);
    ok(defined $boxes, "get sidebar contents");

};

subtest 'messages' => sub {
    my $messages = $users->messages;
    my $rs = $messages->set(uid => $uid, code => 4, mode => 1);
    ok(defined $rs, "set message settings");

    my $settings = $messages->select(uid => $uid);
    ok($settings, "retrice message settings");
    my $setting;
    for my $s (@$settings) {
        if ($s->{code} == 4) {
            $setting = $s;
            last;
        }
    }
    ok($setting, "setting exists");
    is($setting->{mode}, 1, "setting set correctly");

    $setting = $messages->select(uid => $uid, code => 4);
    ok($setting, "select a message setting");
    is($setting->{code}, 4, "select a message code correctly");
    is($setting->{mode}, 1, "select a message mode correctly");

    $rs = $messages->set(uid => $uid, code => 4, mode => 0);
    ok(defined $rs, "update message settings");
    diag $messages->last_error if !$rs;

    $setting = $messages->select(uid => $uid, code => 4);
    is($setting->{code}, 4, "select a updated message setting correctly");
    is($setting->{mode}, 0, "setting update correctly");

    $rs = $messages->set(uid => $uid, code => 4, mode => -1);
    ok($rs, "delete message settings");
    diag $messages->last_error if !$rs;

    $setting = $messages->select(uid => $uid, code => 4);
    is($setting->{mode}, -1, "setting delete correctly");
};

subtest 'param' => sub {
    my $param = $users->param;
    
    # select test
    my $params = $param->select(uid => $uid);
    ok($params, "select param");

    # insert test
    my $rs = $param->insert(uid => $uid,
                            name => "test_param",
                            value => "this_is_test_param");
    ok($rs, "insert param");
    diag $param->last_error if !$rs;
    $params = $param->select(uid => $uid);
    is($params->{test_param}, "this_is_test_param", "insert param correctly");


    # update test
    $rs = $param->update(uid => $uid,
                         name => "test_param",
                         value => "updated_param");
    ok($rs, "update param");
    diag $param->last_error if !$rs;
    $params = $param->select(uid => $uid);
    is($params->{test_param}, "updated_param", "update param correctly");

    # set test (update)
    $rs = $param->set(uid => $uid,
                      name => "test_param",
                      value => "updated_param2");
    ok($rs, "set param1");
    diag $param->last_error if !$rs;
    $params = $param->select(uid => $uid);
    is($params->{test_param}, "updated_param2", "set param (update) correctly");

    # set test (insert)
    $rs = $param->set(uid => $uid,
                      name => "test_param2",
                      value => "test_param2");
    ok($rs, "set param2");
    diag $param->last_error if !$rs;
    $params = $param->select(uid => $uid);
    is($params->{test_param2}, "test_param2", "set param (insert) correctly");

    # delete test
    $rs = $param->delete(uid => $uid,
                         name => "test_param");
    ok($rs, "delete param");
    diag $param->last_error if !$rs;
    $params = $param->select(uid => $uid);
    ok(!defined $params->{test_param}, "delete param correctly");
};

subtest 'achievements' => sub {
    my $rs = $users->achievements->select(uid => $uid);
    ok($rs, "select achievements");
};

subtest 'delete users' => sub {
    plan skip_all => "mode is not 'test'" if ($t->app->mode ne 'test');

    my $ret = $users->hard_delete($uid);
    ok($ret, "delete user");


};

$test_man->cleanup;
done_testing();
