# -*-Perl-*-
# users plugin tests

use Mojo::Base -strict;
use Test::More;
use Test::Mojo;

use Newslash::Util::TestMan;
use Data::Dumper;
use Newslash::Eventd::Handlers;

my $t = Test::Mojo->new('Newslash::Web');
my $test_man = Newslash::Util::TestMan->new($t);
my $test_nickname = "userstest";
my $default_passwd = "foobar_usertest";
my $user;

# prepare for test
if ($t->app->mode eq 'test') {
    # at first, create test users.
    $user = $test_man->create_user($test_nickname, $default_passwd);
}

subtest 'change email related test' => sub {
    my $app = $t->app;
    if ($app->mode ne 'test') {
        fail("mode is not 'test'");
        return;
    }

    # parameter check
    my $bad_address = "foobar";
    my $new_address = 'foobarhogehoge@example.com';
    my $users = $app->model('users');

    ok(!$app->users->request_change_email($user, $bad_address),
       "request_change_email() with bad address");
    is($app->users->last_error, "INVALID_ADDRESS");

    ok(!$app->users->request_change_email(0, $new_address),
       "request_change_email() with bad user");
    is($app->users->last_error, "INVALID_USER");

    my $anonymous = $users->anonymous_user;
    ok(!$app->users->request_change_email($anonymous, $new_address),
       "request_change_email() with anonymous user");
    is($app->users->last_error, "INVALID_USER");

    ok(!$app->users->request_change_email($user, $user->{realemail}),
       "request_change_email() with current address");
    is($app->users->last_error, "ADDRESS_EXISTS");

    # call function
    ok($app->users->request_change_email($user, $new_address),
       "execute request_change_email()") or diag $app->users->last_error;

    # check if database updated
    my $params = $users->param->select(uid => $user->{uid});
    is($params->{new_email}, $new_address, "insert new address to params table");

    # check if event emited
    my $events = $app->model('events');
    my $ev = $events->select(event_type => "user",
                             event_subtype => "change_email",
                             source_user_id => $user->{uid},
                             active => 1,
                            );
    ok($ev && @$ev > 0, "event emited");

    # check if event processed correctly
    my $handlers = Newslash::Eventd::Handlers->new($app, { test => 1 });
    my $rs = $handlers->execute($ev->[0]);
    ok($rs, "event processed correctly");

    # check email que
    my $user_handler = $handlers->{_handlers}->{user};
    ok($user_handler, "retrieve user event handler");
    my $que = $user_handler->{_test_mailque};

    # check validate mail
    my $validate_mail = pop @$que;
    ok($validate_mail, "retrieve validate mail");
    is($validate_mail->{to}, $new_address, "validate mail address is correct");
    my $url_regex = qr|/my/change_email\?token=(\S+)|;
    ok($validate_mail->{body} =~ $url_regex, "validate mail has correct url");
    my $token = $1;
    ok($token, "extract token from mail");
    diag Dumper $validate_mail if !$1;

    # check notify mail
    my $notify_mail = pop @$que;
    ok($notify_mail, "retrieve notify mail");
    is($notify_mail->{to}, $user->{realemail}, "notify mail address is correct");
    my $notify_url = "/my/change_email?cancel=1";
    my $index = index($notify_mail->{body}, $notify_url);
    ok($index != -1, "notify mail has correct url");
    diag Dumper $notify_mail if $index == 1;

    # check ts param set
    $params = $users->param->select(uid => $user->{uid});
    my $dt = eval { DateTime::Format::MySQL->parse_datetime($params->{new_email_token_ts}); };
    my $e = $@;
    ok(!$e, "check new_email_ts");
    diag $e if $e;
    diag Dumper $params if $e;

    if (!$e) {
        my $now_dt = DateTime->now;
        my $delta = $now_dt->epoch - $dt->epoch;
        ok($delta < 3600, "new_email_ts is updated");
    }

    # cancel request
    ok(!$app->users->cancel_change_email_request(0),
       "cancel_change_email_request() with invalid uid");
    is($app->users->last_error, "INVALID_UID");

    ok(!$app->users->cancel_change_email_request($anonymous->{uid}),
       "cancel_change_email_request() with anonymous uid");
    is($app->users->last_error, "NO_ADDRESS");

    ok($app->users->cancel_change_email_request($user->{uid}),
       "cancel_change_email_request()");

    # check if update done correctly
    my $u = $users->select(uid => $user->{uid});
    isnt($u->{realemail}, $new_address, "realemail didn't update");
    $params = $users->param->select(uid => $u->{uid});
    ok(!$params->{new_email}, "new_email cleared");
    ok(!$params->{new_email_token}, "new_email_token cleared");
    ok(!$params->{new_email_token_ts}, "new_email_token_ts cleared");

    # re-request function
    # use users_param table to save temporary new email address
    ok($users->param->set(uid => $user->{uid},
                          name => "new_email",
                          value => $new_address),
       "set new_email parameter");
    $token = $app->users->generate_email_change_token($user->{uid});
    ok($token, "generate_email_change_token()");

    # change email
    # parameter check
    ok(!$app->users->verify_and_change_email(0, $token),
       "verify_and_change_email() with invalid user");
    is($app->users->last_error, "INVALID_TOKEN");

    ok(!$app->users->verify_and_change_email($user->{uid}, "hogehoge"),
       "verify_and_change_email() with invalid token");
    is($app->users->last_error, "INVALID_TOKEN");

    # TODO: check if token is expired

    # execute with correct arguments
    ok($app->users->verify_and_change_email($user->{uid}, $token),
       "verify_and_change_email()");

    # check if update done correctly
    $u = $users->select(uid => $user->{uid});
    is($u->{realemail}, $new_address, "realemail updated correctly");
    $params = $users->param->select(uid => $u->{uid});
    ok(!$params->{new_email}, "new_email cleared");
    ok(!$params->{new_email_token}, "new_email_token cleared");
    ok(!$params->{new_email_token_ts}, "new_email_token_ts cleared");

};

subtest 'reset password related test' => sub {
    my $app = $t->app;
    if ($app->mode ne 'test') {
        fail("mode is not 'test'");
        return;
    }

    # parameter check
    my $bad_address = "foobar";
    my $new_address = 'foobarhogehoge@example.com';
    my $users = $app->model('users');
    my $anonymous = $users->anonymous_user;

    ok(!$app->users->reset_password($anonymous),
       "reset_password() with invalid user");
    is($app->users->last_error, "INVALID_USER");

    # execute with correct arguments
    ok($app->users->reset_password($user),
       "reset_password()") or diag $app->users->last_error;

    # check if event emited
    my $events = $app->model('events');
    my $ev = $events->select(event_type => "user",
                             event_subtype => "reset_password",
                             source_user_id => $user->{uid},
                             active => 1,
                            );
    ok($ev && @$ev > 0, "event emited");

    # check if event processed correctly
    ## generate_password_reset_token() called.
    my $handlers = Newslash::Eventd::Handlers->new($app, { test => 1 });
    my $rs = $handlers->execute($ev->[0]);
    ok($rs, "event processed correctly");

    # check if update done correctly
    my $u = $users->select(uid => $user->{uid});
    isnt($u->{newpassswd}, "", "newpassword updated correctly");
    isnt($u->{newpassswd_ts}, "", "newpassword_ts updated correctly");

    # check newpasswd_ts param set
    my $dt = eval { DateTime::Format::MySQL->parse_datetime($u->{newpasswd_ts}); };
    my $e = $@;
    ok(!$e, "check newpasswd_ts");
    if (!$e) {
        my $now_dt = DateTime->now;
        my $delta = $now_dt->epoch - $dt->epoch;
        ok($delta < 3600, "newpasswd_ts is updated");
    }
    else {
        diag $e;
        diag Dumper $u;
    }

    # check email que
    my $user_handler = $handlers->{_handlers}->{user};
    ok($user_handler, "retrieve user event handler");
    my $que = $user_handler->{_test_mailque};

    # check validate mail
    my $validate_mail = pop @$que;
    ok($validate_mail, "retrieve validate mail");
    is($validate_mail->{to}, $u->{realemail}, "validate mail address is correct");
    my $url_regex = qr|/my/activation\?nickname=([^&]+)&token=(\S+)|;
    ok($validate_mail->{body} =~ $url_regex, "validate mail has correct url");
    my $nickname = $1;
    my $token = $2;
    ok($nickname, "extract nickname from mail");
    ok($token, "extract token from mail");
    diag Dumper $validate_mail if !$1;

    # cancel request
    ok(!$app->users->cancel_activation(0),
       "cancel_activation() with invalid uid");
    is($app->users->last_error, "INVALID_USER");

    ok($app->users->cancel_activation($user),
       "cancel_activation()") or diag $app->users->last_error;

    # check if update done correctly
    $u = $users->select(uid => $user->{uid});
    ok(!$u->{newpassswd}, "newpassword updated correctly");
    ok(!$u->{newpassswd_ts}, "newpassword_ts updated correctly");

    # re-request function
    $token = $app->users->generate_password_reset_token($user->{uid});
    ok($token, "generate_password_reset_token()");

    # update password
    my $new_password = "foobarhogehoge";
    ok(!$app->users->update_password_by_token($u->{nickname}, "hogehoge", $new_password),
       "update_password_by_token() with invalid token");
    is($app->users->last_error, "INVALID_TOKEN");
    # TODO: check expired case

    ok($app->users->update_password_by_token($u->{nickname}, $token, $new_password),
       "update_password_by_token()");

    $u = $users->select(uid => $user->{uid});
    ok($users->passwords->verify_password(user => $u, password => $new_password),
       "password update correctly");
    ok(!$u->{newpassswd}, "newpassword updated correctly");
    ok(!$u->{newpassswd_ts}, "newpassword_ts updated correctly");


};


subtest 'create new user related test' => sub {
    my $app = $t->app;
    if ($app->mode ne 'test') {
        fail("mode is not 'test'");
        return;
    }

    my $new_address = 'hjkhkjadsf@example.com';
    my $new_nickname = 'newusertest00';
    my $users = $app->model('users');
    my $u = $users->select(nickname => $user->{nickname});
    ok($u, "retrieve registered user");

    # validate new user information - nickname
    ok(!$app->users->validate_nickname($u->{nickname}),
       "validate_nickname() with exists nickname");
    is($app->users->last_error, "ID_EXISTS");

    ok(!$app->users->validate_nickname(""),
       "validate_nickname() with blank nickname");
    is($app->users->last_error, "BLANK_ID");

    ok(!$app->users->validate_nickname("[][]"),
       "validate_nickname() with invalid nickname");
    is($app->users->last_error, "INVALID_ID");

    # validate new user information - email
    ok(!$app->users->validate_email($u->{realemail}),
       "validate_email() with invalid email");
    is($app->users->last_error, "ADDRESS_EXISTS");

    ok(!$app->users->validate_email(""),
       "validate_email() with blank email");
    is($app->users->last_error, "BLANK_ADDRESS");

    ok(!$app->users->validate_email("foobar"),
       "validate_email() with invalid email");
    is($app->users->last_error, "INVALID_ADDRESS");

    # create new user
    my $uid = $app->users->create_new_user($new_nickname, $new_address);
    ok($uid, "create_new_user") or diag explain $app->users->last_error;

    # check if database updated
    $u = $users->select(uid => $uid);
    ok($u, "uid exists") or diag $users->last_error;
    is($u->{nickname}, $new_nickname, "correct nickname");
    is($u->{realemail}, $new_address, "correct email");

    # check if event emited
    ok($test_man->process_event(event_type => "user",
                                event_subtype => "create",
                                source_user_id => $u->{uid}),
       "process event");

    # check validate mail
    my $mail = $test_man->pop_mailque()
      ->to_is($new_address)
      ->body_matches(qr|/my/activation\?nickname=(?<nickname>[^&]+)&token=(?<token>\S+)|);
    my $nickname = $mail->last_match->{nickname};
    my $token = $mail->last_match->{token};
    is($nickname, $new_nickname) or diag $mail->body;
    ok($token, "extract token from mail") or diag $mail->body;

    # activate
    ok(!$app->users->check_activation_token("1234", $token),
       "activation() with invalid nickname");
    is($app->users->last_error, "INVALID_TOKEN");

    ok(!$app->users->check_activation_token($nickname, "hogehoge"),
       "activation() with invalid tomen");
    is($app->users->last_error, "INVALID_TOKEN");

    ok($app->users->check_activation_token($nickname, $token),
       "activation()") or diag $app->users->last_error;

    my $new_password = "ThisIsNewPassword";
    ok($app->users->update_password_by_token($nickname, $token, $new_password),
       "update_password_by_token()") or diag $app->users->last_error;

    # verify password
    ok($users->passwords->verify_password(uid => $u->{uid},
                                          password => $new_password),
       "verify password") or diag $users->passwords->last_error;

    # done, delete created user
    ok($test_man->delete_user($u), "delete test user");

};

$test_man->cleanup;
done_testing();
