# coding: utf-8
'''stories.py: newslash journals related database module'''

import random
from base import NewslashDB, NewslashDBError

CREATE_JOURNALS_TABLE_SQL = '''
CREATE TABLE IF NOT EXISTS `journals` (
  `id`          mediumint(8) unsigned              NOT NULL AUTO_INCREMENT,
  `uid`         mediumint(8) unsigned              NOT NULL DEFAULT '0',
  `date`        datetime                           NOT NULL DEFAULT '1970-01-01 00:00:00',
  `description` varchar(80)                        NOT NULL DEFAULT '',
  `posttype`    tinyint(4)                         NOT NULL DEFAULT '2',
  `discussion`  mediumint(8) unsigned              DEFAULT NULL,
  `tid`         smallint(5) unsigned               NOT NULL DEFAULT '0',
  `last_update` timestamp                          NOT NULL DEFAULT CURRENT_TIMESTAMP ON UPDATE CURRENT_TIMESTAMP,
  `promotetype` enum('publicize','publish','post') NOT NULL DEFAULT 'publish',
  `srcid_32`    bigint(20) unsigned                NOT NULL DEFAULT '0',
  `srcid_24`    bigint(20) unsigned                NOT NULL DEFAULT '0',
  PRIMARY KEY       (`id`),
  KEY `uidarticle`  (`uid`),
  KEY `IDandUID`    (`id`,`uid`),
  KEY `tid`         (`tid`),
  KEY `discussion`  (`discussion`),
  KEY `uid_date_id` (`uid`,`date`,`id`),
  KEY `srcid_32`    (`srcid_32`),
  KEY `srcid_24`    (`srcid_24`),
  KEY `id_date`     (`id`,`date`),
  KEY `date`        (`date`)
) ENGINE=InnoDB DEFAULT CHARSET=utf8mb4 ROW_FORMAT=DYNAMIC
'''

CREATE_JOURNALS_TEXT_TABLE_SQL = '''
CREATE TABLE IF NOT EXISTS `journals_text` (
  `id`        mediumint(8) unsigned NOT NULL DEFAULT '0',
  `article`   mediumtext            NOT NULL,
  `introtext` mediumtext            NOT NULL,
  PRIMARY KEY (`id`)
) ENGINE=InnoDB DEFAULT CHARSET=utf8mb4 ROW_FORMAT=DYNAMIC
'''

class Journals(NewslashDB):
    def __init__(self, config):
        self.db = NewslashDB(config)

    def select(self, **kwargs):
        if not "limit" in kwargs:
            kwargs["limit"] = self.db.config("default_limit")

        if not "offset" in kwargs:
            kwargs["offset"] = 0

        if "id" in kwargs:
            query = ("SELECT journals.*, journals_text.article, journals_text.introtext, firehose.public FROM journals"
                     "  LEFT JOIN journals_text USING(id)"
                     "  LEFT JOIN firehose ON firehose.type = 'journal' AND firehose.srcid = journals.id"
                     "  WHERE journals.id = %(id)s"
                     "  LIMIT %(limit)s OFFSET %(offset)s")
        elif "last_update_ge" in kwargs:
            query = ("SELECT journals.*, journals_text.article, journals_text.introtext, firehose.public FROM journals"
                     "  LEFT JOIN journals_text USING(id)"
                     "  LEFT JOIN firehose ON firehose.type = 'journal' AND firehose.srcid = journals.id"
                     "  WHERE journals.id <= %(id_le)s AND journals.last_update >= %(last_update_ge)s"
                     "  ORDER BY journals.id ASC LIMIT %(limit)s OFFSET %(offset)s")
        elif "id_gt" in kwargs:
            query = ("SELECT journals.*, journals_text.article, journals_text.introtext, firehose.public FROM journals"
                     "  LEFT JOIN journals_text USING(id)"
                     "  LEFT JOIN firehose ON firehose.type = 'journal' AND firehose.srcid = journals.id"
                     "  WHERE journals.id > %(id_gt)s"
                     "  ORDER BY journals.id ASC LIMIT %(limit)s OFFSET %(offset)s")
        else:
            query = ("SELECT journals.*, journals_text.*, firehose.public FROM journals"
                     "  LEFT JOIN journals_text USING(id)"
                     "  LEFT JOIN firehose ON firehose.type = 'journal' AND firehose.srcid = journals.id"
                     "  ORDER BY journals.id DESC LIMIT %(limit)s OFFSET %(offset)s")

        cur = self.db.execute(query, **kwargs)
        col_names = cur.column_names
        result = [dict(zip(col_names, x)) for x in cur]

        self.db.close()
        return result

    def create_tables(self):
        '''this function implemented for test purpose...'''
        self.db.execute_multi((CREATE_JOURNALS_TABLE_SQL,
                               CREATE_JOURNALS_TEXT_TABLE_SQL))

    def drop_tables(self):
        '''this function implemented for test purpose...'''
        self.db.execute_multi(('DROP TABLE journals',
                               'DROP TABLE journals_text'))

    def insert(self, **kwargs):
        '''this function implemented for test purpose...'''
        db = self.db.start_transaction()
        kwargs["globjid"] = random.randrange(1, 60000)
        kwargs["type"] = "journal";
        kwargs["name"] = "foobar";
        kwargs["sprite_info"] = "";
        try:
            cur = db.generic_insert("journals", ("description", "uid"), kwargs)
            kwargs["id"] = cur.lastrowid
            kwargs["srcid"] = cur.lastrowid
            db.generic_insert("journals_text", ("id", "article", "introtext"), kwargs)
            db.generic_insert('firehose', ('type', 'globjid', 'public', 'name', 'sprite_info', 'srcid'), kwargs)
        except NewslashDBError as e:
            db.rollback()
            raise e


        db.commit()
        db.close()
        return kwargs["id"]

    def update(self, **kwargs):
        '''this function implemented for test purpose...'''
        db = self.db.start_transaction()

        try:
            cur = db.execute("UPDATE journals SET description=%(description)s, last_update = NOW()"
                             "  WHERE id=%(id)s", **kwargs)
            #cur = db.generic_update("journals", ('description',), ('id',), kwargs)
            cur = db.generic_update("journals_text", ('article', 'introtext'), ('id',), kwargs)
            if 'public' in kwargs:
                kwargs["srcid"] = kwargs["id"]
                kwargs["type"] = "journal"
                db.generic_update('firehose', ('public',), ('srcid', 'type'), kwargs)
        except NewslashDBError as e:
            db.rollback()
            raise e

        db.commit()
        db.close()
        return


    
