# coding: utf-8
'''stories.py: newslash story related database module'''

from base import NewslashDB, NewslashDBError

CREATE_STORIES_TABLE_SQL = '''
CREATE TABLE IF NOT EXISTS `stories` (
`stoid`               mediumint(8) unsigned NOT NULL AUTO_INCREMENT,
`sid`                 varchar(16)           NOT NULL DEFAULT '',
`uid`                 mediumint(8) unsigned NOT NULL DEFAULT '0',
`dept`                varchar(100)          NOT NULL DEFAULT '',
`time`                datetime              NOT NULL DEFAULT '1900-01-01 00:00:00',
`hits`                mediumint(8) unsigned NOT NULL DEFAULT '0',
`discussion`          mediumint(8) unsigned DEFAULT NULL,
`primaryskid`         smallint(6)           NOT NULL DEFAULT '1',
`tid`                 int(11)               NOT NULL DEFAULT '49',
`submitter`           mediumint(8) unsigned NOT NULL DEFAULT '0',
`commentcount`        smallint(5) unsigned  NOT NULL DEFAULT '0',
`hitparade`           varchar(64)           NOT NULL DEFAULT '0,0,0,0,0,0,0',
`is_archived`         enum('no','yes')      NOT NULL DEFAULT 'no',
`in_trash`            enum('no','yes')      NOT NULL DEFAULT 'no',
`day_published`       date                  NOT NULL DEFAULT '1900-01-01',
`qid`                 mediumint(8) unsigned DEFAULT NULL,
`last_update`         timestamp             NOT NULL DEFAULT CURRENT_TIMESTAMP ON UPDATE CURRENT_TIMESTAMP,
`body_length`         mediumint(8) unsigned NOT NULL DEFAULT '0',
`word_count`          mediumint(8) unsigned NOT NULL DEFAULT '0',
`archive_last_update` datetime              NOT NULL DEFAULT '1970-01-01 00:00:00',
`sponsor`             tinyint(3) unsigned   DEFAULT '0',
`stuck`               enum('no','yes')      DEFAULT 'no',
`stuckpos`            tinyint(3) unsigned   DEFAULT '0',
`stuckendtime`        datetime              NOT NULL DEFAULT '1900-01-01 00:00:00',
PRIMARY KEY             (`stoid`),
UNIQUE KEY `sid`        (`sid`),
KEY `uid`               (`uid`),
KEY `is_archived`       (`is_archived`),
KEY `time`              (`time`),
KEY `submitter`         (`submitter`),
KEY `day_published`     (`day_published`),
KEY `skidtid`           (`primaryskid`,`tid`),
KEY `discussion_stoid`  (`discussion`,`stoid`),
KEY `time_stoid`        (`time`,`stoid`),
KEY `trash_time`        (`in_trash`,`time`),
KEY `ibfk_converttid_4` (`tid`),
KEY `stuckendtime`      (`stuckendtime`)
) ENGINE=InnoDB DEFAULT CHARSET=utf8mb4 ROW_FORMAT=DYNAMIC
'''

CREATE_STORY_TEXT_TABLE_SQL = '''
CREATE TABLE IF NOT EXISTS `story_text` (
`stoid`       mediumint(8) unsigned NOT NULL DEFAULT '0',
`title`       varchar(100)          NOT NULL DEFAULT '',
`introtext`   mediumtext,
`bodytext`    mediumtext,
`relatedtext` mediumtext,
`rendered`    mediumtext,
PRIMARY KEY (`stoid`)
) ENGINE=InnoDB DEFAULT CHARSET=utf8mb4 ROW_FORMAT=DYNAMIC
'''

CREATE_STORY_PARAM_TABLE_SQL = '''
CREATE TABLE IF NOT EXISTS `story_param` (
      `param_id` mediumint(8) unsigned NOT NULL AUTO_INCREMENT,
      `stoid`    mediumint(8) unsigned NOT NULL DEFAULT '0',
      `name`     varchar(32)           NOT NULL DEFAULT '',
      `value`    mediumtext            NOT NULL,
      PRIMARY KEY (`param_id`),
      UNIQUE KEY  `story_key` (`stoid`,`name`)
    ) ENGINE=InnoDB DEFAULT CHARSET=utf8mb4 ROW_FORMAT=DYNAMIC
'''

class Stories(NewslashDB):
    def __init__(self, config):
        super(Stories, self).__init__(config)

    def create_tables(self):
        '''this function implemented for test purpose...'''
        self.start_transaction()

        try:
            self.execute(CREATE_STORIES_TABLE_SQL)
            self.execute(CREATE_STORY_TEXT_TABLE_SQL)
            self.execute(CREATE_STORY_PARAM_TABLE_SQL)
        except NewslashDBError as e:
            self.rollback()
            raise e

        self.commit()
        self.close()

    def drop_tables(self):
        '''this function implemented for test purpose...'''
        self.start_transaction()

        try:
            self.execute('DROP TABLE stories')
            self.execute('DROP TABLE story_text')
            self.execute('DROP TABLE story_param')
        except NewslashDBError as e:
            self.rollback()
            raise e

        self.commit()
        self.close()

    def insert(self, **kwargs):
        '''this function implemented for test purpose...'''
        sql_stories = ('INSERT INTO stories (sid, uid, dept, time)'
                       '  VALUES (%(sid)s, %(uid)s, %(dept)s, %(time)s)')
        sql_story_text = ('INSERT INTO story_text (stoid, title, introtext, bodytext, relatedtext, rendered)'
                          '  VALUES (%(stoid)s, %(title)s, %(introtext)s, "", "", "")')

        self.start_transaction()

        try:
            cur = self.execute(sql_stories, **kwargs)
            kwargs["stoid"] = cur.lastrowid
            self.execute(sql_story_text, **kwargs)
        except NewslashDBError as e:
            self.rollback()
            raise e

        self.commit()
        self.close()
        return kwargs["stoid"]

    def update(self, **kwargs):
        '''this function implemented for test purpose...'''
        
        sql_stories = ('UPDATE stories'
                       '  SET uid=%(uid)s, dept=%(dept)s, last_update=NOW()'
                       '  WHERE stoid=%(stoid)s')
        sql_story_text = ('UPDATE story_text'
                          '  SET title=%(title)s, introtext=%(introtext)s'
                          '  WHERE stoid=%(stoid)s')
        sql_story_param = ('INSERT INTO story_param'
                           '  (stoid, name, value)'
                           '  VALUE (%(stoid)s, "neverdisplay", %(neverdisplay)s)'
                           '  ON DUPLICATE KEY UPDATE value = %(neverdisplay)s')

        self.start_transaction()

        try:
            self.execute(sql_stories, **kwargs)
            self.execute(sql_story_text, **kwargs)
            if "neverdisplay" in kwargs:
                self.execute(sql_story_param, **kwargs)

        except NewslashDBError as e:
            self.rollback()
            raise e

        self.commit()
        self.close()
        return

    def select(self, **kwargs):
        if not "limit" in kwargs:
            kwargs["limit"] = self.config("default_limit")

        if not "offset" in kwargs:
            kwargs["offset"] = 0

        if "stoid" in kwargs:
            query = ("SELECT stories.*, story_text.*, story_param.value AS neverdisplay FROM stories"
                     "  LEFT JOIN story_text USING(stoid)"
                     '  LEFT JOIN story_param ON story_param.name = "neverdisplay"'
                     '                          AND story_text.stoid = story_param.stoid'
                     "  WHERE stories.stoid = %(stoid)s AND stories.time <= NOW()"
                     "  LIMIT %(limit)s OFFSET %(offset)s")
        elif "last_update_ge" in kwargs:
            query = ("SELECT stories.*, story_text.*, story_param.value AS neverdisplay FROM stories"
                     "  LEFT JOIN story_text USING(stoid)"
                     '  LEFT JOIN story_param ON story_param.name = "neverdisplay"'
                     '                          AND story_text.stoid = story_param.stoid'
                     "  WHERE stories.stoid <= %(stoid_le)s AND stories.last_update >= %(last_update_ge)s"
                     "  ORDER BY stories.stoid ASC LIMIT %(limit)s OFFSET %(offset)s")
        elif "stoid_gt" in kwargs:
            query = ("SELECT stories.*, story_text.*, story_param.value AS neverdisplay FROM stories"
                     "  LEFT JOIN story_text USING(stoid)"
                     '  LEFT JOIN story_param ON story_param.name = "neverdisplay"'
                     '                          AND story_text.stoid = story_param.stoid'
                     "  WHERE stories.stoid > %(stoid_gt)s AND stories.time <= NOW()"
                     "  ORDER BY stories.stoid ASC LIMIT %(limit)s OFFSET %(offset)s")
            
        else:
            query = ("SELECT stories.*, story_text.*, story_param.value AS neverdisplay FROM stories"
                     "  LEFT JOIN story_text USING(stoid)"
                     '  LEFT JOIN story_param ON story_param.name = "neverdisplay"'
                     '                          AND story_text.stoid = story_param.stoid'
                     '  WHERE stories.time <= NOW()'
                     "  ORDER BY stories.stoid DESC LIMIT %(limit)s OFFSET %(offset)s")

        cur = self.execute(query, **kwargs)
        col_names = cur.column_names
        result = [dict(zip(col_names, x)) for x in cur]

        self.close()
        return result


    
