# coding: utf-8
'''users.py: newslash users related database module'''

import random
from base import NewslashDB, NewslashDBError

CREATE_USERS_TABLE_SQL = '''
CREATE TABLE IF NOT EXISTS `users` (
  `uid`                     mediumint(8) unsigned NOT NULL AUTO_INCREMENT,
  `nickname`                varchar(20)           NOT NULL DEFAULT '',
  `realemail`               varchar(50)           NOT NULL DEFAULT '',
  `fakeemail`               varchar(75)           DEFAULT NULL,
  `homepage`                varchar(100)          DEFAULT NULL,
  `passwd`                  varchar(32)           NOT NULL DEFAULT '',
  `sig`                     varchar(200)          DEFAULT NULL,
  `seclev`                  mediumint(8) unsigned NOT NULL DEFAULT '0',
  `matchname`               varchar(20)           DEFAULT NULL,
  `newpasswd`               varchar(32)           DEFAULT '',
  `newpasswd_ts`            datetime              DEFAULT NULL,
  `journal_last_entry_date` datetime              DEFAULT NULL,
  `author`                  tinyint(4)            NOT NULL DEFAULT '0',
  `shill_id`                tinyint(3) unsigned   NOT NULL DEFAULT '0',
  PRIMARY KEY           (`uid`),
  KEY `login`           (`uid`,`passwd`,`nickname`),
  KEY `chk4user`        (`nickname`,`realemail`),
  KEY `nickname_lookup` (`nickname`),
  KEY `chk4email`       (`realemail`),
  KEY `chk4matchname`   (`matchname`),
  KEY `author_lookup`   (`author`),
  KEY `nickname_uid`    (`nickname`,`uid`),
  KEY `seclev`          (`seclev`)
) ENGINE=InnoDB DEFAULT CHARSET=utf8mb4 ROW_FORMAT=DYNAMIC
'''

CREATE_USERS_INFO_TABLE_SQL = '''
CREATE TABLE IF NOT EXISTS `users_info` (
  `uid`                 mediumint(8) unsigned                     NOT NULL DEFAULT '0',
  `totalmods`           mediumint(9)                              NOT NULL DEFAULT '0',
  `realname`            varchar(50)                               DEFAULT NULL,
  `bio`                 mediumtext                                NOT NULL,
  `tokens`              mediumint(9)                              NOT NULL DEFAULT '0',
  `lastgranted`         datetime                                  NOT NULL DEFAULT '1900-01-01 00:00:00',
  `m2info`              varchar(64)                               NOT NULL DEFAULT '',
  `karma`               mediumint(9)                              NOT NULL DEFAULT '0',
  `maillist`            tinyint(4)                                NOT NULL DEFAULT '0',
  `totalcomments`       mediumint(8) unsigned                     DEFAULT '0',
  `lastm2`              datetime                                  NOT NULL DEFAULT '1970-01-01 00:00:00',
  `m2_mods_saved`       varchar(120)                              NOT NULL DEFAULT '',
  `lastaccess`          date                                      NOT NULL DEFAULT '1970-01-01',
  `m2fair`              mediumint(8) unsigned                     NOT NULL DEFAULT '0',
  `up_fair`             mediumint(8) unsigned                     NOT NULL DEFAULT '0',
  `down_fair`           mediumint(8) unsigned                     NOT NULL DEFAULT '0',
  `m2unfair`            mediumint(8) unsigned                     NOT NULL DEFAULT '0',
  `up_unfair`           mediumint(8) unsigned                     NOT NULL DEFAULT '0',
  `down_unfair`         mediumint(8) unsigned                     NOT NULL DEFAULT '0',
  `m2fairvotes`         mediumint(8) unsigned                     NOT NULL DEFAULT '0',
  `m2voted_up_fair`     mediumint(8) unsigned                     NOT NULL DEFAULT '0',
  `m2voted_down_fair`   mediumint(8) unsigned                     NOT NULL DEFAULT '0',
  `m2unfairvotes`       mediumint(8) unsigned                     NOT NULL DEFAULT '0',
  `m2voted_up_unfair`   mediumint(8) unsigned                     NOT NULL DEFAULT '0',
  `m2voted_down_unfair` mediumint(8) unsigned                     NOT NULL DEFAULT '0',
  `m2voted_lonedissent` mediumint(8) unsigned                     NOT NULL DEFAULT '0',
  `m2voted_majority`    mediumint(8) unsigned                     NOT NULL DEFAULT '0',
  `upmods`              mediumint(8) unsigned                     NOT NULL DEFAULT '0',
  `downmods`            mediumint(8) unsigned                     NOT NULL DEFAULT '0',
  `stirred`             mediumint(8) unsigned                     NOT NULL DEFAULT '0',
  `session_login`       tinyint(4)                                NOT NULL DEFAULT '0',
  `cookie_location`     enum('classbid','subnetid','ipid','none') NOT NULL DEFAULT 'none',
  `registered`          tinyint(3) unsigned                       NOT NULL DEFAULT '1',
  `reg_id`              varchar(32)                               NOT NULL DEFAULT '',
  `expiry_days`         mediumint(9)                              NOT NULL DEFAULT '1',
  `expiry_comm`         mediumint(9)                              NOT NULL DEFAULT '1',
  `user_expiry_days`    smallint(5) unsigned                      NOT NULL DEFAULT '1',
  `user_expiry_comm`    smallint(5) unsigned                      NOT NULL DEFAULT '1',
  `initdomain`          varchar(30)                               NOT NULL DEFAULT '',
  `created_ipid`        varchar(32)                               NOT NULL DEFAULT '',
  `created_at`          datetime                                  NOT NULL DEFAULT '1900-01-01 00:00:00',
  `tag_clout`           float unsigned                            NOT NULL DEFAULT '1',
  `people`              mediumblob,                               
  `people_status`       enum('ok','dirty')                        NOT NULL DEFAULT 'ok',
  `csq_bonuses`         float                                     NOT NULL DEFAULT '0',
  PRIMARY KEY         (`uid`),
  KEY `people_status` (`people_status`),
  KEY `initdomain`    (`initdomain`),
  KEY `created_ipid`  (`created_ipid`)
) ENGINE=InnoDB DEFAULT CHARSET=utf8mb4 ROW_FORMAT=DYNAMIC
'''

class Users(NewslashDB):
    def __init__(self, config):
        self.db = NewslashDB(config)

    def select(self, **kwargs):
        if not "limit" in kwargs:
            kwargs["limit"] = self.db.config("default_limit")

        if not "offset" in kwargs:
            kwargs["offset"] = 0

        if "uid" in kwargs:
            query = ("SELECT users.*, users_info.bio, users_info.created_at, users_info.realname FROM users"
                     "  LEFT JOIN users_info USING(uid)"
                     "  WHERE users.uid = %(uid)s"
                     "  LIMIT %(limit)s OFFSET %(offset)s")
        elif "last_update_ge" in kwargs:
            query = ("SELECT users.*, users_info.bio, users_info.created_at, users_info.realname FROM users"
                     "  LEFT JOIN users_info USING(uid)"
                     "  WHERE users.uid <= %(uid_le)s AND users_info.created_at >= %(last_update_ge)s"
                     "  ORDER BY users.uid ASC LIMIT %(limit)s OFFSET %(offset)s")
        elif "uid_gt" in kwargs:
            query = ("SELECT users.*, users_info.bio, users_info.created_at, users_info.realname FROM users"
                     "  LEFT JOIN users_info USING(uid)"
                     "  WHERE users.uid > %(uid_gt)s"
                     "  ORDER BY users.uid ASC LIMIT %(limit)s OFFSET %(offset)s")
        else:
            query = ("SELECT users.*, users_info.bio, users_info.created_at, users_info.realname FROM users"
                     "  LEFT JOIN users_info USING(uid)"
                     "  ORDER BY users.uid DESC LIMIT %(limit)s OFFSET %(offset)s")

        cur = self.db.execute(query, **kwargs)
        col_names = cur.column_names
        result = [dict(zip(col_names, x)) for x in cur]

        self.db.close()
        return result

    def create_tables(self):
        '''this function implemented for test purpose...'''
        self.db.execute_multi((CREATE_USERS_TABLE_SQL,
                               CREATE_USERS_INFO_TABLE_SQL))

    def drop_tables(self):
        '''this function implemented for test purpose...'''
        self.db.execute_multi(('DROP TABLE users',
                               'DROP TABLE users_info'))

    def insert(self, **kwargs):
        '''this function implemented for test purpose...'''
        db = self.db.start_transaction()
        try:
            cur = db.generic_insert("users", ("nickname", "homepage", "sig"), kwargs)
            kwargs["uid"] = cur.lastrowid
            db.generic_insert("users_info", ("uid", "bio"), kwargs)
        except NewslashDBError as e:
            db.rollback()
            raise e

        db.commit()
        return kwargs["uid"]



    
