# -*- coding: utf-8 -*-
"""index: index model for newslash search system"""

import calendar
from datetime import datetime
import os.path

import exceptions
import lucene_wrapper
import newslash_db
import htmlutil
from document import Document
from metadata import Metadata

class Index(object):
    def __init__(self, **kwargs):
        if "config" in kwargs:
            self.config = kwargs["config"]
        else:
            self.config = {}

        for k in kwargs:
            if k == "config":
                continue
            self.config[k] = kwargs[k]

        self.metadata = Metadata(self.config)
        p = self.config.get("index_path")
        self._index_path = os.path.join(p, "lucene_index");


    def _get_searcher(self):
        return lucene_wrapper.Searcher(index_directory=self._index_path)

    def _db_config(self):
        db_cfg = self.config.get("database")
        if db_cfg is None:
            raise ConfigFileError(ConfigFileError.SECTION_NOT_FOUND, "Database")
        return db_cfg

    def delete(self, target, unique_id, indexer=None):
        '''delete document from index'''
        # create query to identify target document
        q = lucene_wrapper.BooleanQuery()
        q.add_must(lucene_wrapper.TermQuery("type", target))
        q.add_must(lucene_wrapper.IntRangeQuery("unique_id", unique_id))
        
        # check if deletable
        #searcher = self._get_searcher()
        #result = searcher.search(q)

        # delete document
        if indexer: 
            indexer.delete(q)
        else:
            with lucene_wrapper.Indexer(index_directory=self._index_path) as indexer:
                indexer.delete(q)

    def delete_all(self):
        '''delete all document from index'''
        with lucene_wrapper.Indexer(index_directory=self._index_path) as indexer:
            indexer.delete_all()

    def _get_unique_id(self, target, item):
        if target == "user":
            return item["uid"]
        if target == "poll":
            return item["qud"]
        if target == "submission":
            return item["subid"]
        if target == "journal":
            return item["id"]
        if target == "comment":
            return item["cid"]
        if target == "story":
            return item["stoid"]
        return 0
        
    def _update(self, selector, target, batch_size=1000, progress_cb=None, error_cb=None):
        query_done = False

        # at first, get last indexed id  and timestamp
        metadata = self.metadata.get(target)
        latest_id = metadata.get('latest_id', 0)
        last_update = metadata.get('last_update')

        # add document to index
        # don't use datetime.now(), because Database server's timestamp may be not this server's one.
        start_update = self.metadata.get_current_timestamp()
        add_success = 0
        add_errors = 0
        max_unique_id = latest_id

        force_exit = False
        with lucene_wrapper.Indexer(index_directory=self._index_path) as indexer:
            try:
                while not query_done:
                    items = selector(batch_size, 0, max_unique_id)
                    if len(items) == 0:
                        query_done = True
                        break

                    for item in items:
                        try:
                            doc = Document(target, item)
                        except exceptions.DocumentMakingError:
                            if self._get_unique_id(target, item) > max_unique_id:
                                max_unique_id = self._get_unique_id(target, item)
                            add_errors += 1
                            if error_cb is not None:
                                error_cb(target, 'add', item)
                            continue

                        if not doc.indexable:
                            add_success += 1
                            continue

                        try:
                            indexer.add(doc.doc)
                        except KeyboardInterrupt:
                            # revert insertion
                            self.delete(target, doc.unique_id, index)
                            raise

                        try:
                            add_success += 1
                            if doc.unique_id > max_unique_id:
                                max_unique_id = doc.unique_id
                        except KeyboardInterrupt:
                            if doc.unique_id > max_unique_id:
                                max_unique_id = doc.unique_id
                            raise

                    if progress_cb is not None:
                        progress_cb(target, 'add', add_success, add_errors)

            except KeyboardInterrupt:
                force_exit = True

        # update metadata
        self.metadata.update(target=target, last_update=start_update, latest_id=max_unique_id)

        if force_exit:
            print("interrupted.")
            raise KeyboardInterrupt

        # if no previous update, done
        if last_update is None:
            return (add_success, add_errors)

        # update index for updated stories
        update_success = 0
        update_errors = 0
        offset = 0
        with lucene_wrapper.Indexer(index_directory=self._index_path) as indexer:
            items = selector(batch_size, offset, latest_id, last_update)
            offset += len(items)
            if len(items) < batch_size:
                query_done = True

            for item in items:
                # at first, create new document
                try:
                    doc = Document(target, item)
                except exceptions.DocumentMakingError:
                    update_errors += 1
                    if error_cb is not None:
                        error_cb(target, 'update', item)
                    continue

                # create query to identify target document
                self.delete(target, doc.unique_id, indexer) 

                # add document
                if doc.indexable:
                    indexer.add(doc.doc)

                update_success += 1

        if progress_cb is not None:
            progress_cb(target, 'update', update_success, update_errors)

        success = add_success + update_success
        errors = add_errors + update_errors

        # done
        return (success, errors)

    def update_story(self, batch_size=1000, progress_cb=None, error_cb=None):
        '''update story index'''
        stories = newslash_db.Stories(self._db_config())
        def selector(limit, offset, latest_id, last_update=None):
            if last_update:
                return stories.select(limit=limit, stoid_le=latest_id, last_update_ge=last_update)
            else:
                return stories.select(limit=limit, offset=offset, stoid_gt=latest_id)

        return self._update(selector, 'story', batch_size, progress_cb, error_cb)
            
    def update_comment(self, batch_size=1000, progress_cb=None, error_cb=None):
        '''update comment index'''
        comments = newslash_db.Comments(self._db_config())
        def selector(limit, offset, latest_id, last_update=None):
            if last_update:
                return comments.select(limit=limit, cid_le=latest_id, last_update_ge=last_update)
            else:
                return comments.select(limit=limit, offset=offset, cid_gt=latest_id)

        return self._update(selector, 'comment', batch_size, progress_cb, error_cb)

    def update_journal(self, batch_size=1000, progress_cb=None, error_cb=None):
        '''update journal index'''
        journals = newslash_db.Journals(self._db_config())
        def selector(limit, offset, latest_id, last_update=None):
            if last_update:
                return journals.select(limit=limit, id_le=latest_id, last_update_ge=last_update)
            else:
                return journals.select(limit=limit, offset=offset, id_gt=latest_id)

        return self._update(selector, 'journal', batch_size, progress_cb, error_cb)

    def update_submission(self, batch_size=1000, progress_cb=None, error_cb=None):
        '''update submission index'''
        submissions = newslash_db.Submissions(self._db_config())
        def selector(limit, offset, latest_id, last_update=None):
            if last_update:
                return submissions.select(limit=limit, subid_le=latest_id, last_update_ge=last_update)
            else:
                return submissions.select(limit=limit, offset=offset, subid_gt=latest_id)

        return self._update(selector, 'submission', batch_size, progress_cb, error_cb)

    def update_poll(self, batch_size=1000, progress_cb=None, error_cb=None):
        '''update poll index'''
        polls = newslash_db.Polls(self._db_config())
        def selector(limit, offset, latest_id, last_update=None):
            if last_update:
                return polls.select(limit=limit, qid_le=latest_id, date_ge=last_update)
            else:
                return polls.select(limit=limit, offset=offset, qid_gt=latest_id)

        return self._update(selector, 'poll', batch_size, progress_cb, error_cb)

    def update_user(self, batch_size=1000, progress_cb=None, error_cb=None):
        '''update user index'''
        users = newslash_db.Users(self._db_config())
        def selector(limit, offset, latest_id, last_update=None):
            if last_update:
                return []
            else:
                return users.select(limit=limit, offset=offset, uid_gt=latest_id)

        return self._update(selector, 'user', batch_size, progress_cb, error_cb)

    def update_all(self, batch_size=10000, progress_cb=None, error_cb=None):
        '''update index'''
        results = []
        success = 0
        errors = 0

        results.append(self.update_story(batch_size, progress_cb, error_cb))
        print("update story index done.")

        results.append(self.update_comment(batch_size, progress_cb, error_cb))
        print("update comment index done.")

        results.append(self.update_journal(batch_size, progress_cb, error_cb))
        print("update journal index done.")

        results.append(self.update_submission(batch_size, progress_cb, error_cb))
        print("update submission index done.")

        results.append(self.update_poll(batch_size, progress_cb, error_cb))
        print("update poll index done.")

        results.append(self.update_user(batch_size, progress_cb, error_cb))
        print("update user index done.")

        for i in range(0, len(results)):
            success += results[i][0]
            errors += results[i][1]

        # done
        return (success, errors)

    def get(self, item_type, item_id):
        '''get document match iten_type and item_id from index'''
        searcher = self._get_searcher()

        type_query = lucene_wrapper.TermQuery("type",item_type)
        id_query = lucene_wrapper.TermQuery("id", item_id)

        query = lucene_wrapper.BooleanQuery()
        query.add_must(id_query)
        query.add_must(type_query)
        
        result = searcher.search(query)
        if result.total_hits == 0:
            return None
        return result[0]
        
    def query(self, query_string):
        '''query index'''
        searcher = self._get_searcher()
        try:
            query = lucene_wrapper.Query("content_text", query_string)
        except lucene_wrapper.QueryParseError as e:
            sys.stderr.write("query parse error\n")
            return

        result = searcher.search(query)
        return result
