# -*- coding: utf-8
'''router: Routing and preprocess system for Web applications'''

import httplib
import json
import sys

class Response(object):
    def __init__(self, start_response):
        self.status = None
        self.response_headers = []
        self.finished = False
        self.response_bodies = []
        self.start_response = start_response

    def add_response(self, response):
        self.response_bodies.append(response)

    def add_header(self, name, value):
        self.response_headers.append((name, value))

    def set_status_code(self, code):
        self.status = "{} {}".format(code, httplib.responses[code])

    def render_json(self, data):
        self.add_header("Content-Type", "application/json")
        raw_json = json.dumps(data, ensure_ascii=False)
        self.add_response(raw_json.encode('utf-8'))

    def render(self, code=None, **kwargs):
        if code is not None:
            self.set_status_code(code)
        if 'json' in kwargs:
            return self.render_json(kwargs["json"])

    def finish(self, code=None, body=None):
        if code is not None:
            self.set_status_code(code)

        if self.status is None:
            self.set_status_code(200)

        if body is not None:
            self.add_response(body)

        self.start_response(self.status, self.response_headers)
        self.finished = True


class Request(object):
    def __init__(self, environ):
        self.environ = environ
        self.body = self.decode_request_body()

    def decode_request_body(self):
        content_type = self.environ.get('CONTENT_TYPE', "")
        if content_type == 'application/json':
            return self._decode_json()
        return None

    def _decode_json(self):
        body_len = int(self.environ.get('CONTENT_LENGTH', 0))
        if body_len > 0:
            raw_req = self.environ['wsgi.input']
            raw_json = raw_req.read(body_len)
            return json.loads(raw_json)
        return None


class Route(object):
    pass


class Router(object):
    def __init__(self, environ, start_response):
        self.routes = []
        self.environ = environ
        self.start_response = start_response

    def __iter__(self):
        environ = self.environ
        start_response = self.start_response

        req = Request(environ)
        resp = Response(start_response)

        # find route
        path_info = environ['PATH_INFO']
        receiver = self._default_route
        for (rex, recv) in self.routes:
            if rex.search(path_info):
                receiver = recv

        # routing by method
        method = environ['REQUEST_METHOD']
        try:
            if method == 'GET':
                receiver.get(req, resp)
            elif method == 'POST':
                receiver.post(req, resp)
            else:
                receiver.fallback(req, resp)
        except AttributeError:
            self.fallback(req, resp)
        except Exception as e:
            req.environ['wsgi.errors'].write("exception occured: {}".format(e))
            self.fallback(req, resp)

        if not resp.finished:
            resp.finish()

        # done
        for res in resp.response_bodies:
            yield res

    def route(self, rex, receiver):
        self.routes.append((rex, receiver))

    def default_route(self, receiver):
        self._default_route = receiver

    def fallback(self, req, resp):
        resp.finish(500, "Internal Server Error")

        

