#!/usr/bin/python
# -*- coding: utf-8
from __future__ import print_function

import argparse
import os.path
import os
import sys
import time
import inspect

from yaml import load
try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper

import lucene_wrapper
from newslash_index import Index, DatabaseError
from newslash_db import NewslashDB

class SearchCLIError(Exception):
    def __init__(self, message):
        self.message = message

class SearchCLI(object):
    def __init__(self):
        self.index_dir = None
        self._parse_args()

        config_path = os.environ.get("SEARCHD_CONFIG", "/etc/newslash/searchd.conf")
        if not os.path.exists(config_path):
            base_dir = os.path.dirname(os.path.realpath(__file__))
            config_path = os.path.join(base_dir, 'searchd.conf')
        if not os.path.exists(config_path):
            raise SearchCLIError("config file not found")

        self._load_config(config_path)

    def _make_parser(self):
        # parse command line option
        parser = argparse.ArgumentParser(description='search daemon for Newslash')
        parser.add_argument('-i', '--index-dir', help='lucene index directory')

        # subcommands
        subparsers = parser.add_subparsers(help="subcommands")

        # 'query' subcommand
        p_query = subparsers.add_parser("query", help="execute query")
        p_query.add_argument('subcommand', action='store_const', const='query')
        p_query.add_argument('query_string', help='query string')

        # 'index' subcommand
        p_index = subparsers.add_parser("index", help="manipulate lucene index")
        p_index.add_argument('subcommand', action='store_const', const='index')
        p_index.add_argument('action', help='action')
        p_index.add_argument('target', help='target', default='all')

        # 'getdocument' subcommand
        p_getdoc = subparsers.add_parser("getdocument", help="get document from lucene index")
        p_getdoc.add_argument('subcommand', action='store_const', const='getdocument')
        p_getdoc.add_argument('target_type', help='target type')
        p_getdoc.add_argument('target_id', help='target id')

        # 'initdb' subcommand
        p_initdb = subparsers.add_parser("initdb", help="create table which stores index related information")
        p_initdb.add_argument('subcommand', action='store_const', const='initdb')

        # 'analyze' subcommand
        p_analyze = subparsers.add_parser("analyze", help="analyze index")
        p_analyze.add_argument('subcommand', action='store_const', const='analyze')
        p_analyze.add_argument('query_string', help='query string')
        p_analyze.add_argument('-n', '--number-of-result', default=10, type=int, help='number of result output')

        # 'metadata' subcommand
        p_metadata = subparsers.add_parser("metadata", help="manipulate metadata")
        p_metadata.add_argument('subcommand', action='store_const', const='metadata')
        p_metadata.add_argument('action', help='action')

        return parser

    def _parse_args(self):
        self.conf = {}
        parser = self._make_parser()
        self.args = parser.parse_args()

        # index directory
        if self.args.index_dir:
            self.index_dir = os.path.abspath(self.args.index_dir)

        # command
        try:
            self.sub_command = self.args.subcommand or ""
        except AttributeError:
            self.sub_command = ""
            return

    def _load_config(self, pathname):
        try:
            fh = open(pathname)
        except IOError:
            sys.stderr.write('config file ({}) not found...'.format(pathname))
            self.config = {}
            return

        self.config = load(fh, Loader=Loader)
        fh.close()

        d = self.config.get("Searchd", {})
        try:
            self.index_dir = os.path.abspath(d["index_path"])
        except KeyError:
            raise SearchCLIError("index_path not given")

        self.database = self.config.get("Database", {})
            
    def show_help(self):
        parser = self._make_parser()
        parser.print_help()

    def metadata(self):
        if self.args.action == 'show':
            index = Index(database=self.database, index_path=self.index_dir)
            for data in index.metadata.get_all():
                print(data)

    def index(self):
        action = self.args.action
        target = self.args.target

        lucene_wrapper.init_vm()
        index = Index(database=self.database, index_path=self.index_dir)

        def progress_cb(target, phase, success, errors):
            print("{}: {} to index {} items... ({} errors)".format(target, phase, success, errors))

        def error_cb(target, phase, item):
            print("{}: indexing {} error: id={}".format(target, phase, item["sid"]))

        if action == 'clear':
            index.delete_all()
            index.metadata.delete_all()
            print("clear all index and metadata done.")
        elif action == 'update':
            start_time = time.time()
            if target == 'all':
                index.update_all(progress_cb=progress_cb, error_cb=error_cb)
            elif target == 'story':
                index.update_story(progress_cb=progress_cb, error_cb=error_cb)
            elif target == 'comment':
                index.update_comment(progress_cb=progress_cb, error_cb=error_cb)
            elif target == 'journal':
                index.update_journal(progress_cb=progress_cb, error_cb=error_cb)
            elif target == 'submission':
                index.update_submission(progress_cb=progress_cb, error_cb=error_cb)
            elif target == 'poll':
                index.update_poll(progress_cb=progress_cb, error_cb=error_cb)
            elif target == 'user':
                index.update_user(progress_cb=progress_cb, error_cb=error_cb)
            print("indexing done. total time: {}s".format(time.time() - start_time))

    def query(self):
        if self.index_dir is None:
            sys.stderr.write("error: index directory not given\n")
            return

        query_string = self.args.query_string
        lucene_wrapper.init_vm()
        
        searcher = lucene_wrapper.Searcher(index_directory=self.index_dir)
        try:
            query = lucene_wrapper.Query("content_text", query_string)
        except lucene_wrapper.QueryParseError as e:
            sys.stderr.write("query parse error\n")
            return

        result = searcher.search(query)
       
        print("total hits: {}".format(result.total_hits))
        for item in result:
            print("#{} - {}: {}".format(item.number, item.id, item.content_text.encode('utf-8')))

    def analyze(self):
        if self.index_dir is None:
            sys.stderr.write("error: index directory not given\n")
            return

        query_string = self.args.query_string
        limit = self.args.number_of_result

        lucene_wrapper.init_vm()
        
        searcher = lucene_wrapper.Searcher(index_directory=self.index_dir)
        try:
            content_query = lucene_wrapper.Query("content_text", query_string)
            title_query = lucene_wrapper.Query("title", query_string)
            query = lucene_wrapper.BooleanQuery()
            query.set_minimum_nubmber_should_match(1)
            query.add_should(content_query)
            query.add_should(title_query)
        except lucene_wrapper.QueryParseError as e:
            sys.stderr.write("query parse error\n")
            return

        sort = lucene_wrapper.Sort("create_time", lucene_wrapper.Sort.INT, True)
        result = searcher.search(query, limit, 0, sort)
        
        print("total hits: {}".format(result.total_hits))
        for item in result:
            print("#{} - {}:".format(item.number, item.id))

            # get fields
            fields = item.get_fields()
            for field in fields:
                if field.name == "content_text":
                    print("  {}: {}".format(field.name, field.value.encode("utf8")))
                    print("----")
                    for term in field.get_tokens():
                        sys.stdout.write("{} ".format(term.encode("utf8")))
                    print("----")
                else:
                    print("  {}: {}".format(field.name, field.value.encode("utf8")))
            
            print("\n")

    def getdocument(self):
        lucene_wrapper.init_vm()
        index = Index(database=self.database, index_path=self.index_dir)
        result = index.get(self.args.target_type, self.args.target_id)
        if result is None:
            print("no item")
            return

        if result.totalhits > 1:
            print("warning: hits multiple items")

        print("#{} - {}: {}".format(result.number, result.id, result.content_text.encode('utf-8')))

    def initdb(self):
        index = Index(database=self.database, index_path=self.index_dir)
        try:
            index.metadata.create_table()
        except DatabaseError as e:
            print('error: {}'.format(str(e)))
            
    def run(self):
        if self.sub_command == 'query':
            return self.query()

        if self.sub_command == 'index':
            return self.index()

        if self.sub_command == 'getdocument':
            return self.getdocument()

        if self.sub_command == 'initdb':
            return self.initdb()

        if self.sub_command == 'analyze':
            return self.analyze()

        if self.sub_command == 'metadata':
            return self.metadata()

        return self.show_help()


if __name__ == '__main__':
    cli = SearchCLI()
    cli.run()
